//==============================================================================
//
//    lossyWAV: Added noise WAV bit reduction method by David Robinson;
//              Noise shaping coefficients by Sebastian Gesemann;
//
//    Copyright (C) 2007-2011 Nick Currie, Copyleft.
//
//    This program is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program.  If not, see <http://www.gnu.org/licenses/>.
//
//    Contact: lossywav <at> hotmail <dot> co <dot> uk
//
//==============================================================================
unit nCore;
//==============================================================================
interface
//==============================================================================
var
//==============================================================================
  Unit_Data_Packer1 : Packed Record
    Data : Array[0..3] of Integer;
  End;
//==============================================================================
CONST
  version_string='lossyWAV ';

  CRLF                  = #13#10;

  EXPONENT_BIAS         = 1024;

  PRECALC_ANALYSES      =  6;
  BITS_TO_CALCULATE     = 32;

  WORD_SIZE             =  2;
  WORD_SHIFT            =  1;

  SINGLE_SIZE           =  4;
  SINGLE_SHIFT          =  2;

  DOUBLE_SIZE           =  8;
  DOUBLE_SHIFT          =  3;

  COMPLEX_SIZE          = 16;
  COMPLEX_SHIFT         =  4;

  INTEGER_SIZE          =  4;
  INTEGER_SHIFT         =  2;

  CARDINAL_SIZE         =  4;
  CARDINAL_SHIFT        =  2;

  INT64_SIZE            =  8;
  INT64_SHIFT           =  3;

  POINTER_SIZE          =  4;
  POINTER_SHIFT         =  2;

  e                     = 2.7182818284590452353602874713527;    { e               }

  OneOverPi             = 0.31830988618379067153776752674503;   { 1/Pi            }
  OneOverTwoPi          = 0.15915494309189533576888376337251;   { 1/(2*Pi)        }
  HalfPi                = 1.5707963267948966192313216916398;    { Pi/2            }
  Pi                    = 3.1415926535897932384626433832795;    { Pi              }
  TwoPi                 = 6.2831853071795864769252867665590;    { Pi*2            }

  OneOverRoot2          = 0.70710678118654752440084436210485;   { 2^-0.5          }
  RootTwo               = 1.4142135623730950488016887242097;    { 2^0.5           }
  log2_10               = 3.3219280948873623478703194294894;    { 1/log10(2)      }
  log10_2               = 0.30102999566398119521373889472449;   { log10(2)        }
  log10_2x20            = 6.0205999132796239042747778944899;    { 20 * log10(2)   }

  MaxDb                 = 5623413251903.4908039495103977648;    { 10^(255/20)     }

  GoldenRatio           = 1.61803398874989484820458683436564;   { Golden Ratio    }

  Second_div_Day        = 1.1574074074074074074074074074074e-5; {1/(24*60*60)     }

  //============================================================================
  // Core settings relating to WAV limits and processing limit.
  //============================================================================
  MAX_WAVE_SIZE     = 4294967278;
  MAX_CHANNELS      =    8;
  MAX_BLOCK_BITS    =   12;  // LOG2(4096)
  MAX_BLOCK_SIZE    =    1 shl MAX_BLOCK_BITS;
  BUFFER_SIZE       = MAX_CHANNELS*MAX_BLOCK_SIZE*INTEGER_SIZE;
  CHANNEL_BYTE_SIZE = MAX_BLOCK_SIZE*INTEGER_SIZE;

  //============================================================================

  //============================================================================
  // MAX_FFT_LENGTH **MUST** be double (or greater) MAX_BLOCK_SIZE (!!)
  //============================================================================
  MAX_FFT_BIT_LENGTH    = MAX_BLOCK_BITS+1;  //LOG2(8192)
  MAX_FFT_LENGTH        = 1 shl MAX_FFT_BIT_LENGTH;
  MAX_FFT_LENGTH_HALF   = MAX_FFT_LENGTH div 2;
  MAX_FFT_LENGTH_RECIP  = 1/MAX_FFT_LENGTH;
  //============================================================================

  LONG_FFT_LENGTH         = 1024; // for 44.1/48kHz content
  ADDED_FFT_LENGTH_512    =  512; // for 44.1/48kHz content
  ADDED_FFT_LENGTH_256    =  256; // for 44.1/48kHz content
  ADDED_FFT_LENGTH_128    =  128; // for 44.1/48kHz content
  SHORT_FFT_LENGTH        =   64; // for 44.1/48kHz content
  IMPULSE_FFT_LENGTH      =   32; // for 44.1/48kHz content

  LONG_FFT_BIT_LENGTH     =   10;
  SHORT_FFT_BIT_LENGTH    =    6;

  MAX_MUL                 =    8; // shift for 352.4/384kHz content
  MAX_SHIFT               =    3; // shift for 352.4/384kHz content

  MAX_HIST_FFT_SHORT_LENGTH  = SHORT_FFT_LENGTH shl MAX_SHIFT;

  //============================================================================
  // Core Settings.
  //============================================================================
  STATIC_MINIMUM_BITS_TO_KEEP = 6;
  //============================================================================
  // Frequency range over which to check for lowest amplitude signal
  //============================================================================
  LOWER_FREQ_LIMIT      =    20;
  MID_FREQ_LIMIT        =  3700;
  UPPER_FREQ_LIMIT      = 16000;
  HUMAN_FREQ_LIMIT      = 20000;
  //============================================================================

//==============================================================================
type
//==============================================================================
  tDComplex      = Packed Record
    Re, Im : Double;
  End;

  FFT_results_rec = packed record
    analysis  : ShortInt;
    btr       : ShortInt;
    start     : Integer;
    spreading : Double;
  end;

  channel_data_rec = Packed Record
    maximum_bits_to_remove  : Integer;
    bits_to_remove          : Integer;
    bits_lost               : Integer;
    clips                   : Integer;
    min_FFT_result          : FFT_results_rec;
  End;

  FFT_Data_Rec = Packed Record
    bit_length                        : Integer;
    bit_shift_from_max                : Integer;
    length                            : Integer;
    length_m1                         : Integer;
    length_half                       : Integer;
    length_half_m1                    : Integer;
    length_recip                      : Single;
    length_half_recip                 : Single;
    threshold_shift                   : Single;
  End;

  SingleChannelCodecBlock   = array[0..MAX_BLOCK_SIZE-1] of Integer;
  MultiChannelCodecBlock    = array[0..MAX_CHANNELS-1] of SingleChannelCodecBlock;

  MultiChannelCodecBlockPtr = ^MultiChannelCodecBlock;

  Half_FFT_Double_Array          = array[0..MAX_FFT_LENGTH_HALF+1] of Double;
  Half_FFT_Channel_Double_Array  = array[0..MAX_CHANNELS-1] of Half_FFT_Double_Array;

const
  HFDA_Size = SizeOf(Half_FFT_Double_Array);
  HFCDA_Size = SizeOf(Half_FFT_Channel_Double_Array);

//==============================================================================
var

  //============================================================================
  // Array for in-place FFT analysis.
  //============================================================================
  FFT_Array        : Record
  Case Integer of
    0 : (DComplex  : Array[0..MAX_FFT_LENGTH-1]    of tDComplex);
    1 : (DReal     : Array[0..MAX_FFT_LENGTH*2-1]  of Double);
  End;

  //============================================================================
  // Processed FFT Result storage (FFT procedure is in-place).
  //============================================================================
  FFT_raw_result    : array[0..MAX_FFT_LENGTH-1] of Double;
  FFT_last_result   : array[0..MAX_FFT_LENGTH-1] of Double;
  FFT_skew_result   : array[0..MAX_FFT_LENGTH-1] of Double;

  //============================================================================
  // All audio data
  //============================================================================
  BlockData : Record
    Data  : array[0..3] of MultiChannelCodecBlock;  // rotating codec-blocks

    BTRD  : MultiChannelCodecBlock;                 // bit removed codec-block

    CORR  : MultiChannelCodecBlock;                 // correction data block

    Channel_Log2_RMS : Array[0..MAX_CHANNELS-1] of Double;

    Size  : Record
      Prev, Last, This, Next : Integer;
    End;

    CHANNEL_OFFSET : array[0..MAX_CHANNELS-1] of integer;

    Pointer : array[0..3] of MultiChannelCodecBlockPtr;
  End;

  //============================================================================
  // Lookup table for integer reciprocal.
  //============================================================================
  OneOver           : Array[0..MAX_FFT_LENGTH+1] of Double;

  //============================================================================
  // Global variables used by the Units.
  //============================================================================
  REVERSED_NIBBLES  : Array[0..15] of byte = ($00,$08,$04,$0C,$02,$0A,$06,$0E,$01,$09,$05,$0D,$03,$0B,$07,$0F);
  HEXCHARLOOKUP     : Array[0..15] of char = ('0','1','2','3','4','5','6','7','8','9','A','B','C','D','E','F');

  SINGLE_HALF           : Single = 0.50;
  SINGLE_THREE_QUARTERS : Single = 0.75;
  SINGLE_ONE            : Single = 1.00;
  SINGLE_ONE_AND_A_HALF : Single = 1.50;

  db_MaxDB              : double = MaxDB;
  db_lg2x20             : double = log10_2x20;

  LogOutput : Text;

  hSTDIN    : tHandle;
  hSTDOUT   : tHandle;
  hSTDERR   : tHandle;

  ftSTDIN   : Cardinal;
  ftSTDOUT  : Cardinal;
  ftSTDERR  : Cardinal;

  timer : Packed Record
    Frequency : Int64;
    Period    : Double;

    Start     : Int64;
    Stop      : Int64;

    Elapsed   : Double;
  {$IFNDEF USEASM}
    DelayEnd  : Int64;

    Packer    : Int64;
  {$ENDIF}
    StartTime : Double;

    StopTime  : Double;
    DelayTime : Double;
  End;

  Global : Record
    analysis_number        : Integer;

    analysis : Record
      bits          : Array[1..PRECALC_ANALYSES] of Integer;
      length        : Array[1..PRECALC_ANALYSES] of Integer;
    End;

    FFT                     : FFT_DATA_REC;

    channel                 : Integer;
    Codec_Block_size        : Integer;
    Codec_Block_size_recip  : Double;

    Codec_Block_duration    : Double;

    channels                : Integer;

    Codec_Block_bits        : Integer;
    Codec_Block_bit_shift   : Integer;

    sample_rate             : Integer;
    bits_per_sample         : Integer;
    bytes_per_sample        : Integer;
    wave_size               : Cardinal;
    total_samples           : Cardinal;
    Lower_Freq_Limit        : Integer;
    Upper_Freq_Limit        : Integer;

    blocks_processed        : Integer;

    bits_removed            : Integer;
    bits_lost               : Integer;
    last_print              : Integer;

    output_blocks           : Integer;

    sample_rate_recip       : Double;

    blocks_processed_recip  : Double;
    processing_rate         : Double;

    samples_processed       : Int64;
    clipped_samples         : Int64;
    samples_processed_recip : Double;

    jump_location           : Integer;
  End;

  parameters : record
    logfilename               : String;

    wavName                   : String;
    stdinname                 : String;
    lossyName                 : String;
    lwcdfName                 : String;

    WavInpDir                 : string;
    WavOutDir                 : string;

    detail                    : longbool;
    verbosity                 : longbool;
    warnings                  : longbool;

    forcing                   : longbool;
    correction                : longbool;
    checking                  : longbool;
    merging                   : longbool;

    silent                    : longbool;
    linkchannels              : longbool;
    writetolog                : longbool;

    STDINPUT                  : longbool;
    STDOUTPUT                 : longbool;
    perchannel                : longbool;
    bitdist                   : longbool;

    spread                    : integer;
    blockdist                 : longbool;
    sampledist                : longbool;
    logfileopened             : longbool;

    midside                   : longbool;
    dccorrect                 : longbool;
    parameters_checked        : longbool;

    adaptive_shaping          : Integer;
    adaptive_shaping_warp     : longbool;

    static                    : Integer;
    dynamic                   : Double;
    quality                   : Double;
    scaling                   : Double;
    shaping                   : Double;
    
    analyses                  : Integer;
    priority                  : Integer;
    help                      : Integer;

    limit                     : Integer;
    maxclips                  : Integer;
    underlap                  : integer;

    freqdist                  : longbool;
    postanalyse               : longbool;
    histogram                 : longbool;
    width                     : integer;
    longdist                  : LongBool;
  end;

  settings : Record
    noise_threshold_shift_minimum       : Double;
    noise_threshold_shift_average       : Double;
    dccorrect_multiplier                : Double;

    dynamic_minimum_bits_to_keep        : Double;
    dynamic_maximum_bits_to_remove      : Double;

    static_minimum_bits_to_keep         : Integer;
    static_maximum_bits_to_remove       : Integer;

    scaling_factor                      : Double;
    scaling_factor_inv                  : Double;

    quality_double                      : Double;
    quality_fraction                    : double;

    quality_integer                     : Integer;

    maximum_clips_per_channel           : Integer;
    FFT_analysis_switches               : Integer;

    raw_result_short                    : Integer;
    raw_result_sgns                     : Integer;
  End;

  process : Record
    FFT_spreading                 : array[1..precalc_analyses,0..MAX_CHANNELS-1] of record
      old_minimum : Double;
      new_minimum : Double;
      alt_average : Double;
      Masked_Minimum : Double;
    End;

    FFT_underlap_length           : array[0..MAX_FFT_BIT_LENGTH] of Double;
    analysis_blocks               : array[0..MAX_FFT_BIT_LENGTH] of Integer;
    end_overlap_length            : array[0..MAX_FFT_BIT_LENGTH] of Integer;
    actual_analysis_blocks_start  : array[0..MAX_FFT_BIT_LENGTH] of Integer;
    Analyses_Completed            : Array[1..Precalc_Analyses] of Integer;

    Old_Min_Used_History          : Array[1..Precalc_Analyses,0..MAX_FFT_LENGTH_HALF-1] of Integer;
    New_Min_Used_History          : Array[1..Precalc_Analyses,0..MAX_FFT_LENGTH_HALF-1] of Integer;

    Alt_Ave_Used                  : Array[1..Precalc_Analyses] of Integer;
    Old_Min_Used                  : Array[1..Precalc_Analyses] of Integer;
    New_Min_Used                  : Array[1..Precalc_Analyses] of Integer;

    Over_Static                   : Array[1..Precalc_Analyses] of Integer;
    Over_Dynamic                  : Array[1..Precalc_Analyses] of Integer;
    Post_Analysis                 : Array[1..Precalc_Analyses] of Procedure;

    old_min_bin                   : Integer;
    new_min_bin                   : Integer;
    Dynamic_Maximum_Bits_to_Remove: Integer;
    Analyses_Performed            : Integer;
  end;

  results : Record
    saved_FFT_results             : array[0..MAX_CHANNELS-1,1..PRECALC_ANALYSES] of FFT_Results_Rec;
    minima                        : array[0..MAX_CHANNELS-1,1..PRECALC_ANALYSES+2] of Integer;
    this_FFT                      : FFT_Results_Rec;
  End;

  history : Record
    DATA_results                  : Array[0..MAX_CHANNELS-1,0..MAX_HIST_FFT_SHORT_LENGTH div 2 + 1] of double;
    BTRD_results                  : Array[0..MAX_CHANNELS-1,0..MAX_HIST_FFT_SHORT_LENGTH div 2 + 1] of double;
    CORR_results                  : Array[0..MAX_CHANNELS-1,0..MAX_HIST_FFT_SHORT_LENGTH div 2 + 1] of double;
    Last_Results                  : Array[0..MAX_CHANNELS-1,0..MAX_HIST_FFT_SHORT_LENGTH div 2 + 1] of Double;
    FFT                           : FFT_Data_Rec;
    Channel_Offset                : Array[0..MAX_CHANNELS-1] of Integer;

    Histogram_DATA                : Array[0..1025] of Integer;
    Histogram_BTRD                : Array[0..1025] of Integer;
    Histogram_CORR                : Array[0..1025] of Integer;

    Histogram_Multiplier          : Double;
    Histogram_Length              : Integer;
    Histogram_Offset              : Integer;
  end;

  LongDist : Record
    DATA_results                  : Array[0..MAX_CHANNELS-1,0..MAX_FFT_LENGTH_HALF + 1] of double;
    BTRD_results                  : Array[0..MAX_CHANNELS-1,0..MAX_FFT_LENGTH_HALF + 1] of double;
    CORR_results                  : Array[0..MAX_CHANNELS-1,0..MAX_FFT_LENGTH_HALF + 1] of double;
    FFT                           : FFT_Data_Rec;
    Channel_Offset                : Array[0..MAX_CHANNELS-1] of Integer;
  End;

  SGNSFFT : Record
    FFT                           : FFT_Data_Rec;
    Channel_Offset                : Array[0..MAX_CHANNELS-1] of Integer;

    FFT_results_short             : array[0..MAX_FFT_LENGTH_HALF + 1] of Double;
    FFT_results_long              : array[0..MAX_FFT_LENGTH_HALF + 1] of Double;
    Last_Results                  : Array[0..MAX_CHANNELS-1,0..MAX_FFT_LENGTH_HALF + 1] of Double;
  end;

  strings : Record
    parameter     : String;
    datestamp     : String;
    size          : String;
    time          : String;
    elapsed       : String;
    estimate      : String;
    version       : String;
    version_short : String;
  End;

  version : Record
  Case Integer of
    0: (ReleaseBuild,MajorMinor : cardinal);
    1: (Build, Release, Minor, Major : word);
  End;

  //============================================================================
  // Numeric lookup tables.
  //============================================================================
  PowersOf : Record
    TwoInt  : array[    0..  63] of Int64;
    TwoM1   : array[    0..  63] of Int64;
    Two     : array[-1024..1023] of Double;
    Ten     : array[ -308.. 307] of double;
  End;
//==============================================================================

  Stats : Record
    bits_removed : Array[0..MAX_CHANNELS-1,0..33] of Integer;
  End;

//==============================================================================
procedure lossyWAVerror   (lwe_string : string; lwe_value : integer);
procedure lossyWAVwarning (lww_string : string);
procedure gettimer;
procedure setpriority     (spv : integer);
procedure ndelay;
Function  CardinalToHex   (cth_c : cardinal) : string;
procedure time_string_make(var tsms : string; tsmt : double);
procedure size_string_make(var ssms : string; ssmt : double);
Procedure Set_FFT_Data_Rec(var FFTDataRec : FFT_Data_Rec; SFDR_Bits : Integer);
{$IFDEF DISPLAY_VAR_DATA}
Procedure write_unitname_and_datapointers(UnitName : String; UserData1Pointer : Pointer; UserData2Pointer : Pointer; UserData2Size : Cardinal);
{$ENDIF}

//==============================================================================
implementation
//==============================================================================
uses
  Sysutils,
  Math,
  Windows;
const
  UNITNAME='nCore';
var
  Unit_Data_Packer2 : array[0..3] of integer;

  //============================================================================
  procedure lossyWAVerror(lwe_string : string; lwe_value : integer);
  //============================================================================
  Begin
  //============================================================================
    if (not Parameters.silent) then
      write(ErrOutput,'%lossyWAV Error% : ',lwe_string,CRLF);
    halt(lwe_value);
  //============================================================================
  End;
  //============================================================================

  //============================================================================
  procedure lossyWAVwarning(lww_string : string);
  //============================================================================
  Begin
  //============================================================================
    if (Parameters.warnings) and (not Parameters.silent) then
      write(ErrOutput,'%lossyWAV Warning% : ',lww_string,CRLF);
  //============================================================================
  End;
  //============================================================================

  //============================================================================
  Procedure GetVersionInfo;
  //============================================================================
  var
    lptstrFileName  : String;
    dwHandle        : DWORD;
    dwLen           : DWORD;
    puLen           : DWORD;
    lpData       : Pointer;
    lplpBuffer   : Pointer;
  begin
  //============================================================================
    Strings.Version:='';

    lptstrFileName:=GetModuleName(HInstance);

    dwLen:=GetFileVersionInfoSize(PChar(lptstrFileName), dwHandle);

    if (dwLen>0) then
    begin
      GetMem(lpData,dwLen);
      try
        GetFileVersionInfo(PChar(lptstrFileName),dwHandle,dwLen,lpData);

        VerQueryValue(lpData,'\',lplpBuffer,puLen);

        Version.MajorMinor:=PVSFixedFileInfo(lplpBuffer)^.dwFileVersionMS;
        Version.ReleaseBuild:=PVSFixedFileInfo(lplpBuffer)^.dwFileVersionLS;
      finally
        FreeMem(lpData);
      end;
      with version do
        Strings.Version:=Format('%d.%d.%d.%d',[Major,Minor,Release,Build]);
    end
    else
    Begin
      Version.MajorMinor:=0;
      Version.ReleaseBuild:=0;
    End;
  //============================================================================
  end;
  //============================================================================

  //============================================================================
  procedure gettimer;
  //============================================================================
  begin
  //============================================================================
    QueryPerformanceCounter(timer.stop);
    timer.Elapsed:=(timer.Stop-timer.Start)*timer.Period;
    if timer.elapsed<>0 then
      Global.processing_rate:=Global.samples_processed*Global.sample_rate_recip/timer.elapsed;
  //============================================================================
  end;
  //============================================================================

  //============================================================================
  procedure setpriority(spv : integer);
  //============================================================================
  begin
  //============================================================================
    setpriorityclass(getcurrentprocess,spv);
  //============================================================================
  end;
  //============================================================================

  //============================================================================
  procedure ndelay;
  {$IFNDEF USEASM}
  //============================================================================
  Begin
  //============================================================================
    Gettimer;
    timer.DelayEnd:=Floor(timer.stop+timer.frequency*timer.delaytime);
    while timer.Stop<timer.DelayEnd do
      gettimer;
  //============================================================================
  End;
  //============================================================================
  {$ELSE}
  //============================================================================
  asm
                call    gettimer
                fild    timer.stop
                fild    timer.frequency
                fmul    timer.delaytime
                fadd

    @@01:       call    gettimer
                fcom    timer.stop
                fnstsw  ax
                and     ah,$41
                jnz     @@01

                fstp    st
  end;
  //============================================================================
  {$ENDIF}

  //============================================================================
  Function CardinalToHex(cth_c : cardinal) : string;
  //============================================================================
  begin
  //============================================================================
    CardinalToHex:=HEXCHARLOOKUP[HIBYTE(HIWORD(cth_c)) shr 4]+HEXCHARLOOKUP[HIBYTE(HIWORD(cth_c)) and 15]+
                   HEXCHARLOOKUP[LOBYTE(HIWORD(cth_c)) shr 4]+HEXCHARLOOKUP[LOBYTE(HIWORD(cth_c)) and 15]+
                   HEXCHARLOOKUP[HIBYTE(LOWORD(cth_c)) shr 4]+HEXCHARLOOKUP[HIBYTE(LOWORD(cth_c)) and 15]+
                   HEXCHARLOOKUP[LOBYTE(LOWORD(cth_c)) shr 4]+HEXCHARLOOKUP[LOBYTE(LOWORD(cth_c)) and 15];
  //============================================================================
  end;
  //============================================================================

  {$IFDEF DISPLAY_VAR_DATA}
  //============================================================================
  Procedure write_unitname_and_datapointers(UnitName : String; UserData1Pointer : Pointer; UserData2Pointer : Pointer; UserData2Size : Cardinal);
  //============================================================================
  var
    datasize : Cardinal;
  begin
  //============================================================================
    datasize:=cardinal(userdata2pointer)+userdata2size-cardinal(userdata1pointer);
    write(erroutput,UnitName:15,':',cardinaltohex(cardinal(userdata1pointer)),',',cardinaltohex(cardinal(userdata2pointer)+userdata2size),',',
                                    datasize:10,' (',CardinalToHex(datasize),') bytes.',CRLF);
  //============================================================================
  end;
  //============================================================================
  {$ENDIF}

  //============================================================================
  procedure time_string_make(var tsms : string; tsmt : double);
  //============================================================================
  var
    tsmi, tsmk : Integer;
  begin
  //============================================================================
    if tsmt<3600 then
    begin
      tsmi:=4;
      tsmk:=100;
    end
    else
    Begin
      tsmi:=1;
      tsmk:=1;
    end;
    tsmt:=floor(tsmt*tsmk)*oneover[tsmk]*Second_div_Day;
    datetimetostring(tsms,'hh:nn:ss.zzz',tsmt);
    tsms:=copy(tsms,tsmi,8);
  //============================================================================
  end;
  //============================================================================

  //============================================================================
  procedure size_string_make(var ssms : string; ssmt : double);
  //============================================================================
  const
    size_chars : String[7] = 'kMGTPEZ';
  var
    ssmi, ssmj  : integer;
    ssmx        : String;
    ssmy        : double;
  begin
  //============================================================================
    if ssmt>99999 then
    begin
      ssmy:=log2(ssmt)*oneover[10];
      ssmj:=floor(ssmy);
      ssmi:=floor(log2_10*(ssmy-ssmj));

      if ssmi>2 then
      begin
        inc(ssmj);
        ssmi:=0;
      end;

      ssmi:=max(0,3-ssmi);
      ssmt:=floor(ssmt*PowersOf.Two[-ssmj*10]*PowersOf.ten[ssmi])*PowersOf.ten[-ssmi];
      ssmx:=size_chars[ssmj]+'iB';
    end
    else
    begin
      ssmx:='B';
      ssmi:=0;
    end;
    ssms:=floattostrf(ssmt,ffFixed,99,ssmi)+ssmx;
  //============================================================================
  end;
  //============================================================================

  //============================================================================
  Procedure Set_FFT_Data_Rec(var FFTDataRec : FFT_Data_Rec; SFDR_Bits : Integer);
  //============================================================================
  Begin
  //============================================================================
    With FFTDataRec do
    Begin
      bit_length        :=SFDR_Bits;
      bit_shift_from_max:=MAX_FFT_BIT_LENGTH-bit_length;
      threshold_shift   :=(Global.Codec_Block_Bits-bit_length)*SINGLE_HALF;
      length            :=powersof.twoint[bit_length];
      length_recip      :=powersof.two[-bit_length];
      length_m1         :=length-1;
      length_half       :=length shr 1;
      length_half_m1    :=length_half-1;
      length_recip      :=PowersOf.Two[-bit_length];
      length_half_recip :=length_recip*2;
    End;
  //============================================================================
  End;
  //============================================================================

  //============================================================================
  Procedure NSystem_Init();
  //============================================================================
  var
    nt_i, nt_j : Integer;
  Begin
  //============================================================================
    //==========================================================================
    // Initialise Block Data pointers and sizes.
    //==========================================================================
    for nt_i:=0 to 3 do
      BlockData.Pointer[nt_i]:=@BlockData.Data[nt_i];

    BlockData.Size.Prev:=0;
    BlockData.Size.Last:=0;
    BlockData.Size.This:=0;
    BlockData.Size.Next:=0;
    //==========================================================================

    //==========================================================================
    // Ensure initial scaling factor is unity.
    //==========================================================================
    settings.scaling_factor:=1;
    settings.scaling_factor_inv:=1;
    //==========================================================================

    //==========================================================================
    // Get executable version information.
    //==========================================================================
    GetVersionInfo;

    With Version do
    if build<27 then
    begin
      Strings.Version_Short:=Format('%d.%d.%d',[major,minor,release]);
      if build>0 then
        Strings.Version_Short:=Strings.Version_Short+chr(build+96);
    end
    else
      Strings.Version_Short:=Strings.Version;
    //==========================================================================

    //==========================================================================
    // Initialise handles & files;
    //==========================================================================
    hStdIn :=GetStdHandle(STD_INPUT_HANDLE);
    ftSTDIN :=GetFileType(hStdIn);

    hStdOut:=GetStdHandle(STD_OUTPUT_HANDLE);
    ftSTDOUT:=GetFileType(hStdOut);

    hStdErr:=GetStdHandle(STD_ERROR_HANDLE);
    ftSTDERR:=GetFileType(hStdErr);
    //==========================================================================

    //==========================================================================
    // Initialise timer;
    //==========================================================================
    timer.StartTime:=now;
    QueryPerformanceFrequency(timer.frequency);
    QueryPerformanceCounter(timer.Start);
    timer.Period:=1/timer.frequency;
    //==========================================================================

    //==========================================================================
    // Clear frequency distribution history;
    //==========================================================================
    for nt_i:=0 to MAX_FFT_LENGTH-1 do
      FFT_raw_result[nt_i]:=0;

    for nt_j:=0 to MAX_CHANNELS-1 do
    begin
      LongDist.Channel_Offset[nt_j]:=nt_j*Sizeof(LongDist.DATA_Results[0]);
      History.Channel_Offset[nt_j]:=nt_j*Sizeof(History.Last_Results[0]);
      SGNSFFT.Channel_Offset[nt_j]:=nt_j*Sizeof(SGNSFFT.Last_Results[0]);

      for nt_i:=0 to MAX_HIST_FFT_SHORT_LENGTH div 2 + 1 do
      begin
        history.DATA_results[nt_j,nt_i]:=0;
        history.BTRD_results[nt_j,nt_i]:=0;
        history.CORR_results[nt_j,nt_i]:=0;
        History.Last_Results[nt_j,nt_i]:=0;
      end;

      for nt_i:=0 to MAX_FFT_LENGTH_HALF + 1 do
      begin
        LongDist.DATA_results[nt_j,nt_i]:=0;
        LongDist.BTRD_results[nt_j,nt_i]:=0;
        LongDist.CORR_results[nt_j,nt_i]:=0;
      end;

      for nt_i:=0 to 33 do
        stats.bits_removed[nt_j,nt_i]:=0;

      for nt_i:=1 to PRECALC_ANALYSES+2 do
        Results.Minima[nt_j,nt_i]:=0;
    end;

    //==========================================================================
    // Clear Histogram arrays.
    //==========================================================================
    for nt_i:=0 to 1025 do
    Begin
      History.Histogram_DATA[nt_i]:=0;
      History.Histogram_BTRD[nt_i]:=0;
      History.Histogram_CORR[nt_i]:=0;
    End;
    //==========================================================================

    //==========================================================================
    // Create powers of two (integer) table.
    //==========================================================================
    PowersOf.TwoInt[0]:=1;
    PowersOf.Twom1[0]:=0;
    for nt_i:=1 to 62 do
    begin
      PowersOf.TwoInt[nt_i]:=PowersOf.TwoInt[nt_i-1]+PowersOf.TwoInt[nt_i-1];
      PowersOf.Twom1[nt_i]:=PowersOf.TwoInt[nt_i]-1;
    end;
    //==========================================================================

    //==========================================================================
    // Create powers of two (float) table.
    //==========================================================================
    for nt_i:=-1024 to 1023 do
      PowersOf.Two[nt_i]:=Power(2,nt_i);
    //==========================================================================

    //==========================================================================
    // Create powers of ten table.
    //==========================================================================
    for nt_i:=-308 to 307 do
      PowersOf.Ten[nt_i]:=Power(10,nt_i);
    //==========================================================================

    //==========================================================================
    // Create integer inverse table.
    //==========================================================================
    OneOver[0]:=1;
    for nt_i:=1 to MAX_FFT_LENGTH+1 do
      OneOver[nt_i]:=1/nt_i;
    //==========================================================================

  //============================================================================
  End;
  //============================================================================

//==============================================================================
Initialization
//==============================================================================
  Unit_Data_Packer2[0]:=0;
  {$IFDEF DISPLAY_VAR_DATA}
  write_unitname_and_datapointers(UNITNAME,@Unit_Data_Packer1,@Unit_Data_Packer2,sizeof(Unit_Data_Packer2));
  {$ENDIF}

  NSystem_Init();
//==============================================================================
finalization
//==============================================================================
  {$IFDEF WRITE_FINAL}
  Write(Erroutput,UNITNAME,CRLF);
  {$ENDIF}
//==============================================================================
end.
//==============================================================================
