//==============================================================================
//
//    lossyWAV: Added noise WAV bit reduction method by David Robinson;
//              Noise shaping coefficients by Sebastian Gesemann;
//
//    Copyright (C) 2007-2011 Nick Currie, Copyleft.
//
//    This program is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program.  If not, see <http://www.gnu.org/licenses/>.
//
//    Contact: lossywav <at> hotmail <dot> co <dot> uk
//
//==============================================================================
//
//     fourier.pas  -  Don Cross <dcross <at> intersrv <dot> com>
//
//     This is a Turbo Pascal Unit for calculating the Fast Fourier Transform
//     (FFT) and the Inverse Fast Fourier Transform (IFFT).
//     Visit the following URL for the latest version of this code.
//     This page also has a C/C++ version, and a brief discussion of the
//     theory behind the FFT algorithm.
//
//        http://www.intersrv.com/~dcross/fft.html#pascal [!!Dead link!!]
//
//     Revision history [most recent first]:
//
// 2008 May 12 [Nick Currie]
//     All code transcoded to IA-32/x87. FFT_Real implemented to allow calculation
//     of a packed Real array of length 2N to be calculated in a Complex array of
//     length N, with subsequent untangling. ReversedBits, Ai & Ar lookup tables
//     implemented. Calc_Frequency, FFT_Integer and FFT_Integer_Cleanup removed
//     as not used in lossyWAV.
//
// 1996 December 11 [Don Cross]
//     Improved documentation of the procedure CalcFrequency.
//     Fixed some messed up comments in procedure IFFT.
//
// 1996 December 6 [Don Cross]
//     Made procedure 'fft_integer' more efficient when buffer size changes
//     in successive calls:  the buffer is now only resized when the input
//     has more samples, not a differing number of samples.
//     Also changed the way 'fft_integer_cleanup' works so that it is
//     more "bullet-proof".
//
// 1996 December 4 [Don Cross]
//     Adding the procedure 'CalcFrequency', which calculates the FFT
//     at a specific frequency index p=0..n-1, instead of the whole
//     FFT.  This is O(n^2) instead of O(n*log(n)).
//
// 1996 November 30 [Don Cross]
//     Adding a routine to allow FFT of an input array of integers.
//     It is called 'fft_integer'.
//
// 1996 November 18 [Don Cross]
//     Added some comments.
//
// 1996 November 17 [Don Cross]
//     Wrote and debugged first version.
//
//==============================================================================

//==============================================================================
unit nFFT_asm;
//==============================================================================
// Conditional Defines
//==============================================================================
{$DEFINE RB_LUT}
//==============================================================================
interface
//==============================================================================
uses
  nCore;
//==============================================================================
var
  Unit_Data_Packer1 : Packed Record
    Data : Array[0..3] of Integer;
  End;
//==============================================================================

//==============================================================================
// Set the number of samples for FFT analysis;
//==============================================================================
function SetFFTLength(SFL_Samples : Integer) : Integer;

//==============================================================================
//  Calculates the (in place) Fast Fourier Transform of the array of complex
//  numbers represented by FFT_Input to produce the output complex
//  numbers in FFT_Array.
//==============================================================================
procedure FFT_DIT_Complex;

//==============================================================================
//  Calculates the (in place) Fast Fourier Transform of the packed array of real
//  numbers represented by FFT_Input to produce the output complex
//  numbers in FFT_Array, then post-processes the results to unpack the
//  Real output in the full FFT_Array.
//==============================================================================
procedure FFT_DIT_Real;

//==============================================================================
//  Calculates the (in place) Inverse Fast Fourier Transform of the array of
//  complex numbers represented by FFT_Input to produce the output
//  complex numbers in FFT_Array.
//==============================================================================
procedure IFFT_DIT_Complex;
//==============================================================================

//==============================================================================
// Initialise lookup talbes for internal FFT routines.
//==============================================================================
procedure nfft_double_init;
//==============================================================================

//==============================================================================
implementation
//==============================================================================
Uses
  fftw_interface;
//==============================================================================
const
  UNITNAME='nFFT_asm';
var
  {$IFDEF RB_LUT}
  R_Arr                 : array[0..MAX_FFT_LENGTH-1] of Integer;
  {$ELSE}
  REVERSED_WORDS        : array[0..65535] of word;
  {$ENDIF}
  A_Arr                 : Array[0..MAX_FFT_LENGTH-1] of tDComplex;

  REVERSED_BYTES        : array[0..255] of byte;

  NumSamples            : Integer;
  NumberOfBitsNeeded    : Integer;
  MainBlockSize         : Integer;
  BlockBitLen           : Integer;
  PermShiftBits         : Integer;
  PermShiftAdd          : Integer;
  NumSamples_Reciprocal : Double;

  Unit_Data_Packer2     : array[0..6] of integer;

  //============================================================================
  Function SetFFTLength(SFL_Samples : Integer) : Integer; assembler;
  //============================================================================
  asm
                //store NumComplexSamples
                mov     eax,SFL_Samples
                cmp     eax,MAX_FFT_LENGTH
                ja      @@01
                cmp     eax,0
                jb      @@01

                //find highest set bit
                bsr     edx,eax
                jz      @@01
                //find lowest set bit
                bsf     ecx,eax
                jz      @@01
                //if highest=lowest then poweroftwo
                cmp     edx,ecx
                je      @@02
                //else
      @@01:     xor     eax,eax
                xor     edx,edx

      @@02:     mov     ecx,MAX_FFT_BIT_LENGTH
                sub     ecx,edx
                mov     PermShiftBits,ecx
                Mov     ecx,dword ptr PowersOf.TwoInt[ecx*INT64_SIZE]
                Mov     PermShiftAdd,ecx

                mov     NumberOfBitsNeeded,edx
                mov     NumSamples,eax
                fld     qword ptr PowersOf.Two[EXPONENT_BIAS*DOUBLE_SIZE-edx*DOUBLE_SIZE]
                fstp    NumSamples_Reciprocal
  end;
  //============================================================================

  //============================================================================
  procedure Shuffle_In_Place_Complex; assembler;
  //============================================================================
  asm
                mov     esi,NumSamples
                dec     esi
                mov     edx,permshiftadd
                xor     edi,edi
                xor     ecx,ecx
                jmp     @@S2
                nop

      @@S1:       {$IFDEF RB_LUT}
                  mov     eax,dword ptr R_Arr[edi*INTEGER_SIZE]
                  {$ELSE}
                  sub     ebx,ebx
                  mov     eax,edi

                  mov     bx,ax
                  mov     ax,word ptr REVERSED_WORDS[ebx*word_size]
                  ror     eax,16
                  mov     bx,ax
                  mov     ax,word ptr REVERSED_WORDS[ebx*word_size]
                  shr     eax,32-MAX_FFT_BIT_LENGTH-DCOMPLEX_SHIFT
                  {$ENDIF}

                  cmp     eax,ecx
                  jle     @@S2

                  fld     qword ptr FFT_Array[ecx]
                  fld     qword ptr FFT_Array[ecx+DOUBLE_SIZE]

                  fld     qword ptr FFT_Array[eax]
                  fld     qword ptr FFT_Array[eax+DOUBLE_SIZE]

                  fstp    qword ptr FFT_Array[ecx+DOUBLE_SIZE]
                  fstp    qword ptr FFT_Array[ecx]

                  fstp    qword ptr FFT_Array[eax+DOUBLE_SIZE]
                  fstp    qword ptr FFT_Array[eax]

                  {$IFDEF ALIGN08}
                  nop;nop;nop;nop;nop;
                  {$ENDIF}

      @@S2:       add     edi,edx
                  add     ecx,COMPLEX_SIZE
                  sub     esi,1
                  jg      @@S1
  end;
  //============================================================================

  //============================================================================
  procedure FFT_DIT_Complex; assembler;
  //============================================================================
    //==========================================================================
    Procedure FFT_Radix_Four_DIT; Assembler;
    //==========================================================================
    var
      MainBlockSize       : Integer;
      Stride              : Integer;
      ShiftBits           : Integer;
      I, J, N             : Integer;
      Y0, Y1, Y2, Y3      : tDComplex;
    Asm
                mov     blockbitlen,eax

                mov     ecx,dword ptr PowersOf.twoint[eax*INT64_SIZE]
                mov     MainBlockSize,ecx
                mov     ecx,dword ptr PowersOf.twoint[eax*INT64_SIZE-2*INT64_SIZE]
                Mov     Stride,ecx

                Mov     eax,MAX_FFT_BIT_LENGTH
                sub     eax,blockbitlen
                mov     shiftbits,eax

                mov     I,0
              //================================================================
      @@FC0:      mov     eax,I
                  shl     eax,COMPLEX_SHIFT
                  mov     J,eax
                  mov     N,0
                //==============================================================
      @@FC1:        Mov     esi,N
                    mov     ecx,shiftbits
                    add     ecx,COMPLEX_SHIFT
                    shl     esi,cl

                    mov     edi,esi
                    add     edi,esi

                    mov     edx,stride
                    shl     edx,COMPLEX_SHIFT

                    mov     eax,J
                    mov     ebx,eax
                    add     ebx,edx //K
                    mov     ecx,ebx
                    add     ecx,edx //L
                    add     edx,ecx //M

                  //============================================================
                    fld     qword ptr FFT_Array[edx]
                    fld     qword ptr FFT_Array[edx+DOUBLE_SIZE]
                    fld     qword ptr A_Arr[edi+esi+DOUBLE_SIZE]
                    fld     qword ptr A_Arr[edi+esi]

                    fld     st
                    fmul    st,st(3)
                    fld     st(2)
                    fmul    st,st(5)
                    fadd
                    fxch    st(4)

                    fmul
                    fxch    st(2)
                    fmul
                    fsub

                    fld     qword ptr FFT_Array[ecx]
                    fld     qword ptr FFT_Array[ecx+DOUBLE_SIZE]
                    fld     qword ptr A_Arr[esi+DOUBLE_SIZE]
                    fld     qword ptr A_Arr[esi]

                    fld     st
                    fmul    st,st(3)
                    fld     st(2)
                    fmul    st,st(5)
                    fadd
                    fxch    st(4)

                    fmul
                    fxch    st(2)
                    fmul
                    fsub

                    fxch    st(3)

                    fld     st(3)
                    fadd    st,st(3)
                    fstp    y1.re
                    fld     st(1)
                    fadd    st,st(1)
                    fstp    y1.im

                    fsubr
                    fstp    y3.re
                    fsub
                    fstp    y3.im
                  //============================================================
                    fld     qword ptr FFT_Array[ebx]
                    fld     qword ptr FFT_Array[ebx+DOUBLE_SIZE]
                    fld     qword ptr A_Arr[edi+DOUBLE_SIZE]
                    fld     qword ptr A_Arr[edi]

                    fld     st
                    fmul    st,st(3)
                    fld     st(2)
                    fmul    st,st(5)
                    fadd
                    fxch    st(4)

                    fmul
                    fxch    st(2)
                    fmul
                    fsub

                    fld     qword ptr FFT_Array[eax]
                    fld     st
                    fadd    st,st(2)
                    fstp    Y0.Re
                    fsubr
                    fstp    y2.re

                    fld     qword ptr FFT_Array[eax+DOUBLE_SIZE]
                    fld     st
                    fadd    st,st(2)
                    fstp    Y0.Im
                    fsubr
                    fstp    y2.im
                  //============================================================
                    fld     y0.re
                    fld     y1.re
                    fld     y0.im
                    fld     y1.im

                    fld     st
                    fadd    st,st(2)
                    fstp    qword ptr FFT_Array[eax+DOUBLE_SIZE]
                    fsub
                    fstp    qword ptr FFT_Array[ecx+DOUBLE_SIZE]

                    fld     st
                    fadd    st,st(2)
                    fstp    qword ptr FFT_Array[eax]
                    fsub
                    fstp    qword ptr FFT_Array[ecx]
                  //============================================================
                    fld     y3.im
                    fld     y2.im
                    fld     y3.re
                    fld     y2.re

                    fld     st
                    fsub    st,st(2)
                    fstp    qword ptr FFT_Array[ebx]
                    fadd
                    fstp    qword ptr FFT_Array[edx]

                    fld     st
                    fsub    st,st(2)
                    fstp    qword ptr FFT_Array[ebx+DOUBLE_SIZE]
                    fadd
                    fstp    qword ptr FFT_Array[edx+DOUBLE_SIZE]
                  //============================================================
                    add    J,COMPLEX_SIZE
                    mov    eax,N
                    add    eax,1
                    mov    N,eax
                    cmp    eax,stride
                    jl     @@FC1
                //==============================================================
                  mov    eax,I
                  add    eax,mainblocksize
                  mov    I,eax
                  cmp    eax,numsamples
                  jl     @@FC0
    //==========================================================================
    End;
    //==========================================================================

    //==========================================================================
    Procedure FFT_Radix_Two_DIT; Assembler;
    //==========================================================================
    Asm
                mov     blockbitlen,eax

                mov     ecx,dword ptr PowersOf.twoint[eax*INT64_SIZE]
                mov     MainBlockSize,ecx
                shl     ecx,COMPLEX_SHIFT-1

                {$IFDEF ALIGN08}
                nop;nop;nop;
                {$ENDIF}

      @@FC3:      xor     ebp,ebp                                       //i
                  mov     edi,MAX_FFT_BIT_LENGTH+COMPLEX_SHIFT
                  sub     edi,blockbitlen
                  mov     edi,dword ptr PowersOf.twoint[edi*INT64_SIZE]

                  {$IFDEF ALIGN08}
                  nop;nop;nop;nop;
                  {$IFDEF ALIGN16}
                  nop;nop;nop;nop;nop;nop;nop;nop;
                  {$ENDIF}
                  {$ENDIF}

      @@FC4:        xor     esi,esi                                     //n
                    xor     eax,eax

                    {$IFDEF ALIGN08}
                    nop;nop;nop;nop;
                    {$IFDEF ALIGN16}
                    nop;nop;nop;nop;nop;nop;nop;nop;
                    {$ENDIF}
                    {$ENDIF}

      @@FC5:          mov     ebx,ebp
                      shl     ebx,COMPLEX_SHIFT
                      add     ebx,esi                                   //j
                      mov     edx,ebx
                      add     edx,ecx                                   //k:=j+blockend

                      // load FFT_Array[k] into FPU
                      fld     qword ptr FFT_Array[edx]                  // D1.Re
                      fld     qword ptr FFT_Array[edx+DOUBLE_SIZE]      // D1.Im
                      fld     qword ptr A_Arr[eax+DOUBLE_SIZE]          // W1.Im
                      fld     qword ptr A_Arr[eax]                      // W1.Re

                      fld     st                                        // W1.Re
                      fmul    st,st(3)                                  // W1.Re * D1.Im
                      fld     st(2)                                     // W1.Im
                      fmul    st,st(5)                                  // W1.Im * D1.Re
                      fadd                                              // W1.Re * D1.Im + W1.Im * D1.Re
                      fxch    st(4)

                      fmul                                              // D1.Re * W1.Re
                      fxch    st(2)                                     // D1.Im
                      fmul                                              // W1.Im * D1.Im
                      fsub                                              // D1.Re * W1.Re - W1.Im * D1.Im

                      fld     qword ptr FFT_Array[ebx]                  // FFT_Array[j].re
                      fld     st
                      fsub    st,st(2)
                      fstp    qword ptr FFT_Array[edx]                  // FFT_Array[k].re:=FFT_Array[j].re-t.re
                      fadd
                      fstp    qword ptr FFT_Array[ebx]                  // FFT_Array[j].re:=FFT_Array[j].re+t.re

                      fld     qword ptr FFT_Array[ebx+DOUBLE_SIZE]      // FFT_Array[j].im
                      fld     st
                      fsub    st,st(2)
                      fstp    qword ptr FFT_Array[edx+DOUBLE_SIZE]      // FFT_Array[k].im:=FFT_Array[j].im-t.im
                      fadd
                      fstp    qword ptr FFT_Array[ebx+DOUBLE_SIZE]      // FFT_Array[j].im:=FFT_Array[j].im+t.im

                      add     eax,edi
                      add     esi,COMPLEX_SIZE                          // n:=n+1
                      cmp     esi,ecx
                      jl      @@FC5

                    add     ebp,MainBlockSize                           // fft_i:=fft_i+MainBlockSize
                    cmp     ebp,NumSamples
                    jl      @@FC4
    //==========================================================================
    End;
    //==========================================================================

  //============================================================================
  asm
  //============================================================================
                push    ebx
                push    esi
                push    edi
                push    ebp

                call    Shuffle_In_Place_Complex

                mov     blockbitlen,0
                {$IFDEF ALIGN08}
                nop;
                {$IFDEF ALIGN16}
                nop;nop;nop;nop;nop;nop;nop;nop;
                {$ENDIF}
                {$ENDIF}

      @@R4:     mov     eax,blockbitlen
                add     eax,2
                cmp     eax,numberofbitsneeded
                ja      @@R2

                Call    FFT_Radix_Four_DIT
                Jmp     @@R4

                {$IFDEF ALIGN08}
                nop;
                {$IFDEF ALIGN16}
                nop;nop;nop;nop;nop;nop;nop;nop;
                {$ENDIF}
                {$ENDIF}

      @@R2:     mov     eax,blockbitlen
                add     eax,1
                cmp     eax,numberofbitsneeded
                ja      @@End

                Call    FFT_Radix_Two_DIT
                Jmp     @@R2

      @@End:    pop     ebp
                pop     edi
                pop     esi
                pop     ebx
  //============================================================================
  end;
  //============================================================================

  //============================================================================
  procedure FFT_DIT_Real; Assembler;
  //============================================================================
  Asm
                push    ebx
                push    ebp
                push    edi
                push    esi

                call    FFT_DIT_Complex

                mov     ecx,PermShiftBits
                dec     ecx

                Mov     ebx,NumSamples
                shl     ebx,COMPLEX_SHIFT

                Mov     edi,ebx
                shr     edi,1

                fld     single_half                             //0.5        #0#

                {$IFDEF ALIGN08}
                nop;
                {$ENDIF}

      @@FR1:      mov     esi,edi
                  shl     esi,cl

                  mov     eax,ebx
                  sub     eax,edi
                                                                          //  -
                  fld     qword ptr FFT_Array[edi]              //temp1.re   #1#
                  fld     qword ptr FFT_Array[eax]              //temp2.re   #2#
                  fld     qword ptr FFT_Array[edi+DOUBLE_SIZE]  //temp1.im   #3#
                  fld     qword ptr FFT_Array[eax+DOUBLE_SIZE]  //temp2.im   #4#

                  fld     st(3)                                 //temp1.re    5
                  fsub    st,st(3)                              //temp2.re    5
                  fmul    qword ptr A_Arr[esi]                  //ar          5
                  fld     st(2)                                 //temp1.im    6
                  fadd    st,st(2)                              //temp2.im    6
                  fmul    qword ptr A_Arr[esi+DOUBLE_SIZE]      //ai          6
                  fsub                                          //T.im       #5#

                  fld     st(2)                                 //temp1.im    6
                  fsub    st,st(2)                              //temp2.im    6
                  fsub    st,st(1)                              //T.Im        6
                  fmul    st,st(6)                              //0.5         6
                  fst     qword ptr FFT_Array[edi+DOUBLE_SIZE]            //  6
                  fchs
                  fst     qword ptr FFT_Array[ebx+eax+DOUBLE_SIZE]
                  fchs
                  fadd                                                    //  5
                  fst     qword ptr fft_array[ebx+edi+DOUBLE_SIZE]
                  fchs
                  fstp    qword ptr FFT_Array[eax+DOUBLE_SIZE]            // #4#

                  fadd                                                    //  3
                  fmul    qword ptr A_Arr[esi]                  //ar          3
                  fld     st(2)                                 //temp1.re    4
                  fsub    st,st(2)                              //temp2.re    4
                  fmul    qword ptr A_Arr[esi+DOUBLE_SIZE]      //ai          4
                  fadd                                          //T.re       #3#

                  fxch    st(2)                                           //  3
                  fadd                                                    //  2
                  fadd    st,st(1)                                        //  2
                  fmul    st,st(2)                              //0.5         2
                  fst     qword ptr FFT_Array[edi]                        //  2
                  fst     qword ptr FFT_Array[ebx+eax]                        //  2
                  fsubr                                                   //  1
                  fst     qword ptr FFT_Array[ebx+edi]
                  fstp    qword ptr FFT_Array[eax]                        // #0#

                  sub     edi,COMPLEX_SIZE
                  jg      @@FR1

                fstp    st                                                // #-#

                fld     qword ptr FFT_Array[0]
                fld     qword ptr FFT_Array[DOUBLE_SIZE]

                fld     st(1)
                fadd    st,st(1)
                fstp    qword ptr FFT_Array[0]
                fsub
                fstp    qword ptr FFT_Array[ebx]

                fldz
                fst     qword ptr FFT_Array[DOUBLE_SIZE]
                fstp    qword ptr FFT_Array[ebx+DOUBLE_SIZE]

                pop     esi
                pop     edi
                pop     ebp
                pop     ebx
  end;
  //============================================================================

  //============================================================================
  procedure IFFT_DIT_Complex; assembler;
  //============================================================================
  asm
                push    ebx
                push    esi
                push    edi
                push    ebp

                call    Shuffle_In_Place_Complex

                mov     ecx,1                                           // blockend
                mov     blockbitlen,1                                   // blockbitlen
                mov     MainBlockSize,2                                 // MainBlockSize

                {$IFDEF ALIGN08}
                nop;nop;
                {$IFDEF ALIGN16}
                nop;nop;nop;nop;nop;nop;nop;nop;
                {$ENDIF}
                {$ENDIF}

      @@FC3:      xor     ebp,ebp                                       //i
                  mov     edi,MAX_FFT_BIT_LENGTH
                  sub     edi,blockbitlen
                  add     edi,COMPLEX_SHIFT
                  mov     edi,dword ptr PowersOf.twoint[edi*INT64_SIZE]

                  {$IFDEF ALIGN08}
                  nop;
                  {$IFDEF ALIGN16}
                  nop;nop;nop;nop;nop;nop;nop;nop;
                  {$ENDIF}
                  {$ENDIF}

      @@FC4:        xor     esi,esi                                      //n
                    xor     eax,eax
                    {$IFDEF ALIGN08}
                    nop;nop;nop;nop;
                    {$IFDEF ALIGN16}
                    nop;nop;nop;nop;nop;nop;nop;nop;
                    {$ENDIF}
                    {$ENDIF}

      @@FC5:          fld     qword ptr A_Arr[eax+DOUBLE_SIZE]          //a.im:=a_ptr[blockbitlen]^[n].Im
                      fchs
                      fld     qword ptr A_Arr[eax]                      //a.re:=a_ptr[blockbitlen]^[n].Re

                      mov     ebx,esi                                   //j
                      add     ebx,ebp
                      mov     edx,ebx
                      add     edx,ecx                                   //k:=j+blockend

                      shl     ebx,COMPLEX_SHIFT
                      shl     edx,COMPLEX_SHIFT

                      // load FFT_Array[k] into FPU
                      fld     qword ptr FFT_Array[edx+DOUBLE_SIZE]      // FFT_Array[k].im
                      fld     qword ptr FFT_Array[edx]                  // FFT_Array[k].re

                      // T.re := A.re * FFT_Array.Complex[K].Re - A.im * FFT_Array.Complex[K].Im;
                      fld     st(0)
                      fmul    st,st(3)
                      fld     st(2)
                      fmul    st,st(5)
                      fsub                                              // t.re
                      fxch    st(4)

                      // T.im := A.re * FFT_Array.Complex[K].Im + A.im * FFT_Array.Complex[K].Re;
                      fmul
                      fxch    st(2)
                      fmul
                      fadd                                              // t.im

                      fld     qword ptr FFT_Array[ebx+DOUBLE_SIZE]      // FFT_Array[j].im
                      fld     st
                      fsub    st,st(2)
                      fstp    qword ptr FFT_Array[edx+DOUBLE_SIZE]      // FFT_Array[k].im:=FFT_Array[j].im-t.im
                      fadd
                      fstp    qword ptr FFT_Array[ebx+DOUBLE_SIZE]      // FFT_Array[j].im:=FFT_Array[j].im+t.im

                      fld     qword ptr FFT_Array[ebx]                  // FFT_Array[j].re
                      fld     st
                      fsub    st,st(2)
                      fstp    qword ptr FFT_Array[edx]                  // FFT_Array[k].re:=FFT_Array[j].re-t.re
                      fadd
                      fstp    qword ptr FFT_Array[ebx]                  // FFT_Array[j].re:=FFT_Array[j].re+t.re

                      add     esi,1                                     // n:=n+1
                      add     eax,edi
                      cmp     esi,ecx
                      jl      @@FC5

                    add     ebp,MainBlockSize                           // fft_i:=fft_i+MainBlockSize
                    cmp     ebp,NumSamples
                    jl      @@FC4

                  shl     ecx,1                                         // blockend:=2*blockend;
                  shl     MainBlockSize,1                               // MainBlockSize:=2*MainBlockSize;

                  add     blockbitlen,1
                  mov     eax,blockbitlen
                  cmp     eax,numberofbitsneeded
                  jle     @@FC3

                mov     ecx,NumSamples
                shl     ecx,COMPLEX_SHIFT

                pop     ebp
                pop     edi
                pop     esi
                pop     ebx
  end;
  //============================================================================

  //============================================================================
  procedure nfft_double_init; assembler;
  //============================================================================
  asm
      //========================================================================
      // Create 256 byte reversed-bits lookup table
      //========================================================================
                push   ebx

                sub    ebx,ebx
                sub    ecx,ecx
                sub    edx,edx

      @@q1:     mov    al,byte ptr REVERSED_NIBBLES[ecx]
                shl    al,4
                or     al,byte ptr REVERSED_NIBBLES[edx]

                mov    bl,dl
                shl    bl,4
                or     bl,cl
                mov    byte ptr REVERSED_BYTES[ebx],al

                add    edx,1
                cmp    edx,$10
                jl     @@q1

                sub    edx,edx
                add    ecx,1
                cmp    ecx,$10
                jl     @@q1

                pop    ebx
      //========================================================================

      {$IFDEF RB_LUT}
      //========================================================================
      // Fill ReversedBits lookup table (shifted to take into account
      // COMPLEX_SIZE) for maximum FFT length (shift index to suit).
      //========================================================================
      @@01:     mov     edx,MAX_FFT_LENGTH-1

      @@02:       mov     eax,edx
                  sub     ecx,ecx

                  mov     cl,al
                  mov     al,byte ptr REVERSED_BYTES[ecx]
                  mov     cl,ah
                  mov     ah,byte ptr REVERSED_BYTES[ecx]

                  bswap   eax

                  mov     cl,al
                  mov     al,byte ptr REVERSED_BYTES[ecx]
                  mov     cl,ah
                  mov     ah,byte ptr REVERSED_BYTES[ecx]

                  shr     eax,32-MAX_FFT_BIT_LENGTH-COMPLEX_SHIFT
                  mov     dword ptr R_Arr[edx*INTEGER_SIZE],eax

                  sub     edx,1
                  jge     @@02
      //========================================================================
      {$ELSE}
      @@01:     mov     edx,$0000FFFF

      @@02:       sub     ecx,ecx

                  mov     cl,dh
                  mov     al,byte ptr REVERSED_BYTES[ecx]
                  mov     cl,dl
                  mov     ah,byte ptr REVERSED_BYTES[ecx]
                  mov     word ptr REVERSED_WORDS[edx*word_size],ax

                  sub     edx,1
                  jge     @@02
      {$ENDIF}

      //========================================================================
      // Calculate A lookup table for maximum FFT length (shift index to suit).
      //========================================================================
      @@04:     fldpi
                fmul      qword ptr PowersOf.Two[EXPONENT_BIAS*DOUBLE_SIZE-(MAX_FFT_BIT_LENGTH*DOUBLE_SIZE-DOUBLE_SIZE)]
                fchs        // -(2 x pi)/(2^MAX_FFT_BIT_LENGTH)

                fldz

                mov     ecx,MAX_FFT_LENGTH*COMPLEX_SIZE
                sub     eax,eax

      @@05:         fld     st
                    fsincos
                    fstp    qword ptr A_Arr[eax]                //re
                    fstp    qword ptr A_Arr[eax+DOUBLE_SIZE]    //im

                    fadd    st,st(1)
                    add     eax,COMPLEX_SIZE
                    cmp     eax,ecx
                    jl      @@05

      @@06:     //end of the loop
                fstp    st
                fstp    st
      //========================================================================
  end;
  //============================================================================

//==============================================================================
initialization
//==============================================================================
  Unit_Data_Packer2[0]:=0;
  {$IFDEF DISPLAY_VAR_DATA}
  write_unitname_and_datapointers(UNITNAME,@Unit_Data_Packer1,@Unit_Data_Packer2,sizeof(Unit_Data_Packer2));
  {$ENDIF}

  if not (fftw_initialised) then
    nfft_double_init;
//==============================================================================
finalization
//==============================================================================
  {$IFDEF WRITE_FINAL}
  Write(Erroutput,UNITNAME,CRLF);
  {$ENDIF}
//==============================================================================
end.
//==============================================================================

