//==============================================================================
//
//    lossyWAV: Added noise WAV bit reduction method by David Robinson;
//              Noise shaping coefficients by Sebastian Gesemann;
//
//    Copyright (C) 2007-2011 Nick Currie, Copyleft.
//
//    This program is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program.  If not, see <http://www.gnu.org/licenses/>.
//
//    Contact: lossywav <at> hotmail <dot> co <dot> uk
//
//==============================================================================
//
//     fourier.pas  -  Don Cross <dcross <at> intersrv <dot> com>
//
//     This is a Turbo Pascal Unit for calculating the n Fourier Transform
//     (FFT) and the Inverse n Fourier Transform (IFFT).
//     Visit the following URL for the latest version of this code.
//     This page also has a C/C++ version, and a brief discussion of the
//     theory behind the FFT algorithm.
//
//        http://www.intersrv.com/~dcross/fft.html#pascal [!!Dead link!!]
//
//     Revision history [most recent first]:
//
// 2009 May [Nick Currie]
//     Radix Eight added.
//
// 2009 April 24 [Nick Currie]
//     Radix Four added.
//
// 2008 September 29 [Nick Currie]
//     All code transcoded back to Delphi from IA-32/x87.
//
// 2008 May 12 [Nick Currie]
//     All code transcoded to IA-32/x87. FFT_Real implemented to allow calculation
//     of a packed Real array of length 2N to be calculated in a Complex array of
//     length N, with subsequent untangling. ReversedBits, Ai & Ar lookup tables
//     implemented. Calc_Frequency, FFT_Integer and FFT_Integer_Cleanup removed
//     as not used in lossyWAV.
//
// 1996 December 11 [Don Cross]
//     Improved documentation of the procedure CalcFrequency.
//     Fixed some messed up comments in procedure IFFT.
//
// 1996 December 6 [Don Cross]
//     Made procedure 'fft_integer' more efficient when buffer size changes
//     in successive calls:  the buffer is now only resized when the input
//     has more samples, not a differing number of samples.
//     Also changed the way 'fft_integer_cleanup' works so that it is
//     more "bullet-proof".
//
// 1996 December 4 [Don Cross]
//     Adding the procedure 'CalcFrequency', which calculates the FFT
//     at a specific frequency index p=0..n-1, instead of the whole
//     FFT.  This is O(n^2) instead of O(n*log(n)).
//
// 1996 November 30 [Don Cross]
//     Adding a routine to allow FFT of an input array of integers.
//     It is called 'fft_integer'.
//
// 1996 November 18 [Don Cross]
//     Added some comments.
//
// 1996 November 17 [Don Cross]
//     Wrote and debugged first version.
//
//==============================================================================

//==============================================================================
unit nFFT_pas;
//==============================================================================
// Conditional Defines.
//==============================================================================
{$DEFINE RB_LUT}
//==============================================================================
interface
//==============================================================================
uses
  nCore,
  nmaths,
  math;
//==============================================================================
var
  Unit_Data_Packer1 : Packed Record
    Data : Array[0..3] of Integer;
  End;

//==============================================================================
//==============================================================================
// Set the number of samples for FFT analysis;
//==============================================================================
function SetFFTLength(SFL_Samples : Integer) : Integer;

//==============================================================================
//  Calculates the (in place) n Fourier Transform of the array of complex
//  numbers represented by FFT_Input to produce the output complex
//  numbers in FFT_Array.
//==============================================================================
procedure FFT_DIT_Complex;

//==============================================================================
//  Calculates the (in place) n Fourier Transform of the packed array of real
//  numbers represented by FFT_Input to produce the output complex
//  numbers in FFT_Array, then post-processes the results to unpack the
//  Real output in the full FFT_Array.
//==============================================================================
procedure FFT_DIT_Real;

//==============================================================================
//  Calculates the (in place) Inverse n Fourier Transform of the array of
//  complex numbers represented by FFT_Input to produce the output
//  complex numbers in FFT_Array.
//==============================================================================
procedure IFFT_DIT_Complex;
//==============================================================================

//==============================================================================
// Initialise lookup talbes for internal FFT routines.
//==============================================================================
procedure nfft_double_init;
//==============================================================================

//==============================================================================
implementation
//==============================================================================
uses
  fftw_interface;
//==============================================================================
Type
  A_Rec = array[0..MAX_FFT_LENGTH-1] of tDComplex;
  A_Rec_Ptr = ^A_Rec;

  Bytes = Array[0..3] of byte;
  Words = array[0..1] of word;
//==============================================================================
const
  UNITNAME='nFFT_pas';
var
  REVERSED_BYTES        : array[0..255] of byte;

  NumSamples            : Integer;
  NumberOfBitsNeeded    : Integer;
  NumSamples_Reciprocal : Double;
  OmegaStride           : Integer;

  {$IFDEF RB_LUT}
  R_Arr                 : array[0..MAX_FFT_LENGTH-1] of Integer;
  {$ELSE}
  REVERSED_WORDS        : array[0..65535] of word;
  {$ENDIF}
  A_Arr                 : Array[0..MAX_FFT_LENGTH-1] of tDComplex;

  FullOmegaStride       : Integer;

  Unit_Data_Packer2     : array[0..0] of integer;
//==============================================================================

  //============================================================================
  Function SetFFTLength(SFL_Samples : Integer) : Integer;
  //============================================================================
  begin
    if ((SFL_Samples and (SFL_Samples-1))=0) then
    begin
      NumSamples:=SFL_Samples;
      NumberOfBitsNeeded:=floor(nlog2(numsamples));
      NumSamples_Reciprocal:=PowersOf.two[-Numberofbitsneeded];
      SetFFTLength:=NumSamples;
    end
    else
    begin
      NumSamples:=0;
      NumberOfBitsNeeded:=0;
      NumSamples_Reciprocal:=0;
      SetFFTLength:=0;
    end;

    FullOmegaStride:=PowersOf.TwoINT[MAX_FFT_BIT_LENGTH-NumberofBitsNeeded-1];
  end;
  //============================================================================

  //============================================================================
  procedure shuffle_in_place_dcomplex;
  //============================================================================
  var
    sp_i, sp_j : integer;
    sp_c : tDComplex;
  begin
    OmegaStride:=PowersOf.TwoINT[MAX_FFT_BIT_LENGTH-numberofbitsneeded];
    for sp_i:=1 to numsamples-2 do // never need to swap 0 and 2^n-1.
    begin
      {$IFDEF RB_LUT}
      sp_j:=R_Arr[sp_i] shr (MAX_FFT_BIT_LENGTH-numberofbitsneeded);
      {$ELSE}
      sp_j:=((REVERSED_WORDS[Words(sp_i)[0]] shl 16) or REVERSED_WORDS[Words(sp_i)[1]]) shr (32-numberofbitsneeded);
      {$ENDIF}
      if sp_j<sp_i then
      begin
        With FFT_Array.DComplex[sp_i] do
        begin
          sp_c.Re:=Re;
          sp_c.Im:=Im;
          Re:=FFT_Array.DComplex[sp_j].Re;
          Im:=FFT_Array.DComplex[sp_j].Im;
        end;

        With FFT_Array.DComplex[sp_j] do
        Begin
          Re:=sp_c.Re;
          Im:=sp_c.Im;
        End;
      end;
    end;
  end;
  //============================================================================

  //============================================================================
  procedure FFT_DIT_Complex;
  //============================================================================
  var
    BlockBitLen : Integer;

    //==========================================================================
    Procedure RadixEight_DIT;
    //==========================================================================
    Var
      MainBlockSize                   : Integer;
      DataStride                      : Integer;
      D0, D1, D2, D3, D4, D5, D6, D7  : Integer;
      I, N, P, Q                      : Integer;
      V0, V1, V2, V3, V4, V5, V6, V7  : tDComplex;
      X0, X1, X2, X3, X4, X5, X6, X7  : tDComplex;
     {Y0,}Y1, Y2, Y3,{Y4,}Y5, Y6, Y7  : tDComplex;
      U0, U1, U2, U3                  : tDComplex;

    Begin
      DataStride:=PowersOf.TwoINT[blockbitlen];
      blockbitlen:=blockbitlen+3;
      OmegaStride:=PowersOf.TwoINT[MAX_FFT_BIT_LENGTH-BlockBitLen];
      MainBlockSize:=PowersOf.TwoINT[blockbitlen];

      I:=0;
      Repeat
        D0:=I;
        D1:=D0+DataStride;
        D2:=D1+DataStride;
        D3:=D2+DataStride;
        D4:=D3+DataStride;
        D5:=D4+DataStride;
        D6:=D5+DataStride;
        D7:=D6+DataStride;

        V0.Re:=FFT_Array.DCOMPLEX[D0].Re;
        V0.Im:=FFT_Array.DCOMPLEX[D0].Im;
        V1.Re:=FFT_Array.DCOMPLEX[D4].Re;
        V1.Im:=FFT_Array.DCOMPLEX[D4].Im;
        V2.Re:=FFT_Array.DCOMPLEX[D2].Re;
        V2.Im:=FFT_Array.DCOMPLEX[D2].Im;
        V3.Re:=FFT_Array.DCOMPLEX[D6].Re;
        V3.Im:=FFT_Array.DCOMPLEX[D6].Im;
        V4.Re:=FFT_Array.DCOMPLEX[D1].Re;
        V4.Im:=FFT_Array.DCOMPLEX[D1].Im;
        V5.Re:=FFT_Array.DCOMPLEX[D5].Re;
        V5.Im:=FFT_Array.DCOMPLEX[D5].Im;
        V6.Re:=FFT_Array.DCOMPLEX[D3].Re;
        V6.Im:=FFT_Array.DCOMPLEX[D3].Im;
        V7.Re:=FFT_Array.DCOMPLEX[D7].Re;
        V7.Im:=FFT_Array.DCOMPLEX[D7].Im;

        X0.Re:=V0.Re+V4.Re;
        X0.Im:=V0.Im+V4.Im;
        X1.Re:=V1.Re+V7.Re;
        X1.Im:=V1.Im+V7.Im;
        X2.Re:=V2.Re+V6.Re;
        X2.Im:=V2.Im+V6.Im;
        X3.Re:=V3.Re+V5.Re;
        X3.Im:=V3.Im+V5.Im;
        X4.Re:=V0.Re-V4.Re;
        X4.Im:=V0.Im-V4.Im;
        X5.Re:=V1.Re-V7.Re;
        X5.Im:=V1.Im-V7.Im;
        X6.Re:=V2.Re-V6.Re;
        X6.Im:=V2.Im-V6.Im;
        X7.Re:=V3.Re-V5.Re;
        X7.Im:=V3.Im-V5.Im;

        U0.re:=X0.re+X2.re;
        U0.Im:=X0.Im+X2.Im;
        U1.re:=X1.re+X3.re;
        U1.im:=X1.im+X3.im;
        U2.re:=OneOverRoot2*(X1.re-X3.re);
        U2.Im:=OneOverRoot2*(X1.Im-X3.Im);
        U3.re:=OneOverRoot2*(X5.re+X7.re);
        U3.Im:=OneOverRoot2*(X5.Im+X7.Im);

//      Y0.re:=U0.re+U1.re;
//      Y0.im:=U0.im+U1.im;
        Y1.re:=X4.re+U2.re;
        Y1.Im:=X4.Im+U2.Im;
        Y2.re:=X0.re-X2.re;
        Y2.im:=X0.im-X2.im;
        Y3.re:=X4.re-U2.re;
        Y3.Im:=X4.Im-U2.Im;
//      Y4.re:=U0.re-U1.re;
//      Y4.im:=U0.im-U1.im;
        Y5.Re:=U3.Re+X6.re;
        Y5.Im:=U3.Im+X6.Im;
        Y6.Re:=X5.re-X7.re;
        Y6.Im:=X5.im-X7.im;
        Y7.Re:=U3.Re-X6.re;
        Y7.Im:=U3.Im-X6.Im;

        FFT_Array.DCOMPLEX[D0].Re:=(U0.re+U1.re);  //(Y0.Re);
        FFT_Array.DCOMPLEX[D0].Im:=(U0.im+U1.im);  //(Y0.Im);
        FFT_Array.DCOMPLEX[D1].Re:=(Y1.Re+Y5.Im);
        FFT_Array.DCOMPLEX[D1].Im:=(Y1.Im-Y5.Re);
        FFT_Array.DCOMPLEX[D2].Re:=(Y2.Re+Y6.Im);
        FFT_Array.DCOMPLEX[D2].Im:=(Y2.Im-Y6.Re);
        FFT_Array.DCOMPLEX[D3].Re:=(Y3.Re+Y7.Im);
        FFT_Array.DCOMPLEX[D3].Im:=(Y3.Im-Y7.Re);
        FFT_Array.DCOMPLEX[D4].Re:=(U0.re-U1.re);  //(Y4.Re);
        FFT_Array.DCOMPLEX[D4].Im:=(U0.im-U1.im);  //(Y4.Im);
        FFT_Array.DCOMPLEX[D5].Re:=(Y3.Re-Y7.Im);
        FFT_Array.DCOMPLEX[D5].Im:=(Y3.Im+Y7.Re);
        FFT_Array.DCOMPLEX[D6].Re:=(Y2.Re-Y6.Im);
        FFT_Array.DCOMPLEX[D6].Im:=(Y2.Im+Y6.Re);
        FFT_Array.DCOMPLEX[D7].Re:=(Y1.Re-Y5.Im);
        FFT_Array.DCOMPLEX[D7].Im:=(Y1.Im+Y5.Re);

        For N:=1 to DataStride-1 do
        Begin
          D0:=I+N;
          D1:=D0+DataStride;
          D2:=D1+DataStride;
          D3:=D2+DataStride;
          D4:=D3+DataStride;
          D5:=D4+DataStride;
          D6:=D5+DataStride;
          D7:=D6+DataStride;

          P:=N*OmegaStride;
          Q:=P;
          V0.Re:=FFT_Array.DCOMPLEX[D0].Re;
          V0.Im:=FFT_Array.DCOMPLEX[D0].Im;

          V1.Re:=A_Arr[P].Re*FFT_Array.DCOMPLEX[D4].Re-A_Arr[P].Im*FFT_Array.DCOMPLEX[D4].Im;
          V1.Im:=A_Arr[P].Re*FFT_Array.DCOMPLEX[D4].Im+A_Arr[P].Im*FFT_Array.DCOMPLEX[D4].Re;
          P:=P+Q;
          V2.Re:=A_Arr[P].Re*FFT_Array.DCOMPLEX[D2].Re-A_Arr[P].Im*FFT_Array.DCOMPLEX[D2].Im;
          V2.Im:=A_Arr[P].Re*FFT_Array.DCOMPLEX[D2].Im+A_Arr[P].Im*FFT_Array.DCOMPLEX[D2].Re;
          P:=P+Q;
          V3.Re:=A_Arr[P].Re*FFT_Array.DCOMPLEX[D6].Re-A_Arr[P].Im*FFT_Array.DCOMPLEX[D6].Im;
          V3.Im:=A_Arr[P].Re*FFT_Array.DCOMPLEX[D6].Im+A_Arr[P].Im*FFT_Array.DCOMPLEX[D6].Re;
          P:=P+Q;
          V4.Re:=A_Arr[P].Re*FFT_Array.DCOMPLEX[D1].Re-A_Arr[P].Im*FFT_Array.DCOMPLEX[D1].Im;
          V4.Im:=A_Arr[P].Re*FFT_Array.DCOMPLEX[D1].Im+A_Arr[P].Im*FFT_Array.DCOMPLEX[D1].Re;
          P:=P+Q;
          V5.Re:=A_Arr[P].Re*FFT_Array.DCOMPLEX[D5].Re-A_Arr[P].Im*FFT_Array.DCOMPLEX[D5].Im;
          V5.Im:=A_Arr[P].Re*FFT_Array.DCOMPLEX[D5].Im+A_Arr[P].Im*FFT_Array.DCOMPLEX[D5].Re;
          P:=P+Q;
          V6.Re:=A_Arr[P].Re*FFT_Array.DCOMPLEX[D3].Re-A_Arr[P].Im*FFT_Array.DCOMPLEX[D3].Im;
          V6.Im:=A_Arr[P].Re*FFT_Array.DCOMPLEX[D3].Im+A_Arr[P].Im*FFT_Array.DCOMPLEX[D3].Re;
          P:=P+Q;
          V7.Re:=A_Arr[P].Re*FFT_Array.DCOMPLEX[D7].Re-A_Arr[P].Im*FFT_Array.DCOMPLEX[D7].Im;
          V7.Im:=A_Arr[P].Re*FFT_Array.DCOMPLEX[D7].Im+A_Arr[P].Im*FFT_Array.DCOMPLEX[D7].Re;

          X0.Re:=V0.Re+V4.Re;
          X0.Im:=V0.Im+V4.Im;
          X1.Re:=V1.Re+V7.Re;
          X1.Im:=V1.Im+V7.Im;
          X2.Re:=V2.Re+V6.Re;
          X2.Im:=V2.Im+V6.Im;
          X3.Re:=V3.Re+V5.Re;
          X3.Im:=V3.Im+V5.Im;
          X4.Re:=V0.Re-V4.Re;
          X4.Im:=V0.Im-V4.Im;
          X5.Re:=V1.Re-V7.Re;
          X5.Im:=V1.Im-V7.Im;
          X6.Re:=V2.Re-V6.Re;
          X6.Im:=V2.Im-V6.Im;
          X7.Re:=V3.Re-V5.Re;
          X7.Im:=V3.Im-V5.Im;

          U0.re:=X0.re+X2.re;
          U0.Im:=X0.Im+X2.Im;
          U1.re:=X1.re+X3.re;
          U1.im:=X1.im+X3.im;
          U2.re:=OneOverRoot2*(X1.re-X3.re);
          U2.Im:=OneOverRoot2*(X1.Im-X3.Im);
          U3.re:=OneOverRoot2*(X5.re+X7.re);
          U3.Im:=OneOverRoot2*(X5.Im+X7.Im);

//        Y0.re:=U0.re+U1.re;
//        Y0.im:=U0.im+U1.im;
          Y1.re:=X4.re+U2.re;
          Y1.Im:=X4.Im+U2.Im;
          Y2.re:=X0.re-X2.re;
          Y2.im:=X0.im-X2.im;
          Y3.re:=X4.re-U2.re;
          Y3.Im:=X4.Im-U2.Im;
//        Y4.re:=U0.re-U1.re;
//        Y4.im:=U0.im-U1.im;
          Y5.Re:=U3.Re+X6.re;
          Y5.Im:=U3.Im+X6.Im;
          Y6.Re:=X5.re-X7.re;
          Y6.Im:=X5.im-X7.im;
          Y7.Re:=U3.Re-X6.re;
          Y7.Im:=U3.Im-X6.Im;

          FFT_Array.DCOMPLEX[D0].Re:=(U0.re+U1.re);  //(Y0.Re);
          FFT_Array.DCOMPLEX[D0].Im:=(U0.im+U1.im);  //(Y0.Im);
          FFT_Array.DCOMPLEX[D1].Re:=(Y1.Re+Y5.Im);
          FFT_Array.DCOMPLEX[D1].Im:=(Y1.Im-Y5.Re);
          FFT_Array.DCOMPLEX[D2].Re:=(Y2.Re+Y6.Im);
          FFT_Array.DCOMPLEX[D2].Im:=(Y2.Im-Y6.Re);
          FFT_Array.DCOMPLEX[D3].Re:=(Y3.Re+Y7.Im);
          FFT_Array.DCOMPLEX[D3].Im:=(Y3.Im-Y7.Re);
          FFT_Array.DCOMPLEX[D4].Re:=(U0.re-U1.re);  //(Y4.Re);
          FFT_Array.DCOMPLEX[D4].Im:=(U0.im-U1.im);  //(Y4.Im);
          FFT_Array.DCOMPLEX[D5].Re:=(Y3.Re-Y7.Im);
          FFT_Array.DCOMPLEX[D5].Im:=(Y3.Im+Y7.Re);
          FFT_Array.DCOMPLEX[D6].Re:=(Y2.Re-Y6.Im);
          FFT_Array.DCOMPLEX[D6].Im:=(Y2.Im+Y6.Re);
          FFT_Array.DCOMPLEX[D7].Re:=(Y1.Re-Y5.Im);
          FFT_Array.DCOMPLEX[D7].Im:=(Y1.Im+Y5.Re);
        End;
        I:=I+MainBlockSize;
      Until I=NumSamples;
    //==========================================================================
    End;
    //==========================================================================

    //==========================================================================
    Procedure RadixFour_DIT;
    //==========================================================================
    Var
      MainBlockSize       : Integer;
      DataStride          : Integer;
      D0, D1, D2, D3      : Integer;
      I, N                : Integer;
      V0, V1, V2, V3      : tDComplex;
      X0, X1, X2, X3      : tDComplex;
    Begin
      DataStride    :=PowersOf.TwoINT[blockbitlen];
      blockbitlen   :=blockbitlen+2;
      OmegaStride   :=PowersOf.TwoINT[MAX_FFT_BIT_LENGTH-BlockBitLen];
      MainBlockSize :=PowersOf.TwoINT[blockbitlen];

      I:=0;
      Repeat
        D0:=I;
        D1:=D0+DataStride;
        D2:=D1+DataStride;
        D3:=D2+DataStride;

        V0.Re:=FFT_Array.DCOMPLEX[D0].Re;
        V0.Im:=FFT_Array.DCOMPLEX[D0].Im;
        V1.Re:=FFT_Array.DCOMPLEX[D2].Re;
        V1.Im:=FFT_Array.DCOMPLEX[D2].Im;
        V2.Re:=FFT_Array.DCOMPLEX[D1].Re;
        V2.Im:=FFT_Array.DCOMPLEX[D1].Im;
        V3.Re:=FFT_Array.DCOMPLEX[D3].Re;
        V3.Im:=FFT_Array.DCOMPLEX[D3].Im;

        X0.Re:=V0.Re+V2.Re;
        X0.Im:=V0.Im+V2.Im;
        X1.Re:=V1.Re+V3.Re;
        X1.Im:=V1.Im+V3.Im;
        X2.Re:=V0.Re-V2.Re;
        X2.Im:=V0.Im-V2.Im;
        X3.Re:=V3.Im-V1.Im;
        X3.Im:=V1.Re-V3.Re;

        FFT_Array.DCOMPLEX[D0].Re:=(X0.Re+X1.Re);
        FFT_Array.DCOMPLEX[D0].Im:=(X0.Im+X1.Im);
        FFT_Array.DCOMPLEX[D1].Re:=(X2.Re-X3.Re);
        FFT_Array.DCOMPLEX[D1].Im:=(X2.Im-X3.Im);
        FFT_Array.DCOMPLEX[D2].Re:=(X0.Re-X1.Re);
        FFT_Array.DCOMPLEX[D2].Im:=(X0.Im-X1.Im);
        FFT_Array.DCOMPLEX[D3].Re:=(X2.Re+X3.Re);
        FFT_Array.DCOMPLEX[D3].Im:=(X2.Im+X3.Im);

        For N:=1 to DataStride-1 do
        Begin
          D0:=I+N;
          D1:=D0+DataStride;
          D2:=D1+DataStride;
          D3:=D2+DataStride;

          V0.Re:=FFT_Array.DCOMPLEX[D0].Re;
          V0.Im:=FFT_Array.DCOMPLEX[D0].Im;

          V1.Re:=A_Arr[N*OmegaStride].Re*FFT_Array.DCOMPLEX[D2].Re-A_Arr[N*OmegaStride].Im*FFT_Array.DCOMPLEX[D2].Im;
          V1.Im:=A_Arr[N*OmegaStride].Re*FFT_Array.DCOMPLEX[D2].Im+A_Arr[N*OmegaStride].Im*FFT_Array.DCOMPLEX[D2].Re;

          V2.Re:=A_Arr[N*2*OmegaStride].Re*FFT_Array.DCOMPLEX[D1].Re-A_Arr[N*2*OmegaStride].Im*FFT_Array.DCOMPLEX[D1].Im;
          V2.Im:=A_Arr[N*2*OmegaStride].Re*FFT_Array.DCOMPLEX[D1].Im+A_Arr[N*2*OmegaStride].Im*FFT_Array.DCOMPLEX[D1].Re;

          V3.Re:=A_Arr[N*3*OmegaStride].Re*FFT_Array.DCOMPLEX[D3].Re-A_Arr[N*3*OmegaStride].Im*FFT_Array.DCOMPLEX[D3].Im;
          V3.Im:=A_Arr[N*3*OmegaStride].Re*FFT_Array.DCOMPLEX[D3].Im+A_Arr[N*3*OmegaStride].Im*FFT_Array.DCOMPLEX[D3].Re;

          X0.Re:=V0.Re+V2.Re;
          X0.Im:=V0.Im+V2.Im;
          X1.Re:=V1.Re+V3.Re;
          X1.Im:=V1.Im+V3.Im;
          X2.Re:=V0.Re-V2.Re;
          X2.Im:=V0.Im-V2.Im;
          X3.Re:=V3.Im-V1.Im;
          X3.Im:=V1.Re-V3.Re;

          FFT_Array.DCOMPLEX[D0].Re:=(X0.Re+X1.Re);
          FFT_Array.DCOMPLEX[D0].Im:=(X0.Im+X1.Im);
          FFT_Array.DCOMPLEX[D1].Re:=(X2.Re-X3.Re);
          FFT_Array.DCOMPLEX[D1].Im:=(X2.Im-X3.Im);
          FFT_Array.DCOMPLEX[D2].Re:=(X0.Re-X1.Re);
          FFT_Array.DCOMPLEX[D2].Im:=(X0.Im-X1.Im);
          FFT_Array.DCOMPLEX[D3].Re:=(X2.Re+X3.Re);
          FFT_Array.DCOMPLEX[D3].Im:=(X2.Im+X3.Im);
        End;
        I:=I+MainBlockSize;
      Until I=NumSamples;
    //==========================================================================
    End;
    //==========================================================================

    //==========================================================================
    Procedure RadixTwo_DIT;
    //==========================================================================
    var
      MainBlockSize : Integer;
      DataStride    : Integer;
      D0, D1        : Integer;
      I, N          : Integer;
      V0, V1        : tDComplex;
    Begin
      DataStride    :=PowersOf.TwoINT[blockbitlen];
      blockbitlen   :=blockbitlen+1;
      OmegaStride   :=PowersOf.TwoINT[MAX_FFT_BIT_LENGTH-BlockBitLen];
      MainBlockSize :=PowersOf.TwoINT[blockbitlen];

      I:=0;
      repeat
        D0:=I;
        D1:=D0+DataStride;

        V0.Re:=FFT_Array.DCOMPLEX[D0].Re;
        V0.Im:=FFT_Array.DCOMPLEX[D0].Im;
        V1.Re:=FFT_Array.DCOMPLEX[D1].Re;
        V1.Im:=FFT_Array.DCOMPLEX[D1].Im;

        FFT_Array.DCOMPLEX[D0].Re:=V0.Re+V1.Re;
        FFT_Array.DCOMPLEX[D0].Im:=V0.Im+V1.Im;
        FFT_Array.DCOMPLEX[D1].Re:=V0.Re-V1.Re;
        FFT_Array.DCOMPLEX[D1].Im:=V0.Im-V1.Im;

        for N:=1 to DataStride - 1 do
        begin
          D0:=I+N;
          D1:=D0+DataStride;

          V0.Re:=FFT_Array.DCOMPLEX[D0].Re;
          V0.Im:=FFT_Array.DCOMPLEX[D0].Im;

          V1.Re:=A_Arr[N*OmegaStride].Re*FFT_Array.DCOMPLEX[D1].Re-A_Arr[N*OmegaStride].Im*FFT_Array.DCOMPLEX[D1].Im;
          V1.Im:=A_Arr[N*OmegaStride].Re*FFT_Array.DCOMPLEX[D1].Im+A_Arr[N*OmegaStride].Im*FFT_Array.DCOMPLEX[D1].Re;

          FFT_Array.DCOMPLEX[D0].Re:=V0.Re+V1.Re;
          FFT_Array.DCOMPLEX[D0].Im:=V0.Im+V1.Im;
          FFT_Array.DCOMPLEX[D1].Re:=V0.Re-V1.Re;
          FFT_Array.DCOMPLEX[D1].Im:=V0.Im-V1.Im;
        end;
        I:=I+MainBlockSize;
      until (I=NumSamples);
    //==========================================================================
    End;
    //==========================================================================

  //============================================================================
  begin
  //============================================================================
    Shuffle_in_place_DComplex;

    blockbitlen:=0;

    While numberofbitsneeded-blockbitlen>2 do
      RadixEight_DIT;

    While numberofbitsneeded-blockbitlen>1 do
      RadixFour_DIT;

    While Numberofbitsneeded-blockbitlen>0 do
      RadixTwo_DIT;
  //============================================================================
  end;
  //============================================================================

  //============================================================================
  procedure FFT_DIT_Real;
  //============================================================================
  var
    V1, X0, X1 : tDComplex;
    J, K, L, M, N : Integer;
  begin
    FFT_DIT_Complex;

    J:=0;
    N:=Powersof.TwoINT[numberofbitsneeded-1];
    Repeat
      J:=J+1;
      K:=NumSamples-J;
      L:=Numsamples+K;
      M:=Numsamples+J;

      X0.Re:=single_half*(FFT_Array.DCOMPLEX[J].Re+FFT_Array.DCOMPLEX[K].Re);
      X0.Im:=single_half*(FFT_Array.DCOMPLEX[J].Im+FFT_Array.DCOMPLEX[K].Im);

      X1.Re:=single_half*(FFT_Array.DCOMPLEX[J].Re-FFT_Array.DCOMPLEX[K].Re);
      X1.Im:=single_half*(FFT_Array.DCOMPLEX[J].Im-FFT_Array.DCOMPLEX[K].Im);

      V1.Re:=A_Arr[J*FullOmegaStride].Re*X0.Im+A_Arr[J*FullOmegaStride].Im*X1.Re;
      V1.Im:=A_Arr[J*FullOmegaStride].Re*X1.Re-A_Arr[J*FullOmegaStride].Im*X0.Im;

	    FFT_Array.DCOMPLEX[J].Re:=(X0.Re+V1.Re);
      FFT_Array.DCOMPLEX[J].Im:=(X1.Im-V1.Im);

      FFT_Array.DCOMPLEX[L].Re:=FFT_Array.DCOMPLEX[J].Re;
      FFT_Array.DCOMPLEX[L].Im:=-FFT_Array.DCOMPLEX[J].Im;

      FFT_Array.DCOMPLEX[M].Re:=(X0.Re-V1.Re);
      FFT_Array.DCOMPLEX[M].Im:=(X1.Im+V1.Im);

      FFT_Array.DCOMPLEX[K].Re:=FFT_Array.DCOMPLEX[M].Re;
      FFT_Array.DCOMPLEX[K].Im:=-FFT_Array.DCOMPLEX[M].Im;
    Until J=N;

    FFT_Array.DCOMPLEX[NumSamples].Re:=FFT_Array.DCOMPLEX[0].Re-FFT_Array.DCOMPLEX[0].Im;
    FFT_Array.DCOMPLEX[Numsamples].Im:=0;

    FFT_Array.DCOMPLEX[0].Re:=FFT_Array.DCOMPLEX[0].Re+FFT_Array.DCOMPLEX[0].Im;
    FFT_Array.DCOMPLEX[0].Im:=0;
  end;
  //============================================================================

  //============================================================================
  procedure IFFT_DIT_Complex;
  //============================================================================
  var
    BlockBitLen   : Integer;
    MainBlockSize : Integer;
    DataStride    : Integer;
    I, J, K, N    : Integer;
    V             : tDComplex;
  begin
    Shuffle_in_Place_DComplex;

    for blockBitLen:=1 to NumberOfBitsNeeded do
    begin
      I:=0;
      OmegaStride:=PowersOf.TwoINT[MAX_FFT_BIT_LENGTH-BlockBitLen];

      MainBlockSize:=PowersOf.TwoINT[blockbitlen];

      DataStride:=MainBlockSize shr 1;

      repeat
        J:=I;
        for N:=0 to DataStride - 1 do
        begin
          K:=J + DataStride;

          V.Re:=A_Arr[N*OmegaStride].Re*FFT_Array.DCOMPLEX[K].Re+A_Arr[N*OmegaStride].Im*FFT_Array.DCOMPLEX[K].Im;
          V.Im:=A_Arr[N*OmegaStride].Re*FFT_Array.DCOMPLEX[K].Im-A_Arr[N*OmegaStride].Im*FFT_Array.DCOMPLEX[K].Re;

          FFT_Array.DCOMPLEX[K].Re:=FFT_Array.DCOMPLEX[J].Re-V.Re;
          FFT_Array.DCOMPLEX[K].Im:=FFT_Array.DCOMPLEX[J].Im-V.Im;

          FFT_Array.DCOMPLEX[J].Re:=FFT_Array.DCOMPLEX[J].Re+V.Re;
          FFT_Array.DCOMPLEX[J].Im:=FFT_Array.DCOMPLEX[J].Im+V.Im;

          J:=J+1;
        end;
        I:=I+MainBlockSize;
      until (I=NumSamples);
    end;
  //============================================================================
  end;
  //============================================================================

  //============================================================================
  procedure nfft_double_init;
  //============================================================================
  var
    nf_i, nf_j : Integer;
    delta : Double;
  begin
    //for nf_i:=0 to MAX_FFT_LENGTH*2-1 do
    //  FFT_Array.Real[nf_i]:=0;

    //==========================================================================
    // Create 256 byte reversed-bits lookup table
    //==========================================================================
    for nf_i:=0 to 15 do
      for nf_j:=0 to 15 do
        REVERSED_BYTES[nf_j shl 4 + nf_i]:=REVERSED_NIBBLES[nf_j] or (REVERSED_NIBBLES[nf_i] shl 4);
    //==========================================================================
    {$IFNDEF RB_LUT}
    for nf_i:=0 to 65535 do
      REVERSED_WORDS[nf_i]:=(REVERSED_BYTES[nf_i and 255] shl 8) or REVERSED_BYTES[nf_i shr 8];
    {$ENDIF}

    delta:=-TwoPi*PowersOf.two[-MAX_FFT_BIT_LENGTH];
    for nf_i:=0 to MAX_FFT_LENGTH-1 do
    begin
      {$IFDEF RB_LUT}
      R_Arr[nf_i]:=(((REVERSED_BYTES[Bytes(nf_i)[0]] shl 8 + REVERSED_BYTES[Bytes(nf_i)[1]]) shl 8 + REVERSED_BYTES[Bytes(nf_i)[2]]) shl 8 + REVERSED_BYTES[Bytes(nf_i)[3]]) shr (32-MAX_FFT_BIT_LENGTH);
      {$ENDIF}
      With A_Arr[nf_i] do
      begin
        Re:=Cos(nf_i*delta);
        Im:=Sin(nf_i*delta);
      End;
    end;
  //============================================================================
  End;
  //============================================================================

//==============================================================================
initialization
//==============================================================================
  Unit_Data_Packer2[0]:=0;
  {$IFDEF DISPLAY_VAR_DATA}
  write_unitname_and_datapointers(UNITNAME,@Unit_Data_Packer1,@Unit_Data_Packer2,sizeof(Unit_Data_Packer2));
  {$ENDIF}

  if not fftw_initialised then
    nfft_double_init;
//==============================================================================
finalization
//==============================================================================
  {$IFDEF WRITE_FINAL}
  Write(Erroutput,UNITNAME,CRLF);
  {$ENDIF}
//==============================================================================
end.
//==============================================================================

