//==============================================================================
//
//    lossyWAV: Added noise WAV bit reduction method by David Robinson;
//              Noise shaping coefficients by Sebastian Gesemann;
//
//    Copyright (C) 2007-2011 Nick Currie, Copyleft.
//
//    This program is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program.  If not, see <http://www.gnu.org/licenses/>.
//
//    Contact: lossywav <at> hotmail <dot> co <dot> uk
//
//==============================================================================
unit nFillFFT;
//==============================================================================
interface
//==============================================================================
uses
  nCore;
type
  DoubleArray = array[0..0] of double;
//==============================================================================
var
  Unit_Data_Packer1 : Packed Record
    Data : Array[0..3] of Integer;
  End;

  //============================================================================
  // window function lookup table.
  //============================================================================
  window_function               : array[0..MAX_FFT_BIT_LENGTH] of ^DoubleArray;

  FillFFT : Record
    Window_Function_RMS         : Double;
    Codec_Block_fill_offset     : Integer;
    actual_analysis_block_start : Integer;
  End;

//==============================================================================

Procedure nFillFFT_Init();
Procedure Fill_Audio_Safes;

Function  FillFFT_Input_From_Data : Extended;
Function  FillFFT_Input_From_BTRD : Extended;
Function  FillFFT_Input_From_CORR : Extended;

//==============================================================================
implementation
//==============================================================================
uses
  nmaths,
  math;
//==============================================================================
Const
  UNITNAME='nFillFFT';
var
  //============================================================================
  // Fill_FFT_Input procedure data lookup table.
  //============================================================================
  SavedBlocks : Record
    BTRD  : MultiChannelCodecBlock;
    CORR  : MultiChannelCodecBlock;
  End;

  fill_fft_lookup   : array[0..MAX_BLOCK_SIZE*4-1] of packed record // 4 blocks in memory (prev, last, this, next).
                        block  : integer;
                        offset : integer;
                      end;

  {$A4}
  Unit_Data_Packer2 : Record
    Data : Array[0..0] of Integer;
  End;
  {$A8}
//==============================================================================

  {$IFNDEF USEASM}
  //============================================================================
  Procedure nFillFFT_Init;
  //============================================================================
  Var
    ff_i : Integer;
  //============================================================================
  Begin
    for ff_i:=0 to Global.Codec_Block_size*4-1 do
      with fill_fft_lookup[ff_i] do
      begin
        block:=(ff_i div Global.Codec_Block_size);
        offset:=(ff_i mod Global.Codec_Block_size);
      end;
    FillFFT.Codec_Block_fill_offset:=Global.Codec_Block_size*2;
  //============================================================================
  End;
  //============================================================================
  {$ELSE}
  //============================================================================
  Procedure nFillFFT_Init(); Assembler;
  //============================================================================
  Asm
                push    ebx
                push    edi

                mov     eax,Global.Codec_Block_size
                shl     eax,1
                mov     FillFFT.Codec_Block_fill_offset,eax
                shl     eax,1

                jmp     @@02

      @@01:     mov     ecx,Global.Codec_Block_bits
                mov     edx,dword ptr PowersOf.twom1[ecx*int64_size]

                mov     ebx,eax
                mov     edi,eax
                and     edi,edx
                shr     ebx,cl

                mov     dword ptr fill_fft_lookup[eax*(INTEGER_SIZE*2)],ebx
                mov     dword ptr fill_fft_lookup[eax*(INTEGER_SIZE*2)+INTEGER_SIZE],edi

      @@02:     sub     eax,1
                jge     @@01

                pop     edi
                pop     ebx
  //============================================================================
  End;
  //============================================================================
  {$ENDIF}

  {$IFNDEF USEASM}
  //============================================================================
  Function Apply_Window_Function(Filled_Average : Extended) : Extended;
  //============================================================================
  Var
    ff_i : Integer;
    ff_l, ff_m : Double;
  Begin
  //============================================================================
    ff_l:=0;

    ff_i:=0;
    while ff_i<Global.FFT.Length do
    Begin
      ff_m:=FFT_Array.DReal[ff_i]-Filled_Average;
      ff_l:=ff_l+sqr(ff_m);
      {$R-}
      FFT_Array.DReal[ff_i]:=ff_m*window_function[Global.FFT.Bit_Length]^[ff_i];
      {$R+}
      inc(ff_i);
    End;
    Apply_Window_Function:=nlog2(ff_l*Global.FFT.length_recip)*SINGLE_HALF;
  //============================================================================
  End;
  //============================================================================
  {$ELSE}
  //============================================================================
  Function Apply_Window_Function_Int : Extended;
  //============================================================================
  Asm
  //============================================================================
                push    eax
                push    ecx

                fldz
                fxch    st(1)

                xor     ecx,ecx
                mov     eax,Global.FFT.bit_length
                mov     eax,dword ptr window_function[eax*POINTER_SIZE]

                {$IFDEF ALIGN08}
                  {$IFDEF ALIGN16}
                    {$IFDEF ALIGN32}
                    {$ENDIF}
                  {$ENDIF}
                {$ENDIF}

      @@10:       fld     qword ptr FFT_Array.DReal[ecx*DOUBLE_SIZE]
                  fsub    st,st(1)

                  fld     st
                  fmul    st,st
                  faddp   st(3),st

                  fmul    qword ptr [eax+ecx*DOUBLE_SIZE]
                  fstp    qword ptr FFT_Array[ecx*DOUBLE_SIZE]

                  add     ecx,1
                  cmp     ecx,Global.FFT.length
                  jl      @@10

                fstp    st

                fmul    Global.FFT.Length_Recip
                call    nlog2_int
                fmul    SINGLE_HALF

                pop     ecx
                pop     eax

                //mov     Global.Jump_Location,offset @@10
  //============================================================================
  End;
  //============================================================================
  {$ENDIF}

  //============================================================================
  {$IFNDEF USEASM}
  //============================================================================
  function FillFFT_Input_From_Data : extended;
  //============================================================================
  var
    ff_i, ff_j, ff_n : Integer;
    ff_k, ff_m : Double;
  //============================================================================
  begin
    ff_k:=0;
    ff_n:=FillFFT.Codec_Block_fill_offset+FillFFT.actual_analysis_block_start;

    ff_i:=0;
    while ff_i<Global.FFT.Length do
    begin
      ff_j:=ff_i+ff_n;
      ff_m:=BlockData.Pointer[fill_fft_lookup[ff_j].block]^[Global.channel,fill_fft_lookup[ff_j].offset]*settings.scaling_factor;
      ff_k:=ff_k+ff_m;
      fft_array.DReal[ff_i]:=ff_m;
      inc(ff_i);
    end;

    FillFFT_Input_From_Data:=Apply_Window_Function(ff_k*Global.FFT.Length_Recip*settings.dccorrect_Multiplier);
  end;
  //============================================================================
  {$ELSE}
  //============================================================================
  function FillFFT_Input_From_Data : extended; Assembler;
  //============================================================================
  asm
  //============================================================================
                push      ebx
                push      edi
                push      esi

                xor       edi,edi
                xor       edx,edx

                fldz

                mov       eax,Global.channel
                mov       eax,dword ptr BlockData.Channel_Offset[eax*POINTER_SIZE]

                mov       ebx,FillFFT.Codec_Block_fill_offset
                add       ebx,FillFFT.actual_analysis_block_start
                shl       ebx,(integer_shift+1)

      @@00:       mov       edx,dword ptr fill_fft_lookup[ebx+edi*(2*INTEGER_SIZE)]
                  mov       edx,dword ptr BlockData.Pointer[edx*POINTER_SIZE]
                  add       edx,eax

                  mov       esi,dword ptr fill_fft_lookup[ebx+edi*(2*INTEGER_SIZE)+INTEGER_SIZE]

                  fild      dword ptr [edx+esi*INTEGER_SIZE]
                  fmul      settings.scaling_factor

                  fadd      st(1),st
                  fstp      qword ptr FFT_Array[edi*DOUBLE_SIZE]

                  add       edi,1
                  cmp       edi,Global.FFT.length
                  jl        @@00

                pop       esi
                pop       edi
                pop       ebx

                fmul      Global.FFT.Length_Recip
                fmul      settings.dccorrect_Multiplier
                call      Apply_Window_Function_Int

                //mov       Global.Jump_Location,offset @@00
  //============================================================================
  end;
  //============================================================================
  {$ENDIF}

  {$IFNDEF USEASM}
  //============================================================================
  function FillFFT_Input_from_BTRD : Extended;
  //============================================================================
  Var
    ff_i, ff_j : Integer;
    ff_k, ff_m : Double;
  //============================================================================
  begin
  //============================================================================
    ff_k:=0;
    ff_i:=0;

    while ff_i<Global.FFT.Length do
    begin
      ff_j:=ff_i+FillFFT.actual_analysis_block_start;

      if ff_j<0 then
        ff_m:=SavedBlocks.BTRD[Global.Channel,ff_j+Global.Codec_Block_Size]
      else
        if ff_j>Global.Codec_Block_size then
          ff_m:=BlockData.Pointer[3]^[Global.Channel,ff_j-Global.Codec_Block_Size]*settings.scaling_factor
        else
          ff_m:=BlockData.BTRD[Global.Channel,ff_j];

      ff_k:=ff_k+ff_m;
      fft_array.DReal[ff_i]:=ff_m;
      inc(ff_i);
    end;

    FillFFT_Input_From_BTRD:=Apply_Window_Function(ff_k*Global.FFT.Length_Recip*settings.dccorrect_Multiplier);
  //============================================================================
  end;
  //============================================================================
  {$ELSE}
  //============================================================================
  function FillFFT_Input_from_BTRD : Extended; Assembler;
  //============================================================================
  asm
  //============================================================================
                push    ebx
                push    ecx
                push    edi
                push    esi

                fldz

                mov     edx,Global.Channel
                mov     esi,Global.Codec_Block_Size
                xor     edi,edi

                mov     ecx,dword ptr BlockData.Channel_Offset[edx*POINTER_SIZE]
                mov     ebx,dword ptr BlockData.Pointer[3*POINTER_SIZE]
                add     ebx,ecx

                {$IFDEF ALIGN08}
                nop;
                  {$IFDEF ALIGN16}
                    {$IFDEF ALIGN32}
                    nop;nop;nop;nop;nop;nop;nop;nop;nop;nop;nop;nop;nop;nop;nop;nop;
                    {$ENDIF}
                  {$ENDIF}
                {$ENDIF}

      @@J0:       mov     eax,FillFFT.actual_analysis_block_start
                  add     eax,edi

                  cmp     eax,0
                  jl      @@J1

                  cmp     eax,esi
                  jl      @@J2

                  sub     eax,esi
                  fild    dword ptr [ebx+eax*INTEGER_SIZE]
                  fmul    settings.scaling_factor
                  jmp     @@J4

                  {$IFDEF ALIGN08}
                    {$IFDEF ALIGN16}
                    nop;nop;nop;
                      {$IFDEF ALIGN32}
                      {$ENDIF}
                    {$ENDIF}
                  {$ENDIF}

      @@J1:       add     eax,esi
                  fild    dword ptr SavedBlocks.BTRD[ecx+eax*INTEGER_SIZE]
                  jmp     @@J4

                  {$IFDEF ALIGN08}
                  nop;nop;nop;nop;nop;
                    {$IFDEF ALIGN16}
                      {$IFDEF ALIGN32}
                      nop;nop;nop;nop;nop;nop;nop;nop;nop;nop;nop;nop;nop;nop;nop;nop;
                      {$ENDIF}
                    {$ENDIF}
                  {$ENDIF}

      @@J2:       fild    dword ptr BlockData.BTRD[ecx+eax*INTEGER_SIZE]
                  {$IFDEF ALIGN08}
                  nop;
                    {$IFDEF ALIGN16}
                    nop;nop;nop;nop;nop;nop;nop;nop;
                      {$IFDEF ALIGN32}
                      nop;nop;nop;nop;nop;nop;nop;nop;nop;nop;nop;nop;nop;nop;nop;nop;
                      {$ENDIF}
                    {$ENDIF}
                  {$ENDIF}

      @@J4:       fadd    st(1),st
                  fstp    qword ptr FFT_Array[edi*DOUBLE_SIZE]

                  inc     edi
                  cmp     edi,Global.FFT.Length
                  jl      @@J0

                pop     esi
                pop     edi
                pop     ecx
                pop     ebx

                fmul    Global.FFT.Length_Recip
                fmul    settings.dccorrect_Multiplier
                call    Apply_Window_Function_Int
                //mov     Global.Jump_Location,offset @@J4
  //============================================================================
  end;
  //============================================================================
  {$ENDIF}

  {$IFNDEF USEASM}
  //============================================================================
  function FillFFT_Input_from_CORR : Extended;
  //============================================================================
  Var
    ff_i, ff_j : Integer;
    ff_k, ff_m : Double;
  //============================================================================
  begin
  //============================================================================
    ff_k:=0;

    ff_i:=0;
    while ff_i<Global.FFT.Length do
    begin
      ff_j:=ff_i+FillFFT.actual_analysis_block_start;

      if ff_j<0 then
        ff_m:=SavedBlocks.CORR[Global.Channel,ff_j+Global.Codec_Block_Size]
      else
        if ff_j>Global.Codec_Block_size then
          ff_m:=BlockData.Pointer[3]^[Global.Channel,ff_j-Global.Codec_Block_Size]*settings.scaling_factor
        else
          ff_m:=BlockData.CORR[Global.Channel,ff_j];

      ff_k:=ff_k+ff_m;
      fft_array.DReal[ff_i]:=ff_m;
      inc(ff_i);
    end;

    FillFFT_Input_From_CORR:=Apply_Window_Function(ff_k*Global.FFT.Length_Recip*settings.dccorrect_Multiplier);
  //============================================================================
  end;
  //============================================================================
  {$ELSE}
  //============================================================================
  function FillFFT_Input_from_CORR : Extended; Assembler;
  //============================================================================
  asm
  //============================================================================
                push    ebx
                push    ecx
                push    edi
                push    esi

                fldz

                mov     edx,Global.Channel
                mov     esi,Global.Codec_Block_Size
                xor     edi,edi

                mov     ecx,dword ptr BlockData.Channel_Offset[edx*POINTER_SIZE]
                mov     ebx,dword ptr BlockData.Pointer[3*POINTER_SIZE]
                add     ebx,ecx

                {$IFDEF ALIGN08}
                nop;nop;nop;nop;nop;
                  {$IFDEF ALIGN16}
                    {$IFDEF ALIGN32}
                    nop;nop;nop;nop;nop;nop;nop;nop;nop;nop;nop;nop;nop;nop;nop;nop;
                    {$ENDIF}
                  {$ENDIF}
                {$ENDIF}

      @@J0:       mov     eax,FillFFT.actual_analysis_block_start
                  add     eax,edi

                  cmp     eax,0
                  jl      @@J1

                  cmp     eax,esi
                  jl      @@J2

                  sub     eax,esi
                  fild    dword ptr [ebx+eax*INTEGER_SIZE]
                  fmul    settings.scaling_factor
                  jmp     @@J4

                  {$IFDEF ALIGN08}
                  nop;nop;nop;
                    {$IFDEF ALIGN16}
                      {$IFDEF ALIGN32}
                      {$ENDIF}
                    {$ENDIF}
                  {$ENDIF}

      @@J1:       add     eax,esi
                  fild    dword ptr SavedBlocks.CORR[ecx+eax*INTEGER_SIZE]
                  jmp     @@J4

                  {$IFDEF ALIGN08}
                  nop;nop;nop;nop;nop;
                    {$IFDEF ALIGN16}
                      {$IFDEF ALIGN32}
                      nop;nop;nop;nop;nop;nop;nop;nop;nop;nop;nop;nop;nop;nop;nop;nop;
                      {$ENDIF}
                    {$ENDIF}
                  {$ENDIF}

      @@J2:       fild    dword ptr BlockData.CORR[ecx+eax*INTEGER_SIZE]

                  {$IFDEF ALIGN08}
                  nop;
                    {$IFDEF ALIGN16}
                    nop;nop;nop;nop;nop;nop;nop;nop;
                      {$IFDEF ALIGN32}
                      nop;nop;nop;nop;nop;nop;nop;nop;nop;nop;nop;nop;nop;nop;nop;nop;
                      {$ENDIF}
                    {$ENDIF}
                  {$ENDIF}

      @@J4:       fadd    st(1),st
                  fstp    qword ptr FFT_Array[edi*DOUBLE_SIZE]

                  inc     edi
                  cmp     edi,Global.FFT.Length
                  jl      @@J0

                pop     esi
                pop     edi
                pop     ecx
                pop     ebx

                fmul    Global.FFT.Length_Recip
                fmul    settings.dccorrect_Multiplier
                call    Apply_Window_Function_Int

                //mov     Global.Jump_Location,offset @@J4
  //============================================================================
  end;
  //============================================================================
  {$ENDIF}

  {$IFNDEF USEASM}
  //============================================================================
  Procedure Fill_Audio_Safes;
  //============================================================================
  Var
    ff_i, ff_j : Integer;
  //============================================================================
  Begin
  //============================================================================
    for ff_j:=0 to Global.Channels-1 do
      For ff_i:=0 to Global.Codec_Block_Size-1 do
      Begin
        SavedBlocks.BTRD[ff_j,ff_i]:=BlockData.BTRD[ff_j,ff_i];
        SavedBlocks.CORR[ff_j,ff_i]:=BlockData.CORR[ff_j,ff_i];
      End;
  //============================================================================
  End;
  //============================================================================
  {$ELSE}
  //============================================================================
  Procedure Fill_Audio_Safes;
  //============================================================================
  Asm
  //============================================================================
                push    eax
                push    ebx
                push    ecx
                push    edx

                xor     ecx,ecx // channel
                mov     ebx,Global.Codec_Block_Size
                shl     ebx,INTEGER_SHIFT

                {$IFDEF ALIGN08}
                nop;
                  {$IFDEF ALIGN16}
                  nop;nop;nop;nop;nop;nop;nop;nop;
                    {$IFDEF ALIGN32}
                    {$ENDIF}
                  {$ENDIF}
                {$ENDIF}

    @@00:       mov     eax,dword ptr BlockData.Channel_Offset[ecx*integer_size]
                mov     edx,eax
                add     edx,ebx

                {$IFDEF ALIGN08}
                  nop;nop;nop;nop;nop;
                  {$IFDEF ALIGN16}
                    {$IFDEF ALIGN32}
                    nop;nop;nop;nop;nop;nop;nop;nop;nop;nop;nop;nop;nop;nop;nop;nop;
                    {$ENDIF}
                  {$ENDIF}
                {$ENDIF}

    @@01:       fild    qword ptr blockdata.btrd[eax]
                fild    qword ptr blockdata.corr[eax]
                fistp   qword ptr savedblocks.corr[eax]
                fistp   qword ptr savedblocks.btrd[eax]

                fild    qword ptr blockdata.btrd[eax]+8
                fild    qword ptr blockdata.corr[eax]+8
                fistp   qword ptr savedblocks.corr[eax]+8
                fistp   qword ptr savedblocks.btrd[eax]+8

                fild    qword ptr blockdata.btrd[eax]+16
                fild    qword ptr blockdata.corr[eax]+16
                fistp   qword ptr savedblocks.corr[eax]+16
                fistp   qword ptr savedblocks.btrd[eax]+16

                fild    qword ptr blockdata.btrd[eax]+24
                fild    qword ptr blockdata.corr[eax]+24
                fistp   qword ptr savedblocks.corr[eax]+24
                fistp   qword ptr savedblocks.btrd[eax]+24

                fild    qword ptr blockdata.btrd[eax]+32
                fild    qword ptr blockdata.corr[eax]+32
                fistp   qword ptr savedblocks.corr[eax]+32
                fistp   qword ptr savedblocks.btrd[eax]+32

                fild    qword ptr blockdata.btrd[eax]+40
                fild    qword ptr blockdata.corr[eax]+40
                fistp   qword ptr savedblocks.corr[eax]+40
                fistp   qword ptr savedblocks.btrd[eax]+40

                fild    qword ptr blockdata.btrd[eax]+48
                fild    qword ptr blockdata.corr[eax]+48
                fistp   qword ptr savedblocks.corr[eax]+48
                fistp   qword ptr savedblocks.btrd[eax]+48

                fild    qword ptr blockdata.btrd[eax]+56
                fild    qword ptr blockdata.corr[eax]+56
                fistp   qword ptr savedblocks.corr[eax]+56
                fistp   qword ptr savedblocks.btrd[eax]+56

                add     eax,64

                cmp     eax,edx
                jb      @@01

                add     ecx,1
                cmp     ecx,Global.Channels
                jb      @@00

                pop     edx
                pop     ecx
                pop     ebx
                pop     eax

                //mov     Global.Jump_Location,offset @@01
  //============================================================================
  End;
  //============================================================================
  {$ENDIF}

  //============================================================================
  Procedure Window_Function_Lookup_Init;
  //============================================================================
  var
    ff_i, ff_j : Integer;
  //============================================================================
  begin
    // Generate window_function lookup tables
    for ff_j:=0 to MAX_FFT_BIT_LENGTH do
    begin
      getmem(window_function[ff_j],powersof.twoint[ff_j+3]);
      for ff_i:=0 to PowersOf.twom1[ff_j] do
      Begin
        {$R-}
          window_function[ff_j]^[ff_i]:=Hann(ff_i,ff_j);
        {$R+}
      End;
    end;
    FillFFT.window_function_rms:=log2(3/8)/2*db_lg2x20;     // == RMS of Hann Window x 20 x log10(2).
  //============================================================================
  end;
  //============================================================================

  //============================================================================
  procedure window_function_lookup_final;
  //============================================================================
  var
    ff_i : integer;
  begin
    for ff_i:=0 to max_fft_bit_length do
      freemem(window_function[ff_i]);
  end;
  //============================================================================

//==============================================================================
initialization
//==============================================================================
  Unit_Data_Packer2.Data[0]:=0;
  {$IFDEF DISPLAY_VAR_DATA}
  write_unitname_and_datapointers(UNITNAME,@Unit_Data_Packer1,@Unit_Data_Packer2,sizeof(Unit_Data_Packer2));
  {$ENDIF}

  Window_Function_Lookup_Init();
//==============================================================================

//==============================================================================
finalization
//==============================================================================
  {$IFDEF WRITE_FINAL}
  Write(Erroutput,UNITNAME,CRLF);
  {$ENDIF}
  Window_Function_Lookup_Final;
//==============================================================================

//==============================================================================
end.
//==============================================================================
