//==============================================================================
//
//    lossyWAV: Added noise WAV bit reduction method by David Robinson;
//              Noise shaping coefficients by Sebastian Gesemann;
//
//    Copyright (C) 2007-2011 Nick Currie, Copyleft.
//
//    This program is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program.  If not, see <http://www.gnu.org/licenses/>.
//
//    Contact: lossywav <at> hotmail <dot> co <dot> uk
//
//==============================================================================
unit nInitialise;
//==============================================================================
interface
//==============================================================================
var
  Unit_Data_Packer1 : Packed Record
    Data : Array[0..3] of Integer;
  End;
//==============================================================================
Procedure Check_Switches;
procedure initial_setup;
//==============================================================================
implementation
//==============================================================================
uses
  nCore,
  FFTW_interface,
  nparameter,
  nFillFFT,
  nmaths,
  nprocess,
  nremovebits,
  nSGNS,
  nspreading,
  nwav,
  math,
  sysutils;
//==============================================================================
const
  UNITNAME                        = 'nInitialise';
  //============================================================================
  // Quality Preset Data.
  //============================================================================
  QUALITY_PRESET_MIN              = -5;
  QUALITY_PRESET_MAX              = 10;
  //============================================================================
  // constants directly related to quality_integer settings.
  //============================================================================               //      -5,      -4,      -3,      -2,      -1,       0,       1,       2,       3,       4,       5,       6,       7,       8,       9,       10
  QUALITY_MINIMUM_BITS_TO_KEEP    : array[QUALITY_PRESET_MIN..QUALITY_PRESET_MAX] of Single   = (   2.710,   2.848,   2.986,   3.124,   3.262,   3.400,   3.470,   3.540,   3.610,   3.680,   3.750,   4.000,   4.250,   4.500,   4.750,   5.000); //1.2.3a
  QUALITY_NOISE_THRESHOLD_SHIFTS  : array[QUALITY_PRESET_MIN..QUALITY_PRESET_MAX] of Single   = (  16.667,  14.444,  12.222,  10.000,   7.778,   5.556,   4.444,   3.333,   2.222,   1.111,   0.000,  -2.222,  -4.444,  -6.667,  -8.889, -11.111); //1.2.3a
  QUALITY_SIGNAL_TO_NOISE_RATIOS  : array[QUALITY_PRESET_MIN..QUALITY_PRESET_MAX] of Single   = ( -19.500, -20.500, -21.500, -22.500, -23.500, -24.500, -25.000, -25.500, -26.000, -26.500, -27.000, -28.000, -29.000, -30.000, -31.000, -32.000); //1.2.3a
  QUALITY_CLIPS_PER_CHANNEL       : array[QUALITY_PRESET_MIN..QUALITY_PRESET_MAX] of Integer  = (       3,       3,       3,       3,       3,       2,       2,       2,       2,       2,       1,       1,       1,       0,       0,       0); //1.2.3a
  QUALITY_UPPER_CALC_FREQ_LIMIT   : array[QUALITY_PRESET_MIN..QUALITY_PRESET_MAX] of Integer  = (   15159,   15332,   15504,   15676,   15848,   16021,   16107,   16193,   16279,   16365,   16451,   16538,   16624,   16710,   16796,   16882); //1.2.3a
  QUALITY_AUTO_SHAPING_FACTOR     : array[QUALITY_PRESET_MIN..QUALITY_PRESET_MAX] of Single   = (0.077903,0.104718,0.133335,0.162936,0.195383,0.239690,0.265891,0.304356,0.340499,0.380561,0.422840,0.517277,0.621117,0.734487,0.863206,1.000000);

  //============================================================================
  // Quality_Noise_Threshold_Delta is the correction to the calculated noise
  // average when using DC offset corrected FFT input data (default).
  //============================================================================
  QUALITY_NOISE_THRESHOLD_DELTA   : array[QUALITY_PRESET_MIN..QUALITY_PRESET_MAX] of Single   = (2.420359,2.296165,2.141537,1.964923,1.753599,1.424952,1.337800,1.191097,1.096227,0.923858,0.765617,0.350196,0.000000,0.000000,0.000000,0.000000);
//QUALITY_NOISE_THRESHOLD_DELTA   : array[QUALITY_PRESET_MIN..QUALITY_PRESET_MAX] of Single   = (  2.4240,  2.3010,  2.1370,  1.9580,  1.7710,  1.4130,  1.3200,  1.1850,  1.0970,  0.9290,  0.7740,  0.3780,  0.0000,  0.0000,  0.0000,  0.0000);
  //============================================================================

//==============================================================================
var
  Unit_Data_Packer2 : Array[0..2] of Integer;
//==============================================================================

  //============================================================================
  // Check command line parameters from nParameter;
  //============================================================================
  procedure check_switches;
  //============================================================================
  begin
  //============================================================================
    settings.quality_double:=Parameters.quality;
    settings.quality_integer:=floor(settings.quality_double);
    settings.quality_fraction:=settings.quality_double-settings.quality_integer;

    if Parameters.analyses<>3 then
      Strings.Parameter:=Strings.Parameter+' --analyses '+inttostr(Parameters.analyses);

    if Parameters.scaling<>-1 then
    begin
      settings.scaling_factor:=Parameters.scaling;
      Strings.Parameter:=Strings.Parameter+' --scale '+floattostrf(settings.scaling_factor,fffixed,99,6);
    end
    else
      settings.scaling_factor:=1;
    settings.scaling_factor_inv:=1/settings.scaling_factor;

    if Parameters.linkchannels or Parameters.midside then
      Strings.Parameter:=Strings.Parameter+' --linkchannels';

    if parameters.shaping>-1 then
      Strings.Parameter:=Strings.Parameter+' --shaping '+floattostrf(parameters.shaping,fffixed,99,4);

    if (parameters.adaptive_shaping=-1) or (parameters.adaptive_shaping=-3) then
      Strings.Parameter:=Strings.Parameter+' --adaptive off'
    else
      if (Parameters.adaptive_shaping<>-2) or (not parameters.adaptive_shaping_warp) then
      Begin
        Strings.Parameter:=Strings.Parameter+' --adaptive';

        if Parameters.Adaptive_Shaping>0 then
          Strings.Parameter:=Strings.Parameter + ' '+inttostr(parameters.adaptive_shaping);

        if not parameters.adaptive_shaping_warp then
          Strings.Parameter:=Strings.Parameter + ' nowarp';
      End;

    if Parameters.midside then
      Strings.Parameter:=Strings.Parameter+' --midside';

    if parameters.dccorrect then
      settings.dccorrect_multiplier:=1
    else
    begin
      settings.dccorrect_multiplier:=0;
      Strings.Parameter:=Strings.Parameter+' --nodccorrect';
    end;

    if parameters.underlap>0 then
      Strings.Parameter:=Strings.Parameter + ' --underlap '+inttostr(parameters.underlap);

    if Parameters.maxclips>-1 then
    begin
      settings.maximum_clips_per_channel:=Parameters.maxclips;
      Strings.Parameter:=Strings.Parameter+' --maxclips '+inttostr(Parameters.maxclips);
    end
    else
      settings.maximum_clips_per_channel:=QUALITY_CLIPS_PER_CHANNEL[min(QUALITY_PRESET_MAX,settings.quality_integer+ord(settings.quality_fraction<>0))]

  //============================================================================
  end;
  //============================================================================

  //============================================================================
  // Setup most of the internal arrays of values to be used in the analyses;
  //============================================================================
  procedure initial_setup;
  //============================================================================
  var
    sa_i                                  : Integer;
    this_channel                          : Integer;
    this_analysis_number                  : Integer;
    this_threshold_shift                  : Double;
    this_dccorrect_shift                  : Double;
    this_shaping_factor                   : double;
  begin
    settings.static_minimum_bits_to_keep:=min(Global.bits_per_sample-4,parameters.static);

    if parameters.static>STATIC_MINIMUM_BITS_TO_KEEP then
      Strings.Parameter:=Strings.Parameter+'--static '+inttostr(Parameters.static);

    if Parameters.limit<>-1 then
    begin
      Global.Upper_Freq_Limit:=Min(Floor(Global.Sample_Rate*0.453515),Parameters.limit);
      Strings.Parameter:=Strings.Parameter + ' --limit '+floattostrf(Global.Upper_Freq_Limit,ffFixed,99,0);
    end
    else
      Global.Upper_Freq_Limit:=Min(Floor(Global.Sample_Rate*0.453515),QUALITY_UPPER_CALC_FREQ_LIMIT[settings.quality_integer]);

    Frequency_Limits[Spread_Zones+1]:=Global.Upper_Freq_Limit;

    Global.Lower_Freq_Limit:=LOWER_FREQ_LIMIT;
    if global.channels<=2 then
      Parameters.perchannel:=true;

    History.Histogram_Multiplier:=Powersof.Two[6-Global.Bits_Per_Sample];
    History.Histogram_Length:=64;
    History.Histogram_Offset:=History.Histogram_Length shr 1;

    //==========================================================================
    // Make an sample-rate adjustment to noise_threshold_shift.
    //==========================================================================
    this_threshold_shift:=-db_lg2x20*Global.Codec_Block_Bit_Shift*SINGLE_HALF;

    //==========================================================================
    // Set noise-thresholds, adjusting  for samplerate - NB: most tuning was
    // performed using 44.1kHz samples!! (second tweak)
    //==========================================================================
    if Settings.quality_integer=QUALITY_PRESET_MAX then
    begin
      this_dccorrect_shift:=QUALITY_NOISE_THRESHOLD_DELTA[QUALITY_PRESET_MAX];
      settings.noise_threshold_shift_minimum:=QUALITY_NOISE_THRESHOLD_SHIFTS[QUALITY_PRESET_MAX];
      settings.noise_threshold_shift_average:=QUALITY_SIGNAL_TO_NOISE_RATIOS[QUALITY_PRESET_MAX];
      settings.dynamic_minimum_bits_to_keep:=QUALITY_MINIMUM_BITS_TO_KEEP[QUALITY_PRESET_MAX];
      this_shaping_factor:=QUALITY_AUTO_SHAPING_FACTOR[QUALITY_PRESET_MAX];
    end
    else
    begin
      this_dccorrect_shift:=(Settings.quality_fraction*(QUALITY_NOISE_THRESHOLD_DELTA[Settings.quality_integer+1]-QUALITY_NOISE_THRESHOLD_DELTA[Settings.quality_integer])+QUALITY_NOISE_THRESHOLD_DELTA[Settings.quality_integer]);
      settings.noise_threshold_shift_minimum:=Settings.quality_fraction*(QUALITY_NOISE_THRESHOLD_SHIFTS[Settings.quality_integer+1]-QUALITY_NOISE_THRESHOLD_SHIFTS[Settings.quality_integer])+QUALITY_NOISE_THRESHOLD_SHIFTS[Settings.quality_integer];
      settings.noise_threshold_shift_average:=Settings.quality_fraction*(QUALITY_SIGNAL_TO_NOISE_RATIOS[Settings.quality_integer+1]-QUALITY_SIGNAL_TO_NOISE_RATIOS[Settings.quality_integer])+QUALITY_SIGNAL_TO_NOISE_RATIOS[Settings.quality_integer];
      settings.dynamic_minimum_bits_to_keep :=Settings.quality_fraction*(QUALITY_MINIMUM_BITS_TO_KEEP[Settings.quality_integer+1]-QUALITY_MINIMUM_BITS_TO_KEEP[Settings.quality_integer])+QUALITY_MINIMUM_BITS_TO_KEEP[Settings.quality_integer];
      this_shaping_factor:=Settings.quality_fraction*(QUALITY_AUTO_SHAPING_FACTOR[Settings.quality_integer+1]-QUALITY_AUTO_SHAPING_FACTOR[Settings.quality_integer])+QUALITY_AUTO_SHAPING_FACTOR[Settings.quality_integer];
    end;

    this_dccorrect_shift:=this_dccorrect_shift*settings.dccorrect_multiplier;
    settings.noise_threshold_shift_average:=settings.noise_threshold_shift_average+this_dccorrect_shift+this_threshold_shift;
    settings.noise_threshold_shift_minimum:=settings.noise_threshold_shift_minimum+this_threshold_shift;

    if parameters.dynamic<>-1 then
    Begin
      Strings.Parameter:=Strings.Parameter+' --dynamic '+floattostrf(Parameters.dynamic,fffixed,99,3);
      settings.dynamic_minimum_bits_to_keep:=parameters.dynamic;
    End;

    if parameters.shaping=-2 then
      parameters.shaping:=this_shaping_factor;

    //==========================================================================
    // Select which FFT analyses to perform.
    //==========================================================================
    settings.FFT_analysis_switches:=SHORT_FFT_LENGTH or LONG_FFT_LENGTH;

    if (Parameters.analyses>2) then
      settings.FFT_analysis_switches:=settings.FFT_analysis_switches or IMPULSE_FFT_LENGTH;

    if (Parameters.analyses>3) then
      settings.FFT_analysis_switches:=settings.FFT_analysis_switches or ADDED_FFT_LENGTH_512;

    if (Parameters.analyses>4) then
      settings.FFT_analysis_switches:=settings.FFT_analysis_switches or ADDED_FFT_LENGTH_256;

    if (Parameters.analyses>5) then
      settings.FFT_analysis_switches:=settings.FFT_analysis_switches or ADDED_FFT_LENGTH_128;

    //==========================================================================
    // Shift analysis flags to reflect actual codec-block-size.
    //==========================================================================
    settings.FFT_analysis_switches:=floor(settings.FFT_analysis_switches*powersof.two[Global.Codec_Block_Bit_Shift]);
    //==========================================================================

    Set_FFT_Data_Rec(History.FFT,SHORT_FFT_BIT_LENGTH+Global.Codec_Block_Bit_Shift);
    Set_FFT_Data_Rec(LongDist.FFT,Global.Codec_Block_Bits);

    for sa_i:=1 to PRECALC_ANALYSES do
    Begin
      Global.Analysis.Bits[sa_i]:=History.FFT.Bit_Length-2+sa_i;
      Global.Analysis.Length[sa_i]:=Powersof.TwoInt[Global.Analysis.Bits[sa_i]];
    End;

    if Parameters.freqdist or (Parameters.Adaptive_Shaping<>-1) then
      settings.raw_result_short:=History.FFT.Length
    else
      settings.raw_result_short:=0;

    if Parameters.adaptive_shaping<>-1 then
    Begin
      Set_FFT_Data_Rec(SGNSFFT.FFT,LONG_FFT_BIT_LENGTH+Global.Codec_Block_Bit_Shift);
      settings.raw_result_sgns:=SGNSFFT.FFT.length;
      nSGNS_Initialise;
    End
    else
      settings.raw_result_sgns:=0;

    if Global.Upper_Freq_Limit>Global.Sample_Rate*0.475 then
      Global.Upper_Freq_Limit:=floor(Global.Sample_Rate*0.475);

    if (not Parameters.silent) then
    begin
      if (Parameters.verbosity) then
      begin
        Write(ErrOutput,'Filename  : ',WAVFilePrintName,CRLF,'File Info : ',Global.Sample_Rate*OneOver[1000]:0:2,'kHz; ',Global.channels,' channel; ',Global.Bits_Per_Sample,' bit; ');

        if Global.Wave_Size<MAX_WAVE_SIZE then
        begin
          time_string_make(strings.time,Global.Total_Samples/Global.Sample_Rate);
          size_string_make(strings.size,Global.Total_Samples*Cardinal(Global.channels)*Cardinal(Global.Bytes_Per_Sample));
          Write(ErrOutput,strings.time,', ',strings.size);
        end;

        Write(ErrOutput,CRLF+'Progress  : ');
      end
      else
        Write(ErrOutput,WAVFilePrintName,';');
    end;
    flush(ErrOutput);

    for this_channel:=0 to Global.channels-1 do
    begin
      Global.channel:=this_channel;
      for this_analysis_number:=1 to PRECALC_ANALYSES do
        results.saved_FFT_results[Global.channel,this_analysis_number].start:=-1;
    end;

    Global.clipped_samples:=0;

    //==========================================================================

  //============================================================================
    nProcess_Init;
  //============================================================================
    nSpreading_Init;
  //============================================================================
    nFillFFT_Init;    //dependent on codec_block_size.
  //============================================================================
    nRemoveBits_Init; //bitdepth and samplerate dependent.
  //============================================================================

    Global.output_blocks:=floor(PowersOf.two[20-Global.codec_block_bits]*OneOver[Global.channels]);
    Global.last_print:=0;

    Global.blocks_processed:=1;
    Global.blocks_processed_recip:=1;

    Global.bits_removed:=0;
    Global.bits_lost:=0;
  //============================================================================
  end;
  //============================================================================

//==============================================================================
initialization
//==============================================================================
  Unit_Data_Packer2[0]:=0;
  {$IFDEF DISPLAY_VAR_DATA}
  write_unitname_and_datapointers(UNITNAME,@Unit_Data_Packer1,@Unit_Data_Packer2,sizeof(Unit_Data_Packer2));
  {$ENDIF}
//==============================================================================
finalization
//==============================================================================
  {$IFDEF WRITE_FINAL}
  Write(Erroutput,UNITNAME,CRLF);
  {$ENDIF}
//==============================================================================
end.
//==============================================================================
