//==============================================================================
//
//    lossyWAV: Added noise WAV bit reduction method by David Robinson;
//              Noise shaping coefficients by Sebastian Gesemann;
//
//    Copyright (C) 2007-2011 Nick Currie, Copyleft.
//
//    This program is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program.  If not, see <http://www.gnu.org/licenses/>.
//
//    Contact: lossywav <at> hotmail <dot> co <dot> uk
//
//==============================================================================
unit nMaths;
//==============================================================================
interface
//==============================================================================
uses
  nCore;
//==============================================================================
var
  Unit_Data_Packer1 : Packed Record
    Data : Array[0..3] of Integer;
  End;
//==============================================================================

//==============================================================================
// Window function calculation;
//==============================================================================
function  Hann(index : integer; fft_bit_length : Integer) : Extended;

function  nround10(nr_v : double; nr_d : Integer) : Extended;

{$IFDEF USEASM}
procedure npower2_int;
procedure nlog2_int;
procedure nroot_int;
{$ENDIF}

function  npower2(fp_2v : double) : Extended;
function  nlog2(fl_2v : double) : extended;
function  nroot(fr_v : double) : extended;

//==============================================================================
implementation
//==============================================================================
//{$DEFINE LOW_PRECISION}
uses
  math;
const
  UNITNAME = 'nMaths';
var
  temp_integer        : Integer;
  Unit_Data_Packer2   : Array[0..1] of Integer;
//==============================================================================

  //============================================================================
  function Hann(index : integer; fft_bit_length : Integer) : Extended;
  //============================================================================
  begin
  //============================================================================
    Hann:=single_half*(single_one-cos(index*twopi*PowersOf.two[-fft_bit_length]));
  //============================================================================
  end;
  //============================================================================

  //============================================================================
  function nround10(nr_v : double; nr_d : integer) : Extended;
  //============================================================================
  begin
  //============================================================================
    if (abs(nr_d)<307) then
      nround10:=round(nr_v*PowersOf.ten[nr_d])*PowersOf.ten[-nr_d]
    else
      nround10:=nr_v;
  //============================================================================
  end;
  //============================================================================

  {$IFDEF USEASM}
  //============================================================================
  procedure npower2_int; assembler;
  //============================================================================
  asm
                push    eax
                fist    temp_integer
                fisub   temp_integer
                f2xm1
                mov     eax,temp_integer
                fadd    single_one
                fmul    qword ptr PowersOf.Two[EXPONENT_BIAS*double_size+eax*double_size]
                pop     eax
  end;
  //============================================================================

  //============================================================================
  function npower2(fp_2v : double) : extended; assembler;
  //============================================================================
  asm
                fld     fp_2v
                call    npower2_int
  end;
  //============================================================================

  //============================================================================
  procedure nlog2_int; assembler;
  //============================================================================
  asm
                fxam
                push    eax
                fstsw   ax
                and     ah,$43
                jnz     @@01

                fld1
                pop     eax
                fxch    st(1)
                fyl2x
                ret

  @@01:         fstp    st
                pop     eax
                fldz
  end;
  //============================================================================

  //============================================================================
  function nlog2(fl_2v : double) : Extended; assembler;
  //============================================================================
  asm
                fld     fl_2v
                call    nlog2_int
  end;
  //============================================================================

  //============================================================================
  function npower10(fp_10v : double) : extended; assembler;
  //============================================================================
  asm
                fldl2t
                push    eax
                fmul    fp_10v
                fist    temp_integer
                fisub   temp_integer
                f2xm1
                fadd    single_one
                mov     eax,temp_integer
                fmul    qword ptr PowersOf.Two[EXPONENT_BIAS*double_size+eax*double_size]
                pop     eax
  end;
  //============================================================================

  //============================================================================
  procedure nroot_int; assembler;
  //============================================================================
  asm
                push    eax
                fxam
                fstsw   ax
                and     ah,$43
                jnz     @@01

                fsqrt
                pop     eax
                ret

  @@01:         fstp    st
                pop     eax
                fldz
  end;
  //============================================================================

  //============================================================================
  function nroot(fr_v : double) : extended; assembler;
  //============================================================================
  asm
                fld     fr_v
                call    nroot_int
  end;
  //============================================================================

  {$ELSE}

  //============================================================================
  function npower2(fp_2v : double) : extended;
  //============================================================================
  begin
  //============================================================================
    npower2:=power(2,fp_2v);
  //============================================================================
  end;
  //============================================================================

  //============================================================================
  function nlog2(fl_2v : double) : extended;
  //============================================================================
  begin
  //============================================================================
    nlog2:=0;
    if fl_2v>0 then
      nlog2:=log2(fl_2v);
  //============================================================================
  end;
  //============================================================================

  //============================================================================
  function nlog10(fl_10v : double) : extended;
  //============================================================================
  begin
  //============================================================================
    nlog10:=0;
    if fl_10v>0 then
      nlog10:=log10(fl_10v);
  //============================================================================
  end;
  //============================================================================

  //============================================================================
  function nroot(fr_v : double) : extended; assembler;
  //============================================================================
  begin
  //============================================================================
    nroot:=0;
    if fr_v>0 then
      nroot:=sqrt(fr_v);
  //============================================================================
  end;
  //============================================================================
  {$ENDIF}
  //============================================================================

//==============================================================================
initialization
//==============================================================================
  Unit_Data_Packer2[0]:=0;
  temp_integer:=0;
  {$IFDEF DISPLAY_VAR_DATA}
  write_unitname_and_datapointers(UNITNAME,@Unit_Data_Packer1,@Unit_Data_Packer2,sizeof(Unit_Data_Packer2));
  {$ENDIF}
//==============================================================================
finalization
//==============================================================================
  {$IFDEF WRITE_FINAL}
  Write(Erroutput,UNITNAME,CRLF);
  {$ENDIF}
//==============================================================================
end.
//==============================================================================

