//==============================================================================
//
//    lossyWAV: Added noise WAV bit reduction method by David Robinson;
//              Noise shaping coefficients by Sebastian Gesemann;
//
//    Copyright (C) 2007-2011 Nick Currie, Copyleft.
//
//    This program is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program.  If not, see <http://www.gnu.org/licenses/>.
//
//    Contact: lossywav <at> hotmail <dot> co <dot> uk
//
//==============================================================================
unit nRemoveBits;
//==============================================================================
interface
//==============================================================================
uses
  nCore;
//==============================================================================
var
  Unit_Data_Packer1 : Packed Record
    Data : Array[0..3] of Integer;
  End;

  remove_bits_data  : packed record
    bits_to_remove  : Integer;
    clips           : Integer;
    bits_removed    : Integer;
  end;

  //============================================================================
  // Max sample lookup table for rounding.
  //============================================================================
  sample_maxima                           : array[0..BITS_TO_CALCULATE-1] of Integer;
//==============================================================================

//==============================================================================

Procedure Remove_Bits();

Procedure nRemoveBits_Init();

//==============================================================================
implementation
//==============================================================================
uses
  nMaths,
  nSGNS,
  math;
//==============================================================================
const
  UNITNAME                              = 'nRemoveBits';
var
  RemoveData : Record
    this_two_power                        : Double;
    this_two_power_recip                  : Double;

    this_min_sample                       : Integer;
    this_max_sample                       : Integer;

    this_temp_sample                      : Int64;
  End;

  {$A4}
  Unit_Data_Packer2 : Record
    Data : Array[0..3] of Integer;
  End;
  {$A8}

//==============================================================================
  {$IFNDEF USEASM}
  //============================================================================
  Function Limited_Sample(input_sample : Extended) : Integer;
  //============================================================================
  Begin
  //============================================================================
    if Input_Sample>RemoveData.this_max_sample then
    begin
      Input_Sample:=RemoveData.this_max_sample;
      inc(remove_bits_data.clips);
    end
    else
      if Input_Sample<RemoveData.this_min_sample then
      begin
        Input_Sample:=RemoveData.this_min_sample;
        inc(remove_bits_data.clips);
      end;

    Limited_Sample:=floor(Input_Sample);
  //============================================================================
  End;
  //============================================================================

  //============================================================================
  Procedure Remove_Bits_Proc_Adaptive_Noise_Shaping_On;
  //============================================================================
  Var
    rb_i                  : integer;
    signal                : Int64;
    scaled                : Double;
    output                : Int64;
  Begin
  //============================================================================
    remove_bits_data.clips:=0;

    RemoveData.this_max_sample:=sample_maxima[remove_bits_data.bits_removed];
    RemoveData.this_two_power:=PowersOf.twoint[remove_bits_data.bits_removed];
    RemoveData.this_two_power_recip:=PowersOf.two[-remove_bits_data.bits_removed];

    Restore_Filter;
    for rb_i:=0 to BlockData.Size.This-1 do
    begin
      signal:=BlockData.Pointer[2]^[Global.Channel,rb_i];
      scaled:=signal*settings.scaling_factor;

      output:=Floor(limited_sample(round((scaled + Warped_Lattice_Filter_Evaluate)*RemoveData.this_two_power_recip)*RemoveData.this_two_power));

      Warped_Lattice_Filter_Update(output-scaled);

      BlockData.BTRD[Global.channel,rb_i]:=output;
      BlockData.CORR[Global.channel,rb_i]:=integer(signal-round(output*settings.scaling_factor_inv));
    end;
  //============================================================================
  End;
  //============================================================================

  //============================================================================
  Procedure Remove_Bits_Proc_Shaping_Off;
  //============================================================================
  Var
    rb_i                  : Integer;
    signal                : int64;
    scaled                : Double;
    output                : int64;
  Begin
  //============================================================================
    remove_bits_data.clips:=0;
    RemoveData.this_max_sample:=sample_maxima[remove_bits_data.bits_removed];
    RemoveData.this_two_power:=PowersOf.twoint[remove_bits_data.bits_removed];
    RemoveData.this_two_power_recip:=PowersOf.two[-remove_bits_data.bits_removed];

    for rb_i:=0 to BlockData.Size.This-1 do
    begin
      signal:=BlockData.Pointer[2]^[Global.channel,rb_i];
      scaled:=signal*settings.scaling_factor;

      output:=floor(limited_sample(round(scaled*RemoveData.this_two_power_recip)*RemoveData.this_two_power));

      BlockData.BTRD[Global.channel,rb_i]:=output;
      BlockData.CORR[Global.channel,rb_i]:=signal-round(output*settings.scaling_factor_inv);
    end;
  //============================================================================
  End;
  //============================================================================
  {$ELSE}
  //============================================================================
  Procedure Remove_Bits_Proc_Adaptive_Noise_Shaping_On; Assembler;
  //============================================================================
  Asm
  //============================================================================
                push      ebx
                push      ebp
                push      edi
                push      esi

                xor       esi,esi

                mov       ebx,Global.Channel
                mov       ebx,dword ptr BlockData.Channel_Offset[ebx*integer_size]

                mov       edi,BlockData.Size.This
                shl       edi,INTEGER_SHIFT
                add       edi,ebx

                mov       ebp,dword ptr BlockData.Pointer[2*Pointer_Size]

                mov       eax,remove_bits_data.bits_removed
                mov       edx,dword ptr sample_maxima[eax*INTEGER_SIZE]
                mov       RemoveData.this_max_sample,edx

                fld       qword ptr PowersOf.Two[EXPONENT_BIAS*DOUBLE_SIZE+eax*DOUBLE_SIZE]
                fstp      RemoveData.this_two_power
                neg       eax
                fld       qword ptr PowersOf.Two[EXPONENT_BIAS*DOUBLE_SIZE+eax*DOUBLE_SIZE]
                fstp      RemoveData.this_two_power_recip

                //==============================================================
                // Remove Bits - ANS
                //==============================================================
                call      restore_filter

                {$IFDEF ALIGN08}
                nop;nop;nop;nop;nop;
                {$IFDEF ALIGN16}
                nop;nop;nop;nop;nop;nop;nop;nop;
                {$ENDIF}
                {$ENDIF}

      @@81:     fild      dword ptr [ebp+ebx]
                fmul      settings.scaling_factor

                call      Warped_Lattice_Filter_Evaluate
                fadd      st,st(1)
                fmul      RemoveData.this_two_power_recip
                fistp     RemoveData.this_temp_sample        // temp storage to round st(0)
                fild      RemoveData.this_temp_sample        // and reload rounded value.
                fmul      RemoveData.this_two_power

                ficom     RemoveData.this_max_sample
                fstsw     ax
                and       ah,$41
                jnz       @@82

                fstp      st
                inc       esi
                fild      RemoveData.this_max_sample
                jmp       @@83

                {$IFDEF ALIGN08}
                nop;nop;nop;nop;nop;nop;nop;
                {$IFDEF ALIGN16}
                nop;nop;nop;nop;nop;nop;nop;nop;
                {$ENDIF}
                {$ENDIF}

      @@82:     ficom     RemoveData.this_min_sample
                fstsw     ax
                xor       ah,$01
                and       ah,$41
                jnz       @@83

                fstp      st
                inc       esi
                fild      RemoveData.this_min_sample

                {$IFDEF ALIGN08}
                nop;nop;nop;nop;nop;nop
                {$ENDIF}

      @@83:     fist      dword ptr BlockData.BTRD[ebx]
                fsubr
                call      Warped_Lattice_Filter_Update_Int

                fild      dword ptr BlockData.BTRD[ebx]
                fmul      settings.scaling_factor_inv
                fisubr    dword ptr [ebp+ebx]
                fistp     dword ptr BlockData.CORR[ebx]

                add       ebx,INTEGER_SIZE
                cmp       ebx,edi
                jl        @@81
                //==============================================================
                // End of Remove Bits - ANS
                //==============================================================

      @@F0:     mov       remove_bits_data.clips,esi
                pop       esi
                pop       edi
                pop       ebp
                pop       ebx
  //============================================================================
  End;
  //============================================================================

  //============================================================================
  procedure remove_bits_proc_shaping_off; assembler;
  //============================================================================
  asm
  //============================================================================
                push      ebx
                push      ebp
                push      edi
                push      esi

                xor       esi,esi

                mov       ebx,Global.Channel
                mov       ebx,dword ptr BlockData.Channel_Offset[ebx*integer_size]

                mov       edi,BlockData.Size.This
                shl       edi,INTEGER_SHIFT
                add       edi,ebx

                mov       ebp,dword ptr BlockData.Pointer[2*Pointer_Size]

                mov       eax,remove_bits_data.bits_removed
                mov       edx,dword ptr sample_maxima[eax*INTEGER_SIZE]
                mov       RemoveData.this_max_sample,edx
                fld       qword ptr PowersOf.Two[EXPONENT_BIAS*DOUBLE_SIZE+eax*DOUBLE_SIZE]
                fstp      RemoveData.this_two_power
                neg       eax
                fld       qword ptr PowersOf.Two[EXPONENT_BIAS*DOUBLE_SIZE+eax*DOUBLE_SIZE]
                fstp      RemoveData.this_two_power_recip

                {$IFDEF ALIGN08}
                nop;nop;nop;nop;nop;nop;
                {$ENDIF}

                //==============================================================
                // Remove Bits
                //==============================================================
      @@A1:     fild      dword ptr [ebp+ebx]
                fmul      settings.scaling_factor
                fmul      RemoveData.this_two_power_recip
//                frndint
                fistp     RemoveData.this_temp_sample        // temp storage to round st(0)
                fild      RemoveData.this_temp_sample        // and reload rounded value.
                fmul      RemoveData.this_two_power

                ficom     RemoveData.this_max_sample
                fstsw     ax
                and       ah,$41
                jnz       @@A2

                fstp      st
                inc       esi
                fild      RemoveData.this_max_sample
                jmp       @@A3

                {$IFDEF ALIGN08}
                nop;nop;nop;nop;nop;nop;
                {$ENDIF}

      @@A2:     ficom     RemoveData.this_min_sample
                fstsw     ax
                xor       ah,$01
                and       ah,$41
                jnz       @@A3

                fstp      st
                inc       esi
                fild      RemoveData.this_min_sample

                {$IFDEF ALIGN08}
                nop;nop;nop;nop;nop;nop;
                {$ENDIF}

      @@A3:     fist      dword ptr BlockData.BTRD[ebx]   // store bit-removed-sample
                fmul      settings.scaling_factor_inv
                fisubr    dword ptr [ebp+ebx]
                fistp     dword ptr BlockData.CORR[ebx]   // store correction-value

                add       ebx,INTEGER_SIZE
                cmp       ebx,edi
                jl        @@A1
                //==============================================================
                // End of Remove Bits
                //==============================================================

      @@F0:     mov       remove_bits_data.clips,esi

                pop       esi
                pop       edi
                pop       ebp
                pop       ebx
  //============================================================================
  end;
  //============================================================================
  {$ENDIF}

  //============================================================================
  Procedure remove_bits();
  //============================================================================
  begin
  //============================================================================
    remove_bits_data.bits_removed:=remove_bits_data.bits_to_remove;

    repeat
      if (parameters.shaping<>-1) or ((parameters.adaptive_shaping<>-1) and (Filter_Valid)) then
        remove_bits_proc_adaptive_noise_shaping_on
      else
        remove_bits_proc_shaping_off;

      if (remove_bits_data.clips<=settings.maximum_clips_per_channel) then
        exit
      else
        remove_bits_data.bits_removed:=remove_bits_data.bits_removed-1;
    until (remove_bits_data.bits_removed<0) or (remove_bits_data.clips<=settings.maximum_clips_per_channel);

    if remove_bits_data.bits_removed<0 then
      remove_bits_data.bits_removed:=0;
  //============================================================================
  end;
  //============================================================================

  //============================================================================
  Procedure nRemoveBits_Init(); //bitdepth and samplerate dependent.
  //============================================================================
  var
    fbi_i : Integer;
  begin
  //============================================================================
    RemoveData.this_max_sample:=PowersOf.twom1[Global.Bits_Per_Sample-1];
    RemoveData.this_min_sample:=-PowersOf.twoint[Global.Bits_Per_Sample-1];

    for fbi_i:=0 to BITS_TO_CALCULATE-1 do
      sample_maxima[fbi_i]:=(RemoveData.this_max_sample xor powersof.twom1[fbi_i]);

    settings.static_maximum_bits_to_remove:=max(0,Global.Bits_Per_Sample-Settings.static_minimum_bits_to_keep);
  //============================================================================
  end;
  //============================================================================

//==============================================================================
initialization
//==============================================================================
  Unit_Data_Packer2.Data[0]:=0;
  {$IFDEF DISPLAY_VAR_DATA}
  write_unitname_and_datapointers(UNITNAME,@Unit_Data_Packer1,@Unit_Data_Packer2,sizeof(Unit_Data_Packer2));
  {$ENDIF}
//==============================================================================
finalization
//==============================================================================
  {$IFDEF WRITE_FINAL}
  Write(Erroutput,UNITNAME,CRLF);
  {$ENDIF}
//==============================================================================
end.
//==============================================================================

