//==============================================================================
//
//    lossyWAV: Added noise WAV bit reduction method by David Robinson;
//              Noise shaping coefficients by Sebastian Gesemann;
//
//    Copyright (C) 2007-2011 Nick Currie, Copyleft.
//
//    This program is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program.  If not, see <http://www.gnu.org/licenses/>.
//
//    Contact: lossywav <at> hotmail <dot> co <dot> uk
//
//==============================================================================
//
//    Adaptive Noise Shaping method provided by Sebastian Gesemann.
//
//    Copyright (C) 2008 by Sebastian Gesemann <s.gesemann at gmail>
//
//==============================================================================
unit nSGNS;
//==============================================================================
interface
//==============================================================================
Uses
  nCore;
//==============================================================================
var
  Unit_Data_Packer1 : Packed Record
    Data : Array[0..3] of Integer;
  End;
//==============================================================================

Procedure Restore_Filter;
Procedure Make_Filter;
Function  Filter_Valid : LongBool;
Function  Warped_Lattice_Filter_Evaluate : Extended;
{$IFNDEF USEASM}
Procedure Warped_Lattice_Filter_Update(fqe : Double);
{$ELSE}
Procedure Warped_Lattice_Filter_Update_Int;
{$ENDIF}
Function  Skipped_Filters : Integer;
Procedure nSGNS_Initialise;

//==============================================================================
implementation
//==============================================================================
uses
  fftw_interface,
  {$IFNDEF USEASM}
  nFFT_pas,
  {$ELSE}
  nFFT_asm,
  {$ENDIF}
  nMaths,
  math;
//==============================================================================
Const
  UNITNAME            = 'nSgns';
  //============================================================================
  // Maximum order of FIR filter for Adaptive Noise Shaping
  //============================================================================
  MAX_FILTER_ORDER    = 96;

  tFFT_LENGTH         = MAX_FFT_LENGTH_HALF + 2;
  tFFT_Size : Integer = tFFT_LENGTH*DOUBLE_SIZE;

  FNS_Freq : Array[0..41] of Single = (     0,    20,    30,    40,    50,    60,    70,    80,    90,   100,  200,  300,  400,  500,  600,  700,  800,  900,  1000, 1500, 2000, 2500, 3000, 3700, 4000, 5000, 6000,  7000,  8000,  9000, 10000, 11000, 12000, 13000, 14000, 15000, 16000, 17000, 18000, 19000, 20000,384000);
  FNS_Gain : Array[0..41] of Single = (12.850,12.783,12.750,12.717,12.684,12.651,12.618,12.585,12.552,12.519,12.191,11.865,11.537,11.205, 10.864, 10.511, 10.144, 9.767, 9.383, 7.419, 5.397, 2.873, 0.854, 0.513, 1.642, 5.645, 8.692, 10.972, 12.739, 14.156, 15.433, 16.764, 18.385, 20.904, 24.786, 30, 35.674, 39.51, 40.519, 40.418, 40.401,40.401);
//==============================================================================
type
  tFFT_Array_Double   = array[0..tFFT_LENGTH-1] of Double;
  tFFT_Array_Integer  = array[0..tFFT_LENGTH-1] of Integer;
  tFLT_Array          = array[0..MAX_FILTER_ORDER+1] of Double;

  Filter_Rec = Packed Record
    k                 : tFLT_Array;
    State             : tFLT_Array;
    xState            : tFLT_Array;
    e                 : Double;
    One_Over_Minus_C  : Double;
    Valid             : LongBool;
    Packer            : Array[0..(64-(INTEGER(tFFT_Length)*DOUBLE_SIZE*3+DOUBLE_SIZE+DOUBLE_SIZE+INTEGER_SIZE) and 63) div INTEGER_SIZE-1] of Integer;
  End;
//==============================================================================
Const
  Filter_SIZE : Integer = Sizeof(Filter_Rec);
//==============================================================================
Var
  Filter                      : Filter_Rec;
  Filter_Store                : Array[0..MAX_CHANNELS] of Filter_Rec;

  LD : Record
    a : tFLT_Array;
    b : tFLT_Array;
    r : tFLT_Array;
  End;

  SGNS : Record
    Shape_Curve               : tFFT_Array_Double;
    Shape_Total               : tFFT_Array_Double;

    Negative_Freq_Index       : tFFT_Array_Integer;
    Length_DIV                : tFFT_Array_Integer;

    Window                    : tFFT_Array_Double;
    Freq                      : tFFT_Array_Double;
    Warp                      : tFFT_Array_Double;
    Warp_FFT_Length           : tFFT_Array_Double;
    Warp_Frac                 : tFFT_Array_Double;
    Warp_1_M_Frac             : tFFT_Array_Double;
    Warp_Correction           : tFFT_Array_Double;
    Gain                      : tFFT_Array_Double;
    FNS_Gain                  : tFFT_Array_Double;
    Average_Ratio             : tFFT_Array_Double;

    Length_Factor             : tFFT_Array_Double;
    Length_1_M_Factor         : tFFT_Array_Double;

    Length_MOD_Factor         : tFFT_Array_Double;

    Shape_Channel_Offset      : Array[0..MAX_CHANNELS] of Integer;
    Filter_Offset             : Array[0..MAX_CHANNELS] of Integer;

    Lambda                    : Double;
    OneMinusLambda2           : Double;
    Gain_Amplitude            : Double;
    Shaping_Factor            : Double;

    Warp_Int                  : tFFT_Array_Integer;

    Filter_Order              : Integer;
    Filter_Order_M1           : Integer;

    FFTW_Plan_Inverse         : Pointer;
    Skipped_Filters           : Integer;
    Limit_Freq_Bin            : Integer;
    Upper_Freq_Bin            : Integer;
    Use_Average               : Longbool;
  End;

  {$A4}
  Unit_Data_Packer2 : Record
    Data : Array[0..2] of Integer;
  End;
  {$A8}
//==============================================================================

{$IFNDEF USEASM}
//==============================================================================
Procedure Save_Filter;
//==============================================================================
Begin
//==============================================================================
  Move(Filter,Filter_Store[Global.Channel],Filter_SIZE);
//==============================================================================
End;
//==============================================================================
{$ELSE}
//==============================================================================
Procedure Save_Filter; Assembler;
//==============================================================================
Asm
//==============================================================================
                push      eax
                push      ebx
                push      edx

                mov       ebx,Filter_SIZE
                mov       edx,Global.Channel
                mov       edx,dword ptr SGNS.Filter_Offset[edx*POINTER_SIZE]
                xor       eax,eax

                {$IFDEF ALIGN08}
                nop;nop;nop;nop;
                  {$IFDEF ALIGN16}
                  nop;nop;nop;nop;nop;nop;nop;nop;
                    {$IFDEF ALIGN32}
                    nop;nop;nop;nop;nop;nop;nop;nop;nop;nop;nop;nop;nop;nop;nop;nop;
                    {$ENDIF}
                  {$ENDIF}
                {$ENDIF}

      @@01:     // Partially unrolled Loop, Iteration n+0 ======================
                fild      qword ptr Filter[eax]
                fistp     qword ptr Filter_Store[edx+eax]
                fild      qword ptr Filter[eax]+8
                fistp     qword ptr Filter_Store[edx+eax]+8
                fild      qword ptr Filter[eax]+16
                fistp     qword ptr Filter_Store[edx+eax]+16
                fild      qword ptr Filter[eax]+24
                fistp     qword ptr Filter_Store[edx+eax]+24
                fild      qword ptr Filter[eax]+32
                fistp     qword ptr Filter_Store[edx+eax]+32
                fild      qword ptr Filter[eax]+40
                fistp     qword ptr Filter_Store[edx+eax]+40
                fild      qword ptr Filter[eax]+48
                fistp     qword ptr Filter_Store[edx+eax]+48
                fild      qword ptr Filter[eax]+56
                fistp     qword ptr Filter_Store[edx+eax]+56
                add       eax,64
                //==============================================================

                cmp       eax,ebx
                jl        @@01

                pop       edx
                pop       ebx
                pop       eax
                //mov       global.jump_location,offset @@01
//==============================================================================
End;
//==============================================================================
{$ENDIF}

{$IFNDEF USEASM}
//==============================================================================
Procedure Restore_Filter;
//==============================================================================
Begin
//==============================================================================
  Move(Filter_Store[Global.Channel],Filter,Filter_SIZE);
//==============================================================================
End;
//==============================================================================
{$ELSE}
//==============================================================================
Procedure Restore_Filter; Assembler;
//==============================================================================
Asm
//==============================================================================
                push      eax
                push      ebx
                push      edx

                mov       ebx,Filter_SIZE
                mov       edx,Global.Channel
                mov       edx,dword ptr SGNS.Filter_Offset[edx*POINTER_SIZE]
                xor       eax,eax

                {$IFDEF ALIGN08}
                nop;nop;nop;nop;
                  {$IFDEF ALIGN16}
                    {$IFDEF ALIGN32}
                    nop;nop;nop;nop;nop;nop;nop;nop;nop;nop;nop;nop;nop;nop;nop;nop;
                    {$ENDIF}
                  {$ENDIF}
                {$ENDIF}

      @@01:     // Partially unrolled Loop, Iteration n+0 ======================
                fild      qword ptr Filter_Store[edx+eax]
                fistp     qword ptr Filter[eax]
                fild      qword ptr Filter_Store[edx+eax]+8
                fistp     qword ptr Filter[eax]+8
                fild      qword ptr Filter_Store[edx+eax]+16
                fistp     qword ptr Filter[eax]+16
                fild      qword ptr Filter_Store[edx+eax]+24
                fistp     qword ptr Filter[eax]+24
                fild      qword ptr Filter_Store[edx+eax]+32
                fistp     qword ptr Filter[eax]+32
                fild      qword ptr Filter_Store[edx+eax]+40
                fistp     qword ptr Filter[eax]+40
                fild      qword ptr Filter_Store[edx+eax]+48
                fistp     qword ptr Filter[eax]+48
                fild      qword ptr Filter_Store[edx+eax]+56
                fistp     qword ptr Filter[eax]+56
                add       eax,64
                //==============================================================

                cmp       eax,ebx
                jl        @@01

                pop       edx
                pop       ebx
                pop       eax
                //mov       global.jump_location,offset @@01
//==============================================================================
End;
//==============================================================================
{$ENDIF}

{$IFNDEF USEASM}
//==============================================================================
function ArcTan_Complex(const X: tDComplex): Extended;
//==============================================================================
asm
//==============================================================================
                FLD       X.im
                FLD       X.re
                FPATAN
//==============================================================================
End;
//==============================================================================

//==============================================================================
function warped_frequency(lambda : double; freq : double) : Extended;
//==============================================================================
var
  x, y, z : tDComplex;
  //t : Double;
Begin
//==============================================================================
  //z:=complexexp(f);
  z.re:=cos(freq);
  z.im:=sin(freq);
  //x:=z-complex(lambda,0);
  x.re:=z.re+lambda;
  x.im:=z.im;
  //y:=complexStateubtract(complex(1,0),complexmultiply(z,complex(lambda,0)));
  y.re:=1+z.re*lambda;
  y.im:=+z.im*lambda;

  //============================================================================
  //z:=complexdivide(x,y);
  //============================================================================
  // No need to complete full complex division as both elements divided by t
  //============================================================================
  //  t:=1/(y.re*y.re+y.im*y.im);
  z.re:=(x.re*y.re+x.im*y.im);//*t;
  z.im:=(x.im*y.re-x.re*y.im);//*t;

  //warped_frequency:=complexarg(z));
  warped_frequency:=ArcTan_Complex(Z)
//==============================================================================
End;
//==============================================================================
{$ELSE}
//==============================================================================
function warped_frequency(lambda : double; freq : double) : Extended; Assembler;
//==============================================================================
var
  x, y : tDComplex;
asm
//==============================================================================
                fld     freq
                fsincos

                fld     st
                fadd    Lambda
                fstp    x.re
                fmul    Lambda
                fld1
                fadd
                fstp    y.re

                fst     x.im
                fmul    Lambda
                fstp    y.im

                fld     x.im
                fmul    y.re
                fld     x.re
                fmul    y.im
                fsub

                fld     x.re
                fmul    y.re
                fld     x.im
                fmul    y.im
                fadd

                fpatan
                fstp    @result
//==============================================================================
End;
//==============================================================================
{$ENDIF}

{$IFNDEF USEASM}
//==============================================================================
Procedure Levinson;
//==============================================================================
var
  j, m    : Integer;
  err, km : Extended;
Begin
//==============================================================================
  LD.a[0]:=1;
  Filter.e:=LD.r[0];

  m:=0;
  while (Filter.e<>0) and (m<SGNS.Filter_Order) do
  Begin
    m:=m+1;
    err:=LD.r[m];

    j:=1;
    while j<m do
    begin
      LD.b[m-j]:=LD.a[j];
      err:=err+LD.a[j]*LD.r[m-j];
      j:=j+1;
    end;

    km:=-err/Filter.e;

    j:=1;
    while j<m do
    begin
      LD.a[j]:=LD.a[j]+km*LD.b[j];
      j:=j+1;
    end;

    LD.a[m]:=km;
    Filter.k[m-1]:=km;
    Filter.e:=Filter.e*(1-km*km);
  End;

  Filter.Valid:=(m=SGNS.Filter_Order);
//==============================================================================
End;
//==============================================================================
{$ELSE}
//==============================================================================
Procedure Levinson; Assembler;
//==============================================================================
asm
                push      eax
                push      ebx
                push      ecx
                push      edx

                mov       ebx,SGNS.Filter_Order

                fld1
                fstp      qword ptr LD.a[0]
                fld       qword ptr LD.r[0]

                {$IFDEF ALIGN08}
                  nop;nop;
                  {$IFDEF ALIGN16}
                    {$IFDEF ALIGN32}
                    {$ENDIF}
                  {$ENDIF}
                {$ENDIF}

  //============================================================================
  // Start of Loop
  //============================================================================
                xor       ecx,ecx
  //============================================================================
  @@01:         ftst
                fstsw     ax
                and       ah,$41
                jnz       @@06

                add       ecx,1

                xor       eax,eax
                fld       qword ptr LD.r[ecx*double_size]
                jmp       @@03

                {$IFDEF ALIGN08}
                nop;nop;nop;nop;
                  {$IFDEF ALIGN16}
                    {$IFDEF ALIGN32}
                    {$ENDIF}
                  {$ENDIF}
                {$ENDIF}

  @@02:         mov       edx,ecx
                sub       edx,eax
                fld       qword ptr LD.a[eax*double_size]
                fst       qword ptr LD.b[edx*double_size]
                fmul      qword ptr LD.r[edx*double_size]
                fadd

                {$IFDEF ALIGN08}
                nop;nop;nop;nop;nop;
                  {$IFDEF ALIGN16}
                    {$IFDEF ALIGN32}
                    {$ENDIF}
                  {$ENDIF}
                {$ENDIF}

  @@03:         add       eax,1
                cmp       eax,ecx
                jl        @@02

                fchs
                fld       st(1)
                fdiv

                fst       qword ptr LD.a[ecx*double_size]
                fst       qword ptr Filter.k[ecx*double_size]-DOUBLE_SIZE

                xor       eax,eax
                jmp       @@05

                {$IFDEF ALIGN08}
                nop;
                  {$IFDEF ALIGN16}
                    {$IFDEF ALIGN32}
                    {$ENDIF}
                  {$ENDIF}
                {$ENDIF}

  @@04:         fld       qword ptr LD.a[eax*double_size]
                fld       qword ptr LD.b[eax*double_size]
                fmul      st,st(2)
                fadd
                fstp      qword ptr LD.a[eax*double_size]

                {$IFDEF ALIGN08}
                nop;nop;nop;nop;nop;nop;nop;
                  {$IFDEF ALIGN16}
                    {$IFDEF ALIGN32}
                    {$ENDIF}
                  {$ENDIF}
                {$ENDIF}

  @@05:         add       eax,1
                cmp       eax,ecx
                jl        @@04

                fmul      st,st
                fld1
                fsubr
                fmul

                cmp       ecx,ebx
                jl        @@01
  //============================================================================
  // End of loop
  //============================================================================

                {$IFDEF ALIGN08}
                nop;
                  {$IFDEF ALIGN16}
                  nop;nop;nop;nop;nop;nop;nop;nop;
                    {$IFDEF ALIGN32}
                    {$ENDIF}
                  {$ENDIF}
                {$ENDIF}

  @@06:         xor       eax,eax
                cmp       ecx,ebx
                setne     al
                dec       eax
                mov       Filter.Valid,eax

                fstp      Filter.E

                pop       edx
                pop       ecx
                pop       ebx
                pop       eax
                //mov       global.jump_location,offset @@06
//==============================================================================
End;
//==============================================================================
{$ENDIF}

{$IFNDEF USEASM}
//==============================================================================
Procedure Warped_Lattice_Filter_Init;
//==============================================================================
var
  a, o, u : double;
  i : integer;
Begin
//==============================================================================
  o:=1;
  u:=1;

  for i:=0 to SGNS.Filter_Order_M1 do
  Begin
    Filter.State[i]:=0;
    Filter.xState[i]:=u;
    u:=-SGNS.lambda*u;
    a:=u+Filter.k[i]*o;
    o:=o+Filter.k[i]*u;
    u:=a;
  End;

  Filter.One_Over_Minus_C:=-1/o;
//==============================================================================
End;
//==============================================================================
{$ELSE}
//==============================================================================
procedure Warped_Lattice_Filter_Init; Assembler;
//==============================================================================
asm
//==============================================================================
                push      eax
                push      edx

                xor       eax,eax
                mov       edx,SGNS.Filter_Order
                shl       edx,DOUBLE_SHIFT

                fld1                                          //o
                fld       st                                  //o;u

                {$IFDEF ALIGN08}
                  nop;nop;nop;nop;nop;nop;nop;
                  {$IFDEF ALIGN16}
                    {$IFDEF ALIGN32}
                    nop;nop;nop;nop;nop;nop;nop;nop;nop;nop;nop;nop;nop;nop;nop;nop;
                    {$ENDIF}
                  {$ENDIF}
                {$ENDIF}

    @@01:       // Partially unrolled Loop - Common element optimised===========
                fldz
                fst       qword ptr Filter.State[eax+0*DOUBLE_SIZE]
                fst       qword ptr Filter.State[eax+1*DOUBLE_SIZE]
                fst       qword ptr Filter.State[eax+2*DOUBLE_SIZE]
                fst       qword ptr Filter.State[eax+3*DOUBLE_SIZE]
                fst       qword ptr Filter.State[eax+4*DOUBLE_SIZE]
                fst       qword ptr Filter.State[eax+5*DOUBLE_SIZE]
                fst       qword ptr Filter.State[eax+6*DOUBLE_SIZE]
                fstp      qword ptr Filter.State[eax+7*DOUBLE_SIZE]

                // Partially unrolled Loop, Iteration n+0 ======================
                fst       qword ptr Filter.xState[eax]//o;u
                fmul      SGNS.Lambda                         //o;u
                fchs                                          //o;u

                fld       qword ptr Filter.k[eax]     //o;u;Filter.k

                fld       st                                  //;o;u;Filter.k;Filter.k
                fmul      st,st(2)                            //o;u;Filter.k;Filter.k*u
                fadd      st,st(3)                            //o;u;Filter.k;Filter.k*u+o
                fxch      st(3)                               //o;u;Filter.k;o
                fmul                                          //o;u;Filter.k*o
                fadd                                          //o;u
                add       eax,DOUBLE_SIZE

                // Partially unrolled Loop, Iteration n+1 ======================
                fst       qword ptr Filter.xState[eax]//o;u
                fmul      SGNS.Lambda                         //o;u
                fchs                                          //o;u

                fld       qword ptr Filter.k[eax]     //o;u;Filter.k

                fld       st                                  //;o;u;Filter.k;Filter.k
                fmul      st,st(2)                            //o;u;Filter.k;Filter.k*u
                fadd      st,st(3)                            //o;u;Filter.k;Filter.k*u+o
                fxch      st(3)                               //o;u;Filter.k;o
                fmul                                          //o;u;Filter.k*o
                fadd                                          //o;u
                add       eax,DOUBLE_SIZE

                // Partially unrolled Loop, Iteration n+2 ======================
                fst       qword ptr Filter.xState[eax]//o;u
                fmul      SGNS.Lambda                         //o;u
                fchs                                          //o;u

                fld       qword ptr Filter.k[eax]     //o;u;Filter.k

                fld       st                                  //;o;u;Filter.k;Filter.k
                fmul      st,st(2)                            //o;u;Filter.k;Filter.k*u
                fadd      st,st(3)                            //o;u;Filter.k;Filter.k*u+o
                fxch      st(3)                               //o;u;Filter.k;o
                fmul                                          //o;u;Filter.k*o
                fadd                                          //o;u
                add       eax,DOUBLE_SIZE

                // Partially unrolled Loop, Iteration n+3 ======================
                fst       qword ptr Filter.xState[eax]//o;u
                fmul      SGNS.Lambda                         //o;u
                fchs                                          //o;u

                fld       qword ptr Filter.k[eax]     //o;u;Filter.k

                fld       st                                  //;o;u;Filter.k;Filter.k
                fmul      st,st(2)                            //o;u;Filter.k;Filter.k*u
                fadd      st,st(3)                            //o;u;Filter.k;Filter.k*u+o
                fxch      st(3)                               //o;u;Filter.k;o
                fmul                                          //o;u;Filter.k*o
                fadd                                          //o;u
                add       eax,DOUBLE_SIZE

                // Partially unrolled Loop, Iteration n+4 ======================
                fst       qword ptr Filter.xState[eax]//o;u
                fmul      SGNS.Lambda                         //o;u
                fchs                                          //o;u

                fld       qword ptr Filter.k[eax]     //o;u;Filter.k

                fld       st                                  //;o;u;Filter.k;Filter.k
                fmul      st,st(2)                            //o;u;Filter.k;Filter.k*u
                fadd      st,st(3)                            //o;u;Filter.k;Filter.k*u+o
                fxch      st(3)                               //o;u;Filter.k;o
                fmul                                          //o;u;Filter.k*o
                fadd                                          //o;u
                add       eax,DOUBLE_SIZE

                // Partially unrolled Loop, Iteration n+5 ======================
                fst       qword ptr Filter.xState[eax]//o;u
                fmul      SGNS.Lambda                         //o;u
                fchs                                          //o;u

                fld       qword ptr Filter.k[eax]     //o;u;Filter.k

                fld       st                                  //;o;u;Filter.k;Filter.k
                fmul      st,st(2)                            //o;u;Filter.k;Filter.k*u
                fadd      st,st(3)                            //o;u;Filter.k;Filter.k*u+o
                fxch      st(3)                               //o;u;Filter.k;o
                fmul                                          //o;u;Filter.k*o
                fadd                                          //o;u
                add       eax,DOUBLE_SIZE

                // Partially unrolled Loop, Iteration n+6 ======================
                fst       qword ptr Filter.xState[eax]//o;u
                fmul      SGNS.Lambda                         //o;u
                fchs                                          //o;u

                fld       qword ptr Filter.k[eax]     //o;u;Filter.k

                fld       st                                  //;o;u;Filter.k;Filter.k
                fmul      st,st(2)                            //o;u;Filter.k;Filter.k*u
                fadd      st,st(3)                            //o;u;Filter.k;Filter.k*u+o
                fxch      st(3)                               //o;u;Filter.k;o
                fmul                                          //o;u;Filter.k*o
                fadd                                          //o;u
                add       eax,DOUBLE_SIZE

                // Partially unrolled Loop, Iteration n+7 ======================
                fst       qword ptr Filter.xState[eax]//o;u
                fmul      SGNS.Lambda                         //o;u
                fchs                                          //o;u

                fld       qword ptr Filter.k[eax]     //o;u;Filter.k

                fld       st                                  //;o;u;Filter.k;Filter.k
                fmul      st,st(2)                            //o;u;Filter.k;Filter.k*u
                fadd      st,st(3)                            //o;u;Filter.k;Filter.k*u+o
                fxch      st(3)                               //o;u;Filter.k;o
                fmul                                          //o;u;Filter.k*o
                fadd                                          //o;u
                add       eax,DOUBLE_SIZE

                // Partially unrolled Loop, End ================================
                cmp       eax,edx
                jl        @@01
                //==============================================================

                fstp      st

                fld1
                fdivr
                fchs
                fstp      Filter.One_Over_Minus_C

                pop       edx
                pop       eax
                //mov       global.jump_location,offset @@01
End;
//==============================================================================
{$ENDIF}

{$IFNDEF USEASM}
//==============================================================================
Function Warped_Lattice_Filter_Evaluate : Extended;
//==============================================================================
var
  a     : Double;
  o     : Double;
  u     : double;
  i     : integer;
Begin
//==============================================================================
  o:=0;
  u:=0;

  for i:=0 to SGNS.Filter_Order_M1 do
  Begin
    a:=Filter.State[i];
    Filter.State[i]:=u+SGNS.lambda*a;
    u:=SGNS.OneMinusLambda2*a-SGNS.lambda*u;
    a:=o;
    o:=o+Filter.k[i]*u;
    u:=u+Filter.k[i]*a;
  End;

  Warped_Lattice_Filter_Evaluate:=o*Filter.One_Over_Minus_C;
//==============================================================================
End;
//==============================================================================
{$ELSE}
//==============================================================================
Function Warped_Lattice_Filter_Evaluate : Extended;
//==============================================================================
asm
//==============================================================================
                push      eax
                push      edx
                xor       eax,eax
                mov       edx,SGNS.Filter_Order
                shl       edx,DOUBLE_SHIFT

                {$IFDEF ALIGN08}
                nop;nop;nop;nop;nop;nop;nop;
                  {$IFDEF ALIGN16}
                  nop;nop;nop;nop;nop;nop;nop;nop;
                    {$IFDEF ALIGN32}
                    nop;nop;nop;nop;nop;nop;nop;nop;nop;nop;nop;nop;nop;nop;nop;nop;
                    {$ENDIF}
                  {$ENDIF}
                {$ENDIF}

                fld       SGNS.Lambda
                fld       SGNS.OneMinusLambda2

                fldz                                          //o
                fld       st                                  //o;u

  @@01:         // Partially unrolled Loop, Iteration n+0 ======================
                fld       qword ptr Filter.State[eax] //o;u/Filter.State
                fld       st                                  //o;u/Filter.State;Filter.State
                fmul      st,st(5)                            //o;u;Filter.State;Filter.State*Lambda
                fadd      st,st(2)                            //o;u;Filter.State;u+Filter.State*Lambda
                fstp      qword ptr Filter.State[eax] //o;u;Filter.State

                fmul      st,st(3)                            //o;u;Filter.State*OML2
                fld       st(1)                               //o;u;Filter.State*OML2;u
                fmul      st,st(5)                            //o;u;Filter.State*OML2;u*Lambda
                fsub                                          //o;u;Filter.State*OML2-u*Lambda
                fstp      st(1)                               //o;u

                fld       qword ptr Filter.k[eax]     //o;u;Filter.k
                fld       st                                  //o;u;Filter.k;Filter.k
                fmul      st,st(2)                            //o;u;Filter.k;Filter.k*u
                fadd      st,st(3)                            //o;u;Filter.k;Filter.k*u+o
                fxch      st(3)                               //o;u;Filter.k;o
                fmul                                          //o;u;o*Filter.k
                fadd                                          //o;u
                add       eax,DOUBLE_SIZE

                // Partially unrolled Loop, Iteration n+1 ======================
                fld       qword ptr Filter.State[eax] //o;u/Filter.State
                fld       st                                  //o;u/Filter.State;Filter.State
                fmul      st,st(5)                            //o;u;Filter.State;Filter.State*Lambda
                fadd      st,st(2)                            //o;u;Filter.State;u+Filter.State*Lambda
                fstp      qword ptr Filter.State[eax] //o;u;Filter.State

                fmul      st,st(3)                            //o;u;Filter.State*OML2
                fld       st(1)                               //o;u;Filter.State*OML2;u
                fmul      st,st(5)                            //o;u;Filter.State*OML2;u*Lambda
                fsub                                          //o;u;Filter.State*OML2-u*Lambda
                fstp      st(1)                               //o;u

                fld       qword ptr Filter.k[eax]     //o;u;Filter.k
                fld       st                                  //o;u;Filter.k;Filter.k
                fmul      st,st(2)                            //o;u;Filter.k;Filter.k*u
                fadd      st,st(3)                            //o;u;Filter.k;Filter.k*u+o
                fxch      st(3)                               //o;u;Filter.k;o
                fmul                                          //o;u;o*Filter.k
                fadd                                          //o;u
                add       eax,DOUBLE_SIZE

                // Partially unrolled Loop, Iteration n+2 ======================
                fld       qword ptr Filter.State[eax] //o;u/Filter.State
                fld       st                                  //o;u/Filter.State;Filter.State
                fmul      st,st(5)                            //o;u;Filter.State;Filter.State*Lambda
                fadd      st,st(2)                            //o;u;Filter.State;u+Filter.State*Lambda
                fstp      qword ptr Filter.State[eax] //o;u;Filter.State

                fmul      st,st(3)                            //o;u;Filter.State*OML2
                fld       st(1)                               //o;u;Filter.State*OML2;u
                fmul      st,st(5)                            //o;u;Filter.State*OML2;u*Lambda
                fsub                                          //o;u;Filter.State*OML2-u*Lambda
                fstp      st(1)                               //o;u

                fld       qword ptr Filter.k[eax]     //o;u;Filter.k
                fld       st                                  //o;u;Filter.k;Filter.k
                fmul      st,st(2)                            //o;u;Filter.k;Filter.k*u
                fadd      st,st(3)                            //o;u;Filter.k;Filter.k*u+o
                fxch      st(3)                               //o;u;Filter.k;o
                fmul                                          //o;u;o*Filter.k
                fadd                                          //o;u
                add       eax,DOUBLE_SIZE

                // Partially unrolled Loop, Iteration n+3 ======================
                fld       qword ptr Filter.State[eax] //o;u/Filter.State
                fld       st                                  //o;u/Filter.State;Filter.State
                fmul      st,st(5)                            //o;u;Filter.State;Filter.State*Lambda
                fadd      st,st(2)                            //o;u;Filter.State;u+Filter.State*Lambda
                fstp      qword ptr Filter.State[eax] //o;u;Filter.State

                fmul      st,st(3)                            //o;u;Filter.State*OML2
                fld       st(1)                               //o;u;Filter.State*OML2;u
                fmul      st,st(5)                            //o;u;Filter.State*OML2;u*Lambda
                fsub                                          //o;u;Filter.State*OML2-u*Lambda
                fstp      st(1)                               //o;u

                fld       qword ptr Filter.k[eax]     //o;u;Filter.k
                fld       st                                  //o;u;Filter.k;Filter.k
                fmul      st,st(2)                            //o;u;Filter.k;Filter.k*u
                fadd      st,st(3)                            //o;u;Filter.k;Filter.k*u+o
                fxch      st(3)                               //o;u;Filter.k;o
                fmul                                          //o;u;o*Filter.k
                fadd                                          //o;u
                add       eax,DOUBLE_SIZE

                // Partially unrolled Loop, Iteration n+4 ======================
                fld       qword ptr Filter.State[eax] //o;u/Filter.State
                fld       st                                  //o;u/Filter.State;Filter.State
                fmul      st,st(5)                            //o;u;Filter.State;Filter.State*Lambda
                fadd      st,st(2)                            //o;u;Filter.State;u+Filter.State*Lambda
                fstp      qword ptr Filter.State[eax] //o;u;Filter.State

                fmul      st,st(3)                            //o;u;Filter.State*OML2
                fld       st(1)                               //o;u;Filter.State*OML2;u
                fmul      st,st(5)                            //o;u;Filter.State*OML2;u*Lambda
                fsub                                          //o;u;Filter.State*OML2-u*Lambda
                fstp      st(1)                               //o;u

                fld       qword ptr Filter.k[eax]     //o;u;Filter.k
                fld       st                                  //o;u;Filter.k;Filter.k
                fmul      st,st(2)                            //o;u;Filter.k;Filter.k*u
                fadd      st,st(3)                            //o;u;Filter.k;Filter.k*u+o
                fxch      st(3)                               //o;u;Filter.k;o
                fmul                                          //o;u;o*Filter.k
                fadd                                          //o;u
                add       eax,DOUBLE_SIZE

                // Partially unrolled Loop, Iteration n+5 ======================
                fld       qword ptr Filter.State[eax] //o;u/Filter.State
                fld       st                                  //o;u/Filter.State;Filter.State
                fmul      st,st(5)                            //o;u;Filter.State;Filter.State*Lambda
                fadd      st,st(2)                            //o;u;Filter.State;u+Filter.State*Lambda
                fstp      qword ptr Filter.State[eax] //o;u;Filter.State

                fmul      st,st(3)                            //o;u;Filter.State*OML2
                fld       st(1)                               //o;u;Filter.State*OML2;u
                fmul      st,st(5)                            //o;u;Filter.State*OML2;u*Lambda
                fsub                                          //o;u;Filter.State*OML2-u*Lambda
                fstp      st(1)                               //o;u

                fld       qword ptr Filter.k[eax]     //o;u;Filter.k
                fld       st                                  //o;u;Filter.k;Filter.k
                fmul      st,st(2)                            //o;u;Filter.k;Filter.k*u
                fadd      st,st(3)                            //o;u;Filter.k;Filter.k*u+o
                fxch      st(3)                               //o;u;Filter.k;o
                fmul                                          //o;u;o*Filter.k
                fadd                                          //o;u
                add       eax,DOUBLE_SIZE

                // Partially unrolled Loop, Iteration n+6 ======================
                fld       qword ptr Filter.State[eax] //o;u/Filter.State
                fld       st                                  //o;u/Filter.State;Filter.State
                fmul      st,st(5)                            //o;u;Filter.State;Filter.State*Lambda
                fadd      st,st(2)                            //o;u;Filter.State;u+Filter.State*Lambda
                fstp      qword ptr Filter.State[eax] //o;u;Filter.State

                fmul      st,st(3)                            //o;u;Filter.State*OML2
                fld       st(1)                               //o;u;Filter.State*OML2;u
                fmul      st,st(5)                            //o;u;Filter.State*OML2;u*Lambda
                fsub                                          //o;u;Filter.State*OML2-u*Lambda
                fstp      st(1)                               //o;u

                fld       qword ptr Filter.k[eax]     //o;u;Filter.k
                fld       st                                  //o;u;Filter.k;Filter.k
                fmul      st,st(2)                            //o;u;Filter.k;Filter.k*u
                fadd      st,st(3)                            //o;u;Filter.k;Filter.k*u+o
                fxch      st(3)                               //o;u;Filter.k;o
                fmul                                          //o;u;o*Filter.k
                fadd                                          //o;u
                add       eax,DOUBLE_SIZE

                // Partially unrolled Loop, Iteration n+7 ======================
                fld       qword ptr Filter.State[eax] //o;u/Filter.State
                fld       st                                  //o;u/Filter.State;Filter.State
                fmul      st,st(5)                            //o;u;Filter.State;Filter.State*Lambda
                fadd      st,st(2)                            //o;u;Filter.State;u+Filter.State*Lambda
                fstp      qword ptr Filter.State[eax] //o;u;Filter.State

                fmul      st,st(3)                            //o;u;Filter.State*OML2
                fld       st(1)                               //o;u;Filter.State*OML2;u
                fmul      st,st(5)                            //o;u;Filter.State*OML2;u*Lambda
                fsub                                          //o;u;Filter.State*OML2-u*Lambda
                fstp      st(1)                               //o;u

                fld       qword ptr Filter.k[eax]     //o;u;Filter.k
                fld       st                                  //o;u;Filter.k;Filter.k
                fmul      st,st(2)                            //o;u;Filter.k;Filter.k*u
                fadd      st,st(3)                            //o;u;Filter.k;Filter.k*u+o
                fxch      st(3)                               //o;u;Filter.k;o
                fmul                                          //o;u;o*Filter.k
                fadd                                          //o;u
                add       eax,DOUBLE_SIZE

                // Partially unrolled Loop, End ================================
                cmp       eax,edx
                jl        @@01
                //==============================================================

                fstp      st
                fmul      Filter.One_Over_Minus_C
                fxch      st(2)
                fstp      st
                fstp      st

                pop       edx
                pop       eax

                //mov     global.jump_location,offset @@01
//==============================================================================
End;
//==============================================================================
{$ENDIF}

{$IFNDEF USEASM}
//==============================================================================
Procedure Warped_Lattice_Filter_Update(fqe : Double);
//==============================================================================
Var
  wu_i : Integer;
//==============================================================================
begin
//==============================================================================
  for wu_i:=0 to SGNS.Filter_Order_M1 do
    Filter.State[wu_i]:=(fqe*Filter.xState[wu_i]+Filter.State[wu_i]);
//==============================================================================
end;
//==============================================================================
{$ELSE}
//==============================================================================
Procedure Warped_Lattice_Filter_Update_Int;
//==============================================================================
Asm
//==============================================================================
                push      eax
                push      edx
                xor       eax,eax
                mov       edx,SGNS.Filter_Order
                shl       edx,DOUBLE_SHIFT

                {$IFDEF ALIGN08}
                nop;nop;nop;nop;nop;nop;nop;
                  {$IFDEF ALIGN16}
                  nop;nop;nop;nop;nop;nop;nop;nop;
                    {$IFDEF ALIGN32}
                    {$ENDIF}
                  {$ENDIF}
                {$ENDIF}

  @@01:         // Partially unrolled Loop, Iteration n+0 ======================
                fld       qword ptr Filter.xState[eax]
                fmul      st,st(1)
                fadd      qword ptr Filter.State[eax]
                fstp      qword ptr Filter.State[eax]

                fld       qword ptr Filter.xState[eax]+DOUBLE_SIZE
                fmul      st,st(1)
                fadd      qword ptr Filter.State[eax]+DOUBLE_SIZE
                fstp      qword ptr Filter.State[eax]+DOUBLE_SIZE

                fld       qword ptr Filter.xState[eax]+(DOUBLE_SIZE*2)
                fmul      st,st(1)
                fadd      qword ptr Filter.State[eax]+(DOUBLE_SIZE*2)
                fstp      qword ptr Filter.State[eax]+(DOUBLE_SIZE*2)

                fld       qword ptr Filter.xState[eax]+(DOUBLE_SIZE*3)
                fmul      st,st(1)
                fadd      qword ptr Filter.State[eax]+(DOUBLE_SIZE*3)
                fstp      qword ptr Filter.State[eax]+(DOUBLE_SIZE*3)

                fld       qword ptr Filter.xState[eax]+(DOUBLE_SIZE*4)
                fmul      st,st(1)
                fadd      qword ptr Filter.State[eax]+(DOUBLE_SIZE*4)
                fstp      qword ptr Filter.State[eax]+(DOUBLE_SIZE*4)

                fld       qword ptr Filter.xState[eax]+(DOUBLE_SIZE*5)
                fmul      st,st(1)
                fadd      qword ptr Filter.State[eax]+(DOUBLE_SIZE*5)
                fstp      qword ptr Filter.State[eax]+(DOUBLE_SIZE*5)

                fld       qword ptr Filter.xState[eax]+(DOUBLE_SIZE*6)
                fmul      st,st(1)
                fadd      qword ptr Filter.State[eax]+(DOUBLE_SIZE*6)
                fstp      qword ptr Filter.State[eax]+(DOUBLE_SIZE*6)

                fld       qword ptr Filter.xState[eax]+(DOUBLE_SIZE*7)
                fmul      st,st(1)
                fadd      qword ptr Filter.State[eax]+(DOUBLE_SIZE*7)
                fstp      qword ptr Filter.State[eax]+(DOUBLE_SIZE*7)

                add       eax,(DOUBLE_SIZE*8)

                // Partially unrolled Loop, End ================================
                cmp       eax,edx
                jl        @@01

                pop       edx
                fstp      st
                pop       eax

                //mov       global.jump_location,offset @@01
//==============================================================================
end;
//==============================================================================
{$ENDIF}

{$IFNDEF USEASM}
//==============================================================================
Procedure Process_Stored_Results;
//==============================================================================
var
  ts_i : Integer;
  ts_j : Integer;

  ts_x : Double;
  ts_y : Double;

  hi_fact : Double;
  lo_fact : Double;
//==============================================================================
Begin
  //==========================================================================
  // Fill desired shape curve.
  //==========================================================================
  ts_y:=0;

  for ts_i:=0 to SGNSFFT.FFT.Length_Half do
  Begin
    ts_j:=SGNS.Length_DIV[ts_i];
    hi_fact:=(ts_i and 15)*OneOver[16];
    lo_fact:=1-hi_fact;
    ts_x:=nRoot(((SGNSFFT.FFT_Results_Short[ts_j]*Lo_Fact+SGNSFFT.FFT_Results_Short[ts_j+1]*hi_fact)*SGNS.Length_Factor[ts_i])+
                (SGNSFFT.FFT_Results_Long[ts_i])*SGNS.Length_1_M_Factor[ts_i])*SGNS.Gain[ts_i];

    SGNS.Shape_Curve[ts_i]:=ts_x;
    ts_y:=ts_y+ts_x;
    SGNS.Shape_Total[ts_i]:=ts_y;
  End;

  if SGNS.Use_Average then
  Begin
    //==========================================================================
    // Calculate average up to Limit_Freq_Bin
    //==========================================================================
    ts_y:=SGNS.Shape_Total[SGNS.Upper_Freq_Bin]*OneOver[SGNS.Upper_Freq_Bin+1];

    For ts_i:=SGNS.Upper_Freq_Bin+1 to SGNSFFT.FFT.Length_Half do
    Begin
      ts_x:=SGNS.Shape_Curve[ts_i];
      if ts_x<ts_y then
        SGNS.Shape_Curve[ts_i]:=ts_x+(ts_y-ts_x)*SGNS.Average_Ratio[ts_i];
      SGNS.Shape_Total[ts_i]:=SGNS.Shape_Total[ts_i-1]+SGNS.Shape_Curve[ts_i];
    End;
  End;
//==============================================================================
End;
//==============================================================================
{$ELSE}
//==============================================================================
Procedure Process_Stored_Results;
//==============================================================================
Asm
//==============================================================================
              push      eax
              push      ecx
              push      edx
              push      esi

  //==========================================================================
  // Fill desired shape curve.
  //==========================================================================
              fldz
              xor       edx,edx
              mov       ecx,SGNSFFT.FFT.Length_Half

              {$IFDEF ALIGN08}
              nop;nop;
                {$IFDEF ALIGN16}
                  {$IFDEF ALIGN32}
                  {$ENDIF}
                {$ENDIF}
              {$ENDIF}

    @@01:     mov       esi,dword ptr SGNS.Length_DIV[edx*INTEGER_SIZE]

              fld       qword ptr SGNSFFT.FFT_Results_Short[esi*DOUBLE_SIZE]
              fld       qword ptr SGNSFFT.FFT_Results_Short[esi*DOUBLE_SIZE]+DOUBLE_SIZE
              fsub      st,st(1)
              fmul      qword ptr SGNS.Length_MOD_Factor[edx*DOUBLE_SIZE]
              fadd
              fmul      qword ptr SGNS.Length_Factor[edx*DOUBLE_SIZE]

              fld       qword ptr SGNSFFT.FFT_Results_Long[edx*DOUBLE_SIZE]
              fmul      qword ptr SGNS.Length_1_M_Factor[edx*DOUBLE_SIZE]

              fadd

              call      nroot_int
              fmul      qword ptr SGNS.Gain[edx*DOUBLE_SIZE]

              fst       qword ptr SGNS.Shape_Curve[edx*DOUBLE_SIZE]
              fadd
              fst       qword ptr SGNS.Shape_Total[edx*DOUBLE_SIZE]

              add       edx,1
              cmp       edx,ecx
              jng       @@01

              fstp      st

    //==========================================================================
    // Calculate average up to Limit_Freq_Bin
    //==========================================================================
              test      SGNS.Use_Average,1
              jz        @@Exit

              mov       edx,SGNS.Upper_Freq_Bin
              fld       qword ptr SGNS.Shape_Total[edx*DOUBLE_SIZE]
              inc       edx
              fmul      qword ptr OneOver[edx*DOUBLE_SIZE]

              {$IFDEF ALIGN08}
              nop;nop;nop;
              {$ENDIF}

      @@03:   fld       qword ptr SGNS.Shape_Curve[edx*DOUBLE_SIZE]

              fcom      st(1)
              fstsw     ax
              and       ah,$01
              jz        @@04

              fld       st(1)
              fsub      st,st(1)
              fmul      qword ptr SGNS.Average_Ratio[edx*DOUBLE_SIZE]
              fadd

              {$IFDEF ALIGN08}
              nop;nop;
              {$ENDIF}

      @@04:   fst       qword ptr SGNS.Shape_Curve[edx*DOUBLE_SIZE]
              fadd      qword ptr SGNS.Shape_Total[edx*DOUBLE_SIZE]-DOUBLE_SIZE
              fstp      qword ptr SGNS.Shape_Total[edx*DOUBLE_SIZE]

              add       edx,1
              cmp       edx,ecx
              jng       @@03

              fstp      st

      @@Exit: pop       esi
              pop       edx
              pop       ecx
              pop       eax
              //mov       global.jump_location,offset @@01
//==============================================================================
End;
//==============================================================================
{$ENDIF}

{$IFNDEF USEASM}
//==============================================================================
Procedure Fill_FFT_With_Warped_Spectrum;
//==============================================================================
var
  ts_i : Integer;
  ts_j : Integer;
  ts_y : Extended;
  ts_z : Extended;
  ts_x : Extended;
Begin
//==============================================================================
  ts_x:=0;

  for ts_i:=0 to SGNSFFT.FFT.Length_Half do
  Begin
    ts_j:=SGNS.Warp_Int[ts_i];
    ts_z:=SGNS.Warp_1_M_Frac[ts_i]*SGNS.Shape_Total[ts_j]+SGNS.Warp_Frac[ts_i]*SGNS.Shape_Total[ts_j+1];
    ts_y:=(ts_z-ts_x)*SGNS.Warp_Correction[ts_i];

    if ((ts_i and SGNSFFT.FFT.Length_Half_M1)>0) then
      With FFT_Array.DComplex[SGNSFFT.FFT.Length-ts_i] do
      begin
        re:=ts_y;
        im:=0;
      end;

    With FFT_Array.DComplex[ts_i] do
    begin
      re:=ts_y;
      im:=0;
    end;

    ts_x:=ts_z;
  End;
//==============================================================================
End;
//==============================================================================
{$ELSE}
//==============================================================================
Procedure Fill_FFT_With_Warped_Spectrum;
//==============================================================================
Asm
//==============================================================================
              push      eax
              push      ebx
              push      ecx
              push      edx
              push      esi

              mov       eax,1
              mov       edx,SGNSFFT.FFT.Length_Half

              xor       ecx,ecx
              fldz
              mov       ebx,SGNSFFT.FFT.Length_Half_M1

              {$IFDEF ALIGN08}
              nop;nop;
                {$IFDEF ALIGN16}
                  {$IFDEF ALIGN32}
                  {$ENDIF}
                {$ENDIF}
              {$ENDIF}

    @@02:     mov       esi,dword ptr SGNS.Warp_Int[ecx*INTEGER_SIZE]

              //================================================================
              // Calculate linear interpolated value.
              //================================================================
              fld       qword ptr SGNS.Shape_Total[esi*DOUBLE_SIZE]
              fmul      qword ptr SGNS.Warp_1_M_Frac[ecx*DOUBLE_SIZE]

              fld       qword ptr SGNS.Shape_Total[esi*DOUBLE_SIZE+DOUBLE_SIZE]
              fmul      qword ptr SGNS.Warp_Frac[ecx*DOUBLE_SIZE]
              fadd
              //================================================================
              fsub      st,st(1)
              fadd      st(1),st
              //================================================================
              // Correct desired value by relative bin width ratio.
              //================================================================
              fmul      qword ptr SGNS.Warp_Correction[ecx*DOUBLE_SIZE]
              //================================================================

              mov       esi,ecx
              and       esi,ebx
              jz        @@03

              mov       esi,dword ptr SGNS.Negative_Freq_Index[ecx*Integer_Size]
              fst       qword ptr FFT_Array.DComplex[esi*DOUBLE_SIZE]
              fldz
              fstp      qword ptr FFT_Array.DComplex[esi*DOUBLE_SIZE]+DOUBLE_SIZE

              {$IFDEF ALIGN08}
              nop;nop;nop;
                {$IFDEF ALIGN16}
                  {$IFDEF ALIGN32}
                  {$ENDIF}
                {$ENDIF}
              {$ENDIF}

    @@03:     mov       esi,ecx
              add       esi,esi
              fstp      qword ptr FFT_Array.DComplex[esi*DOUBLE_SIZE]
              fldz
              fstp      qword ptr FFT_Array.DComplex[esi*DOUBLE_SIZE]+DOUBLE_SIZE

              add       ecx,1
              cmp       ecx,edx
              jle       @@02

              fstp      st

              pop       esi
              pop       edx
              pop       ecx
              pop       ebx
              pop       eax
              //mov       global.jump_location,offset @@02
//==============================================================================
End;
//==============================================================================
{$ENDIF}

//==============================================================================
Procedure Fill_LD_r_with_FFT_Results;
//==============================================================================
Var
  ts_i : integer;
Begin
//==============================================================================
  for ts_i:=0 to SGNS.Filter_Order do
    LD.r[ts_i]:=FFT_Array.DComplex[ts_i].re;
//==============================================================================
End;
//==============================================================================

//==============================================================================
Procedure Make_Filter();
//==============================================================================
Begin
//==============================================================================
  if (Parameters.Adaptive_Shaping<>-1) and (parameters.Adaptive_Shaping<>-3) then
  Begin
    Process_Stored_Results;

    Fill_FFT_With_Warped_Spectrum;

    If FFTW_Initialised then
      FFTW.Execute(SGNS.FFTW_Plan_Inverse)
    else
    Begin
      SetFFTLength(SGNSFFT.FFT.Length);
      IFFT_DIT_Complex;
    End;

    Fill_LD_r_with_FFT_Results;

    Levinson;

    if Filter.Valid then
      Warped_Lattice_Filter_Init
    else
      move(Filter_Store[MAX_CHANNELS],Filter,Filter_Size);
  End
  else
    move(Filter_Store[MAX_CHANNELS],Filter,Filter_Size);

  Save_Filter;
//==============================================================================
End;
//==============================================================================

//==============================================================================
Function Filter_Valid : LongBool;
//==============================================================================
Begin
//==============================================================================
  Filter_Valid:=Filter_Store[Global.Channel].Valid;
//==============================================================================
End;
//==============================================================================

//==============================================================================
Function Skipped_Filters : Integer;
//==============================================================================
Begin
//==============================================================================
  Skipped_Filters:=SGNS.Skipped_Filters;
//==============================================================================
End;
//==============================================================================

//==============================================================================
Procedure nSGNS_Initialise;
//==============================================================================
Var
  si_i, si_j, si_k    : Integer;
  this_lambda         : Double;
  this_freq           : Double;
  this_target_bin     : Double;
  this_delta          : Double;
  this_freq_limit     : Double;
  this_average_ratio  : Double;
  this_gain           : Double;
  this_warp           : Double;
  this_total          : Double;
Begin
//==============================================================================

  //============================================================================
  // Set pointer to relevant FFTW plan (if available).
  //============================================================================
  If FFTW_Initialised then
    SGNS.FFTW_Plan_Inverse:=FFTW.plan_dft_1d(SGNSFFT.FFT.Length,@FFT_Array,@FFT_Array,FFTW_BACKWARD,$69);
  //============================================================================

  //============================================================================
  // Use high frequency average tweak?
  //============================================================================
  SGNS.Use_Average:=true;
  //============================================================================

  //============================================================================
  // Set shaping factor for fixed noise shaping.
  //============================================================================
  SGNS.Shaping_Factor:=Parameters.Shaping;

  if SGNS.Shaping_Factor=-1 then
    SGNS.Shaping_Factor:=1;
  //============================================================================

  //============================================================================
  // Zero number of skipped filters.
  //============================================================================
  SGNS.Skipped_Filters:=0;
  //============================================================================

  //============================================================================
  // Set amplitude of default gain curve.
  //============================================================================
  SGNS.Gain_Amplitude:=4.5;
  //============================================================================

  //============================================================================
  // Initialise filters and create channel offset tables.
  //============================================================================
  FillChar(Filter,sizeof(Filter),0);
  FillChar(Filter_Store,sizeof(Filter_Store),0);

  for si_i:=0 to max_channels-1 do
  begin
    Filter_Store[si_i].One_Over_Minus_C:=1;
    SGNS.Shape_Channel_Offset[si_i]:=si_i*integer(TFFT_SIZE);
    SGNS.Filter_Offset[si_i]:=si_i*Filter_Size;
  end;
  SGNS.Filter_Offset[MAX_CHANNELS]:=MAX_CHANNELS*Filter_Size;
  Filter.One_Over_Minus_C:=1;
  //============================================================================

  //============================================================================
  // Set filter order
  //============================================================================
  if parameters.adaptive_shaping<-1 then
    SGNS.Filter_Order:=64
  else
    SGNS.Filter_Order:=parameters.adaptive_shaping;

  SGNS.Filter_Order:=min(MAX_FILTER_ORDER,min(SGNS.Filter_Order,SGNSFFT.FFT.Length-16));
  SGNS.Filter_Order_M1:=Max(0,SGNS.Filter_Order-1);
  //============================================================================

  //============================================================================
  // Set bin range points for "Average Ratio" curve.
  //============================================================================
  SGNS.Upper_Freq_Bin:=round(SGNSFFT.FFT.Length*Min(Global.Upper_Freq_Limit/Global.Sample_Rate,46*OneOver[96]));
  SGNS.Limit_Freq_Bin:=round(SGNSFFT.FFT.Length*Min((HUMAN_FREQ_LIMIT+2000*Global.Codec_Block_Bit_Shift)/Global.Sample_Rate,46*OneOver[96]));
  //============================================================================

  //============================================================================
  // Define "average ratio" curve.
  //============================================================================
  this_average_ratio:=(32+(6*Global.Codec_Block_Bit_Shift))*OneOver[96];

  si_j:=0;
  si_k:=SGNS.Upper_Freq_Bin;
  For si_i:=si_j to si_k do
    SGNS.Average_Ratio[si_i]:=0;

  si_j:=si_k;
  si_k:=SGNS.Limit_Freq_Bin;
  For si_i:=si_j to si_k do
    SGNS.Average_Ratio[si_i]:=Sqr((si_i-si_j)/(Max(si_k-si_j,1)))*this_average_ratio;

  si_j:=si_k;
  si_k:=(SGNSFFT.FFT.Length_Half-(SGNSFFT.FFT.Length_Half-SGNS.Limit_Freq_Bin) div Max(2,Global.Codec_Block_Bit_Shift+2));
  For si_i:=si_j to si_k do
    SGNS.Average_Ratio[si_i]:=this_average_ratio;

  si_j:=si_k;
  si_k:=SGNSFFT.FFT.Length_Half;
  For si_i:=si_j to si_k do
    SGNS.Average_Ratio[si_i]:=Sqr((si_k-si_i)/(Max(si_k-si_j,1)))*this_average_ratio;
  //============================================================================

  //============================================================================
  // Calculate lambda value to use in frequency warping.
  //============================================================================
  if Parameters.Adaptive_Shaping_Warp then
  Begin
    this_delta:=SINGLE_HALF;
    this_freq_limit:=min(HUMAN_FREQ_LIMIT,Global.Sample_Rate*46*OneOver[96]);
    this_lambda:=SINGLE_ONE;
    this_target_bin:=46*OneOver[96];

    repeat
      this_freq:=Warped_Frequency(this_lambda,this_target_bin*twopi)*oneovertwopi*Global.Sample_Rate;

      if this_freq<this_freq_limit then
        this_lambda:=this_lambda-this_delta;

      If this_freq>this_freq_limit then
      Begin
        this_delta:=this_delta*single_half;
        this_lambda:=this_lambda+this_delta;
      End;
    until (this_delta<1e-12) or (abs(this_freq-Global.Upper_Freq_Limit)<1e-6) or (this_lambda>1) or (this_lambda<0);
  End
  Else
    this_lambda:=0;

  SGNS.Lambda:=Min(1,Max(0,this_lambda));
  SGNS.OneMinusLambda2:=1-Sqr(SGNS.Lambda);
  //============================================================================

  //============================================================================
  // Clear stored spectral shapes
  //============================================================================
  for si_i:=0 to SGNSFFT.FFT.Length_Half do
  Begin
    SGNSFFT.FFT_Results_Long[si_i]:=0;
    SGNSFFT.FFT_Results_Short[si_i]:=0;
  End;
  //============================================================================

  //============================================================================
  // Calculate lambda value to use in frequency warping.
  //============================================================================
  this_warp:=0;
  for si_i:=0 to SGNSFFT.FFT.Length_Half do
  Begin
    SGNS.Window[si_i]:=Hann(si_i,SGNSFFT.FFT.Bit_Length);
    SGNS.Freq[si_i]:=si_i*SGNSFFT.FFT.Length_Recip;

    SGNS.Negative_Freq_Index[si_i]:=(SGNSFFT.FFT.Length-si_i)*2;

    SGNS.Warp[si_i]:=Warped_Frequency(SGNS.Lambda,SGNS.Freq[si_i]*twopi)*oneovertwopi;

    SGNS.Warp_FFT_Length[si_i]:=SGNS.Warp[si_i]*SGNSFFT.FFT.Length;
    SGNS.Warp_Int[si_i]:=Floor(SGNS.Warp_FFT_Length[si_i]);
    SGNS.Warp_Frac[si_i]:=SGNS.Warp_FFT_Length[si_i]-SGNS.Warp_Int[si_i];
    SGNS.Warp_1_M_Frac[si_i]:=1-SGNS.Warp_Frac[si_i];

    SGNS.Length_Factor[si_i]:=Power(si_i*SGNSFFT.FFT.Length_Half_Recip,1.5)*SINGLE_HALF;
    SGNS.Length_1_M_Factor[si_i]:=SINGLE_HALF-SGNS.Length_Factor[si_i];

    SGNS.Length_Factor[si_i]:=SGNS.Length_Factor[si_i]*PowersOf.Two[4];

    SGNS.Length_DIV[si_i]:=si_i shr 4;
    SGNS.Length_MOD_Factor[si_i]:=(si_i and 15)*OneOver[16];

    if SGNS.Warp[si_i]-this_warp<>0 then
      SGNS.Warp_Correction[si_i]:=1/((SGNS.Warp[si_i]-this_warp)*SGNSFFT.FFT.Length)
    else
      SGNS.Warp_Correction[si_i]:=1;
    this_warp:=SGNS.Warp[si_i];
  End;
  //============================================================================

  //============================================================================
  // Calculate unwarped gain curve to use with adaptive shaping method.
  //============================================================================
  for si_i:=0 to SGNSFFT.FFT.Length_Half do
  Begin
    this_freq:=SGNS.Freq[si_i]*Global.Sample_Rate;
    case floor(this_freq) of
          0..16000  : this_gain:=0;
      16001..19999  : this_gain:=sqr((this_freq-16000)*OneOver[4000]);
    else
      This_Gain:=1;
    end;
    SGNS.Gain[si_i]:=nPower2(this_gain*oneover[10]*Log2_10*SGNS.Gain_Amplitude);
  End;
  //============================================================================

  //============================================================================
  // Calculate Fixed Noise Shaping curve
  //============================================================================
  this_total:=0;
  for si_i:=0 to SGNSFFT.FFT.Length_Half do
  Begin
    this_freq:=SGNS.Freq[si_i]*Global.Sample_Rate;
    si_j:=0;
    while FNS_Freq[si_j+1]<this_freq do
      inc(si_j);

    this_gain:=(this_freq-FNS_Freq[si_j])/(FNS_Freq[si_j+1]-FNS_Freq[si_j])*(FNS_Gain[si_j+1]-FNS_Gain[si_j])+FNS_Gain[si_j];
    SGNS.FNS_Gain[si_i]:=nPower2(this_gain*SGNS.Shaping_Factor*oneover[10]*Log2_10);

    this_total:=this_total+SGNS.FNS_Gain[si_i];
    SGNS.Shape_Total[si_i]:=this_total;
  End;

  //============================================================================
  // Use the fixed shaping filter when selected (i.e. skip "make_filter") and
  // for cases where calculated adaptive shaping filter invalid.
  // (skip Process_Stored_Results stage as desired curve in SGNS.Shape_Total)
  //============================================================================
  Fill_FFT_with_Warped_Spectrum;

  If FFTW_Initialised then
    FFTW.Execute(SGNS.FFTW_Plan_Inverse)
  else
  Begin
    SetFFTLength(SGNSFFT.FFT.Length);
    IFFT_DIT_Complex;
  End;

  Fill_LD_r_with_FFT_Results;

  Levinson;

  Warped_Lattice_Filter_Init;

  for si_i:=0 to MAX_CHANNELS do
  Begin
    Global.Channel:=si_i;
    Save_Filter;
  End;
  Global.Channel:=0;

//==============================================================================
End;
//==============================================================================

//==============================================================================
initialization
//==============================================================================
  Unit_Data_Packer2.Data[0]:=0;
  {$IFDEF DISPLAY_VAR_DATA}
  write_unitname_and_datapointers(UNITNAME,@Unit_Data_Packer1,@Unit_Data_Packer2,sizeof(Unit_Data_Packer2));
  {$ENDIF}
//==============================================================================
finalization
//==============================================================================
  {$IFDEF WRITE_FINAL}
  Write(Erroutput,UNITNAME,CRLF);
  {$ENDIF}
   if (FFTW_Initialised) and (SGNS.FFTW_Plan_Inverse<>Nil) then
     FFTW.Destroy_Plan(SGNS.FFTW_Plan_Inverse);
//==============================================================================
end.
//==============================================================================

