//==============================================================================
//
//    lossyWAV: Added noise WAV bit reduction method by David Robinson;
//              Noise shaping coefficients by Sebastian Gesemann;
//
//    Copyright (C) 2007-2011 Nick Currie, Copyleft.
//
//    This program is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program.  If not, see <http://www.gnu.org/licenses/>.
//
//    Contact: lossywav <at> hotmail <dot> co <dot> uk
//
//==============================================================================
unit nSpreading;
//==============================================================================
interface
//==============================================================================
Uses
  nCore;
//==============================================================================
var
  Unit_Data_Packer1 : Packed Record
    Data : Array[0..3] of Integer;
  End;
//==============================================================================
Const
  SKEWING_AMPLITUDE             =    36;
  SPREADING_REC_SIZE            =     4;
  SPREADING_STEPS               = 8190;
  SPREADING_STEPS_RECIP         =  1/SPREADING_STEPS;

  THRESHOLD_INDEX_SPREAD        =     64;
  THRESHOLD_INDEX_SPREAD_MUL    =    256;
  THRESHOLD_INDEX_SPREAD_MAX    =    THRESHOLD_INDEX_SPREAD_MUL*THRESHOLD_INDEX_SPREAD;

  //============================================================================
  // Frequency range over which to check for lowest amplitude signal
  //============================================================================
  SPREAD_ZONES                  = 7;
  FREQUENCY_LIMITS              : Array [0..SPREAD_ZONES+1] of Double = (20,1378.125,3445.3125,5512.5,8268.75,10335.9375,12403.125,14470.3125,16000);

  SPREADING_FUNCTION_ARRAY      : Array[1..PRECALC_ANALYSES,0..SPREAD_ZONES] of byte = ((1,2,2,2,2,2,2,2),(1,2,2,2,2,2,2,2),(1,2,2,2,2,2,2,3),
                                                                                        (1,2,2,2,2,2,2,3),(1,2,2,2,2,2,2,4),(1,1,2,2,2,2,3,4));

  SPREADING_FUNCTION_WIDTHS     : array[1..PRECALC_ANALYSES] of double = (0.25,0.17677669529663688110021109052621,0.250,0.35355339059327376220042218105242,0.5,0.70710678118654752440084436210485);
//==============================================================================

//==============================================================================
Type
  Spreading_rec = packed record
    width    : word;      // integer number of bins to average (used in old spread method);
    fractint : word;      // linked with Spreading.widths array to create fraction in range 0..1 (used in new spreading method);
  end;

  sprec_array = array [0..0] of Spreading_rec;
  sprec_array_ptr = ^sprec_array;

//==============================================================================
Var
//==============================================================================

  Spreading : Record
    widths                      : array[0..SPREADING_STEPS+1] of single;
    divisors                    : array[0..SPREADING_STEPS+1] of single;

    alt_average                 : Double;
    old_minimum                 : Double;
    new_minimum                 : Double;

    averages_ptr                : array[1..PRECALC_ANALYSES] of sprec_array_ptr;

    //============================================================================
    // For each FFT bit length store bits_to_remove with respect to calculated minimum dB of FFT result.
    //============================================================================
    threshold_index             : array[0..THRESHOLD_INDEX_SPREAD_MAX-1] of Byte;

    //============================================================================
    // FFT Bin associated with relevant frequencies for each FFT length.
    //============================================================================
    Frequency_Bins              : array[1..PRECALC_ANALYSES,0..SPREAD_ZONES+1] of Integer;

    //============================================================================
    // FFT bins for defining calculation limits.Spreading.
    Bins : Record
      lo        : array[0..MAX_FFT_BIT_LENGTH] of integer;
      hi        : array[0..MAX_FFT_BIT_LENGTH] of integer;
      mid       : Array[0..MAX_FFT_BIT_LENGTH] of integer;
      Recip     : array[0..MAX_FFT_BIT_LENGTH] of double;
    End;
  End;

  //============================================================================
  // skewing function low frequency bin bin attenuation lookup table.
  //============================================================================
  Skewing_Gain  : array[0..(MAX_FFT_LENGTH div 2)+1] of Double;

//==============================================================================

Procedure nSpreading_Init;
Procedure Spreading_Function;

//==============================================================================
implementation
//==============================================================================
uses
//==============================================================================
  nMaths,
  math,
  sysutils;
//==============================================================================
const
  UNITNAME='nSpreading';
var
//==============================================================================

  Unit_Data_Packer2 : array[0..2] of Integer;

//==============================================================================

  {$IFNDEF USEASM}
  //============================================================================
  Procedure Spreading_Function();
  //============================================================================
  var
    sc_i, sc_j, sc_w, sc_f : Integer;
    new_value, old_value, alt_average : Double;
  begin
    Spreading.new_minimum:=db_maxdb;
    Spreading.old_minimum:=db_maxdb;
    alt_average:=0;

    for sc_i:=Spreading.Bins.Lo[Global.FFT.bit_length] to Spreading.Bins.Hi[Global.FFT.bit_length] do
    begin
      {$R-}
      With Spreading.Averages_Ptr[Global.analysis_number]^[sc_i] do
      Begin
        sc_w:=Width;
        sc_f:=FractInt;
      End;
      {$R+}

      alt_average:=alt_average+fft_skew_result[sc_i];

      old_value:=0;
      for sc_j:=0 to sc_w-1 do
        old_value:=old_value+fft_skew_result[sc_i-sc_j];
      old_value:=old_value*OneOver[sc_w];

      new_value:=((fft_skew_result[sc_i-1]+fft_skew_result[sc_i+1])*Spreading.widths[sc_f]+fft_skew_result[sc_i])*Spreading.divisors[sc_f];

      if Spreading.old_minimum>old_value then
      Begin
        Spreading.Old_minimum:=old_value;
        Process.Old_min_bin:=sc_i;
      End;

      if Spreading.new_minimum>new_value then
      Begin
        Spreading.new_minimum:=new_value;
        Process.new_min_bin:=sc_i;
      End;
    end;

    Spreading.alt_average:=(nlog2(alt_average*Spreading.Bins.Recip[Global.FFT.bit_length])+Global.FFT.Threshold_Shift)*db_lg2x20+Settings.Noise_Threshold_Shift_Average;
    Spreading.old_minimum:=(nlog2(Spreading.old_minimum)+Global.FFT.Threshold_Shift)*db_lg2x20+Settings.Noise_Threshold_Shift_Minimum;
    Spreading.new_minimum:=(nlog2(Spreading.new_minimum)+Global.FFT.Threshold_Shift)*db_lg2x20+Settings.Noise_Threshold_Shift_Minimum;
  //============================================================================
  end;
  //============================================================================
  {$ELSE}
  //============================================================================
  Procedure Spreading_Function(); Assembler;
  asm
  //============================================================================
                push    ebx
                push    ebp
                push    edi
                push    esi

                fld     db_maxdB  // min_value
                fst     Spreading.new_minimum
                fstp    Spreading.old_minimum

                fldz    // Alt-Simple Average

                mov     eax,Global.FFT.bit_length
                mov     edi,dword ptr Spreading.Bins.Lo[eax*INTEGER_SIZE]
                mov     esi,dword ptr Spreading.Bins.Hi[eax*INTEGER_SIZE]
                mov     ebp,Global.analysis_number
                mov     ebp,dword ptr Spreading.Averages_Ptr[ebp*pointer_size-pointer_size]

                {$IFDEF ALIGN16}
                nop;nop;nop;nop;nop;nop;nop;nop;
                  {$IFDEF ALIGN32}
                  nop;nop;nop;nop;nop;nop;nop;nop;nop;nop;nop;nop;nop;nop;nop;nop;
                  {$ENDIF}
                {$ENDIF}

      //========================================================================
      // Existing spreading method.
      //========================================================================
      @@J0:       movzx   eax,word ptr [ebp+edi*SPREADING_REC_SIZE]           // load required number of bins into eax
                  mov     edx,edi
                  mov     ecx,eax
                  fldz

                  {$IFDEF ALIGN16}
                  nop;nop;nop;nop;nop;
                    {$IFDEF ALIGN32}
                    nop;nop;nop;nop;nop;nop;nop;nop;nop;nop;nop;nop;nop;nop;nop;nop;
                    {$ENDIF}
                  {$ENDIF}
      //========================================================================
      @@J1:         fadd    qword ptr fft_skew_result[edx*DOUBLE_SIZE]             // add required number of bin values to centre bin
                    dec     edx
                    sub     ecx,1
                    jg      @@J1
      //========================================================================
                  fmul    qword ptr OneOver[eax*DOUBLE_SIZE]                 // divide by number of bins to obtain single bin spread result for old method
      //========================================================================
                  fcom    Spreading.old_minimum
                  fstsw   ax
                  and     ah,$01
                  jz      @@J3

                  fst     Spreading.old_minimum
                  mov     Process.old_min_bin,edi

                  {$IFDEF ALIGN08}
                  nop;nop;
                    {$IFDEF ALIGN32}
                    nop;nop;nop;nop;nop;nop;nop;nop;nop;nop;nop;nop;nop;nop;nop;nop;
                    {$ENDIF}
                  {$ENDIF}

      @@J3:       fstp    st
      //========================================================================
      // Incorporate new spreading function within old spreading function....
      //========================================================================
                  fld     qword ptr fft_skew_result[edi*DOUBLE_SIZE-DOUBLE_SIZE]  // load adjacent bin (lower)
                  movzx   eax,word ptr [ebp+edi*SPREADING_REC_SIZE+WORD_SIZE]     // load fraction "pointer" into eax
                  fadd    qword ptr fft_skew_result[edi*DOUBLE_SIZE+DOUBLE_SIZE]  // add adjacent bin (higher)
                  fmul    dword ptr Spreading.widths[eax*SINGLE_SIZE]             // multiply by "width" (variable, 0..1)
                  fadd    qword ptr fft_skew_result[edi*DOUBLE_SIZE]              // add centre bin
                  fmul    dword ptr Spreading.divisors[eax*SINGLE_SIZE]           // multiply by 1/(1+2*width). Final result for new method.
      //========================================================================

      //========================================================================
      // Compare both results and select lower.
      //========================================================================
                  fcom    Spreading.new_minimum
                  fstsw   ax
                  and     ah,$01
                  jz      @@J5

                  fst     Spreading.new_minimum
                  mov     Process.new_min_bin,edi

                  {$IFDEF ALIGN08}
                  nop;nop;nop;nop;
                    {$IFDEF ALIGN16}
                    nop;nop;nop;nop;nop;nop;nop;nop;
                      {$IFDEF ALIGN32}
                      nop;nop;nop;nop;nop;nop;nop;nop;nop;nop;nop;nop;nop;nop;nop;nop;
                      {$ENDIF}
                    {$ENDIF}
                  {$ENDIF}

     @@J5:        fstp    st
      //========================================================================
      // end of integration of new method within old method.
      //========================================================================

      //========================================================================
      // Prepare simple average
      //========================================================================
                  fld     qword ptr fft_skew_result[edi*DOUBLE_SIZE]               // centre bin
                  fadd
      //========================================================================

                  add     edi,1
                  cmp     edi,esi
                  jle     @@J0

      //========================================================================
      // Process results.
      //========================================================================
                mov     eax,Global.FFT.bit_length

                fmul    qword ptr Spreading.Bins.Recip[eax*DOUBLE_SIZE]                 // multiply running total by 1/(number of bins in running total)
                call    nlog2_int
                fadd    Global.FFT.Threshold_Shift
                fmul    db_lg2x20
                fadd    Settings.Noise_Threshold_Shift_Average
                fstp    Spreading.alt_average

                fld     Spreading.old_minimum
                call    nlog2_int
                fadd    Global.FFT.Threshold_Shift
                fmul    db_lg2x20
                fadd    Settings.Noise_Threshold_Shift_Minimum
                fstp    Spreading.old_minimum

                fld     Spreading.new_minimum
                call    nlog2_int
                fadd    Global.FFT.Threshold_Shift
                fmul    db_lg2x20
                fadd    Settings.Noise_Threshold_Shift_Minimum
                fstp    Spreading.new_minimum

  //============================================================================
                pop     esi
                pop     edi
                pop     ebp
                pop     ebx

                //mov     Global.Jump_Location,offset @@J5
  //============================================================================
  end;
  //============================================================================
  {$ENDIF}

  //============================================================================
  Procedure Spreading_Memory_Allocate();
  //============================================================================
  Var
    sa_i : Integer;
  Begin
  //============================================================================
    for sa_i:=1 to PRECALC_ANALYSES do
      getmem(Spreading.Averages_Ptr[sa_i],Global.Analysis.Length[sa_i]*2+4);
  //============================================================================
  end;
  //============================================================================

  //============================================================================
  Procedure Skewing_Curve_Init(); // Samplerate dependent!!!!
  //============================================================================
  Var
    sa_lfb, sa_mfb, sa_tfb, sa_amp : Double;
    sa_i : Integer;
  Begin
  //============================================================================
    sa_lfb:=log10_2*nlog2(LOWER_FREQ_LIMIT); //   20Hz lower limit for skewing;
    sa_mfb:=log10_2*nlog2(MID_FREQ_LIMIT);   // 3700Hz upper limit for skewing;

    if Parameters.dccorrect then
      sa_amp:=SKEWING_AMPLITUDE*SINGLE_HALF
    else
      sa_amp:=SKEWING_AMPLITUDE;

    for sa_i:=0 to Powersof.twom1[max_fft_bit_length-1]+1 do
    begin
      sa_tfb:=log10_2*nlog2((sa_i*PowersOf.two[-max_fft_bit_length]*Global.sample_rate));

      if sa_tfb<=sa_mfb then
        Skewing_Gain[sa_i]:=npower2((power(sin(HalfPi*max(0,sa_tfb-sa_lfb)/(sa_mfb-sa_lfb)),0.75)-1)*sa_amp*OneOver[20]*Log2_10)
      else
        Skewing_Gain[sa_i]:=1;
    end;
    Skewing_Gain[0]:=0;

  //============================================================================
  end;
  //============================================================================

  //============================================================================
  Procedure Spreading_Widths_Init();
  //============================================================================
  Var
    sa_i : Integer;
  Begin
  //============================================================================
    for sa_i:=0 to SPREADING_STEPS do
    begin
      Spreading.widths[sa_i]:=sa_i*SPREADING_STEPS_RECIP;
      Spreading.divisors[sa_i]:=1/(1+2*Spreading.widths[sa_i]);
    end;
  //============================================================================
  End;
  //============================================================================

  //============================================================================
  Procedure Threshold_Index_Init();
  //============================================================================
  var
    last_filled : Integer;
    this_bit_to_remove : Integer;
    this_reference_threshold : Double;
  Begin
    // Calculate Threshold_Index values using Reference_Threshold Values.
    last_filled:=0;
    for this_bit_to_remove:=1 to BITS_TO_CALCULATE do
    begin
      this_reference_threshold:=-2+this_bit_to_remove+Global.Codec_Block_Bits*single_half+
                                log2(1+PowersOf.Two[1-this_bit_to_remove*2])*single_half+
                                log2(1+PowersOf.Two[-1-Global.Codec_Block_Bits])*0.30769230769230769230769230769231+
                                -0.20751874963942190927313052802609+              //log2(0.75)*0.5+
                                -0.0013444987323232079693278710922535;            //log2((4194304-3907)/4194304);

      this_reference_threshold:=this_reference_threshold*db_lg2x20*THRESHOLD_INDEX_SPREAD;
      while(last_filled<this_reference_threshold) do
      begin
        Spreading.threshold_index[last_filled]:=this_bit_to_remove-1;
        inc(last_filled);
      end;
    end;

    while (last_filled<THRESHOLD_INDEX_SPREAD_MAX) do
    begin
      Spreading.threshold_index[last_filled]:=BITS_TO_CALCULATE-1;
      inc(last_filled);
    end;
  //============================================================================
  End;
  //============================================================================

  //============================================================================
  Procedure Spreading_Memory_Free;
  //============================================================================
  Var
    sa_i : Integer;
  begin
  //============================================================================
    for sa_i:=1 to precalc_analyses do
      freemem(Spreading.Averages_Ptr[sa_i]);
  //============================================================================
  end;
  //============================================================================

  //============================================================================
  Procedure nSpreading_Init;
  //============================================================================
  var
    this_analysis_number  : Integer;
    sa_i, sa_j            : Integer;
  Begin
  //============================================================================
    Spreading_Memory_Allocate;
    Spreading_Widths_Init;
    Skewing_Curve_Init;
    Threshold_Index_Init();

    //==========================================================================
    // calculating for each analysis_number
    //==========================================================================
    for sa_i:=1 to MAX_FFT_BIT_LENGTH do
    Begin
      // Calculate which FFT bins correspond to the respective frequency limits
      Spreading.Bins.Lo[sa_i]:=max(1,round(Global.Lower_Freq_Limit/Global.sample_rate*PowersOf.Two[sa_i])-1);
      Spreading.Bins.Hi[sa_i]:=max(1,round(Global.Upper_Freq_Limit/Global.sample_rate*PowersOf.Two[sa_i])-1);
      Spreading.Bins.Recip[sa_i]:=OneOver[Spreading.Bins.Hi[sa_i]-Spreading.Bins.Lo[sa_i]+1];
      Spreading.Bins.Mid[sa_i]:=max(1,round(MID_FREQ_LIMIT/Global.Sample_Rate*PowersOf.Two[sa_i])-1);
    End;

    for this_analysis_number:=1 to PRECALC_ANALYSES do
    begin
      Global.analysis_number:=this_analysis_number;
      Set_FFT_Data_Rec(Global.FFT,History.FFT.Bit_Length-2+Global.analysis_number);

      // Calculate which FFT bin corresponds to the low frequency limit
      for sa_i:=0 to SPREAD_ZONES+1 do
      begin
        Spreading.frequency_bins[Global.analysis_number,sa_i]:=min(Global.FFT.length div 2-1,max(1,round(FREQUENCY_LIMITS[sa_i]/Global.Sample_Rate*Global.FFT.Length)));
        if Spreading.frequency_bins[Global.analysis_number,sa_i]>Global.FFT.length div 2-1 then
          lossyWAVerror('frequency too high : '+floattostrf(FREQUENCY_LIMITS[sa_i]*OneOver[1000],ffFixed,99,3)+'kHz',$21);
      end;

      for sa_j:=SPREAD_ZONES downto 0 do
        for sa_i:=Spreading.frequency_bins[Global.analysis_number,sa_j] to Spreading.frequency_bins[Global.analysis_number,sa_j+1] do
        {$R-}
          spreading.averages_ptr[Global.analysis_number]^[sa_i].width:=min(sa_i,SPREADING_FUNCTION_ARRAY[Global.analysis_number,sa_j]);
        {$R+}

      for sa_i:=0 to Spreading.Bins.Mid[Global.FFT.Bit_Length] do
      {$R-}
        spreading.averages_ptr[Global.analysis_number]^[sa_i].fractint:=0;
      {$R+}

      for sa_i:=Spreading.Bins.Mid[Global.FFT.bit_length] to Global.FFT.Length_Half_M1 do
      {$R-}
        spreading.averages_ptr[Global.analysis_number]^[sa_i].fractint:=floor(SPREADING_STEPS*max(0,min((UPPER_FREQ_LIMIT-MID_FREQ_LIMIT),sa_i*Global.Sample_Rate/Global.FFT.Length-MID_FREQ_LIMIT))/(UPPER_FREQ_LIMIT-MID_FREQ_LIMIT)*spreading_function_widths[Global.analysis_number]);
      {$R+}
    //==========================================================================
    end; // calculating for each analysis_number
    //==========================================================================

  //============================================================================
  End;
  //============================================================================

//==============================================================================
Initialization
//==============================================================================
  Unit_Data_Packer2[0]:=0;
  {$IFDEF DISPLAY_VAR_DATA}
  write_unitname_and_datapointers(UNITNAME,@Unit_Data_Packer1,@Unit_Data_Packer2,sizeof(Unit_Data_Packer2));
  {$ENDIF}
//==============================================================================

//==============================================================================
finalization
//==============================================================================
  {$IFDEF WRITE_FINAL}
  Write(Erroutput,UNITNAME,CRLF);
  {$ENDIF}
  Spreading_Memory_Free;
//==============================================================================

//==============================================================================
end.
//==============================================================================

