//==============================================================================
//
//    lossyWAV: Added noise WAV bit reduction method by David Robinson;
//              Noise shaping coefficients by Sebastian Gesemann;
//
//    Copyright (C) 2007-2011 Nick Currie, Copyleft.
//
//    This program is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program.  If not, see <http://www.gnu.org/licenses/>.
//
//    Contact: lossywav <at> hotmail <dot> co <dot> uk
//
//==============================================================================
unit nWav;
//==============================================================================
interface
//==============================================================================
uses
  nCore;
//==============================================================================
var
  Unit_Data_Packer1 : Packed Record
    Data : Array[0..3] of Integer;
  End;
//==============================================================================
procedure MergeFiles;                         // Merge lossy.wav and lwcdf.wav files.

function openWavIO : Longbool;                // Zero if operation not succesful.
                                              // Reads the header chunks of inWav and memorizes the
                                              // Headerrmation from the fmt chunk.
                                              // Creates outWav and fills its header chunks with an
                                              // identical copy of that of inWav.
                                              // Memorizes blocksize (nr of samples to read/write with subsequent calls to
                                              // readNextSampleBlock resp. writeNextSampleBlock).

function readNextNextCodecBlock : longbool;   // False if not possible (usually because all the samples have been read).
                                              // sampleBlock: next block of samples from input wav file for all the channels.
                                              //              first sample is in element with index 0.
                                              // readsamplesCount: number of samples read (in each of the channels).

function writeNextbtrdcodecblock : longbool;  // False if not possible.
                                              // Writes sampleBlock as the next sample block of all the channels to the output file.
                                              // The number of samples to be written is given by samplesCountToBeWritten.
                                              // First sample written is sampleBlock element with index overlapsize
                                              // - overlapsize from openWavIO call.
                                              // Last sample written is in element with index samplesCountToBeWritten - 1.
                                              // samplesCountToBeWritten should be the readsamplesCount value from the
                                              // corresponding readNextSampleBlock function call.

function writeNextcorrcodecblock : longbool;  // False if not possible.
                                              // Writes sampleBlock as the next sample block of all the channels to the output file.
                                              // The number of samples to be written is given by samplesCountToBeWritten.
                                              // First sample written is sampleBlock element with index overlapsize
                                              // - overlapsize from openWavIO call.
                                              // Last sample written is in element with index samplesCountToBeWritten - 1.
                                              // samplesCountToBeWritten should be the readsamplesCount value from the
                                              // corresponding readNextSampleBlock function call.

function closeWavIO : longbool;               // False if not possible.
                                              // Finalizes the wavIO operations.

//==============================================================================
implementation
//==============================================================================
{$I-}
uses
  Math,
  SysUtils,
  Windows;
const
  CUEChunkMaxSize   = 65536;
  FACTChunkMaxSize  = 65536;
  FMTChunkMaxSize   = 18+22;

Type
  tChunkID = array[0..3] of Char;

  tChunkHeader = packed record
    Case integer of
     0 : (ID   : tChunkID;
          Size : Cardinal);
     1 : (CNum : Cardinal);
    End;

  tChunkMapRecord = Packed Record
    Header : tChunkHeader;
    CPtr   : Integer;
    Packer : Integer;
  End;

  tWAVHeader = Packed Record
    Header   : tChunkHeader; //"RIFF"
    RIFFType : tChunkID; //"WAVE"
    Packer   : Integer;
  end;

  tFMTChunk = Packed Record
    Header                    : tChunkHeader; //"fmt "
    wFormatTag                : Word;
    wChannels                 : Word;
    nSamplesPerSec            : Cardinal;
    nAvgBytesPerSec           : Cardinal;
    nBlockAlign               : Word;
    wBitsPerSample            : Word;
    cbSize                    : Word;
    Samples : Packed Record
    Case integer of
      0:(wValidBitsPerSample  : Word);
      1:(wSamplesPerBlock     : Word);
      2:(wReserved            : Word);
    End;
    dwChannelMask             : Cardinal;
    SubFormat_Word            : Array[0..3] of Word;
    SubFormat_Byte            : Array[0..7] of Byte;
  End;

  tFACTChunk = Packed Record
    Header : tChunkHeader; //"fact"
    Data : array[0..65535] of char;
  End;

  tDATAChunk = Packed Record
    Header : tChunkHeader; //"data"
  End;

  tlossyWAVChunk = Packed Record
   WAV  : tWAVHeader;
   FMT  : tFMTChunk;
   FACT : tFACTChunk;
   DATA : tDataChunk;
  End;

  tWaveBuffer = packed record
  Case integer of
    1: (w : array[0..(BUFFER_SIZE div 2)-1] of smallint);
    2: (t : array[0..(BUFFER_SIZE div 3)-1] of packed record lo_b : byte; hi_w : smallint; end;);
    3: (l : array[0..(BUFFER_SIZE div 4)-1] of integer);
  end;

  chunkspace_array                    = array[0..2097151] of byte;
//==============================================================================
//nwav constants, variables & types
//==============================================================================
const
  UNITNAME='nWav';
Var
  WAVBuff                             : tWaveBuffer;
  BtrdBuff                            : tWaveBuffer;
  Corrbuff                            : tWaveBuffer;
  MergeBuff                           : tWaveBuffer;

  chunkspace                          : ^chunkspace_array;
  chunkmap                            : array[0..255] of tChunkMapRecord;

  WAVEFile, Btrdfile, Corrfile        : File;

  datetimestring                      : string;

  ChunkHeader                         : tChunkHeader;
  WAVEChunk                           : tlossyWAVChunk;
  BTRDChunk                           : tlossyWAVChunk;
  CORRChunk                           : tlossyWAVChunk;

  filesOpened                         : Cardinal;
  readresult                          : Cardinal;
  writeresult                         : Cardinal;

  BUFFER_SIZEread                     : Cardinal; // number of byte that make up for inbuff
  BUFFER_SIZEwrite                    : Cardinal; // number of byte that make up for outbuff
  nrOfByteInOneBlockInBuff            : Cardinal; // number of byte that make up a complete block in inBuff and outBuff.
  nrOfBlocksInReadBuff                : Cardinal;
  nrOfBlocksInWriteBuff               : Cardinal;

  nrOfFullBlocksReadIntoInBuff        : Cardinal; // with the last BlockRead: how many full blocks managed it into inBuff ?
  nrOfByteOfTheLastIncompleteBlock    : Cardinal; // when EOF was encountered with the last BlockRead: the number of byte that didn't make it
                                                  // into the last full block
  nrOfBlockInBuffNotYetFetched        : Cardinal; // the next number of the block in inBuff that was not fetched yet by readNextSampleBlock.
                                                  // 0: no data in inBuff;
  nrOfByteFilledInBtrdBuff            : Cardinal; // how many byte have managed it into outBuff that are not written to the outputfile yet?
  nrOfByteFilledInCorrBuff            : Cardinal;  // how many byte have managed it into outBuff that are not written to the outputfile yet?

  sampleByteLeftToRead                : Cardinal;
  samplebytesread                     : Cardinal;
  readbyteCount                       : Cardinal;
  byteslefttoread                     : cardinal;
  thisreadbyteCount                   : Cardinal;

  writebytecount                      : Cardinal;
  thiswritebytecount                  : Cardinal;
  firstsamplesread                    : Cardinal;
  lastwinerror                        : Cardinal;
  lastIOResult                        : Cardinal;
  datachunkpos                        : Cardinal;
  ThisFilePos                         : Cardinal;

  retries                             : Cardinal;
  CantReadInWavAnyMore                : longbool;

  numchunks                           : Cardinal = 0;
  endofchunkmap                       : Cardinal;

  BUFFER_SIZEtouse                    : Cardinal;
  last_find_text                      : integer = -1;

  readtransfer                        : Procedure(pB, pEndB : PByte);
  writetransfer                       : Procedure(var outputcodecblock : MultiChannelCodecBlock; pB : PByte);

  Unit_Data_Packer2                   : array[0..3] of integer;

  {$IFNDEF USEASM}
  //============================================================================
  procedure combinelossybuffers(bytes_in_buffer : Integer);
  var
    cb_i, cb_j : integer;
    this_sample, btrd_sample, corr_sample : integer;
  //============================================================================
  begin
  //============================================================================
    cb_i:=bytes_in_buffer div integer(Global.Bytes_Per_Sample);
    case Global.Bytes_Per_Sample of
      1 : lossyWAVerror('Invalid bitdepth.',$12);
      2 : for cb_j:=0 to cb_i-1 do
            mergebuff.w[cb_j]:=round(btrdbuff.w[cb_j]*Settings.Scaling_Factor_inv+corrbuff.w[cb_j]);
      3 : for cb_j:=0 to cb_i-1 do
          begin
            btrd_sample:=btrdbuff.t[cb_j].hi_w*256 or btrdbuff.t[cb_j].lo_b;
            corr_sample:=corrbuff.t[cb_j].hi_w*256 or btrdbuff.t[cb_j].lo_b;
            this_sample:=round(btrd_sample*Settings.Scaling_Factor_inv+corr_sample);
            mergebuff.t[cb_j].lo_B:=byte(this_sample and 255);
            mergebuff.t[cb_j].hi_W:=this_sample div 63356;
          end;
      4 : for cb_j:=0 to cb_i-1 do
            mergebuff.l[cb_j]:=round(btrdbuff.l[cb_j]*Settings.Scaling_Factor_inv+corrbuff.l[cb_j]);
    end;
  //============================================================================
  end;
  //============================================================================
  {$ELSE}
  //============================================================================
  procedure combinelossybuffers(bytes_in_buffer : Integer);
  //============================================================================
  var
    this_sample, btrd_sample, corr_sample : integer;
  begin
  //============================================================================
    case Global.Bytes_Per_Sample of
      1 : lossyWAVerror('Invalid bitdepth.',$12);

      2 : asm
                Push   ebx
                push   edi
                push   esi

                mov    ebx,Offset BtrdBuff
                mov    edi,Offset CorrBuff
                mov    esi,Offset MergeBuff
                mov    ecx,bytes_in_buffer

      @@00:     sub    ecx,2
                cmp    ecx,0
                jl     @@01

                fild   word ptr [ebx+ecx]
                fmul   Settings.Scaling_Factor_inv
                fiadd  word ptr [edi+ecx]
                fistp  word ptr [esi+ecx]
                jmp    @@00

      @@01:     pop    esi
                pop    edi
                pop    ebx
          end;

      3 : asm
                Push   ebx
                push   edi
                push   esi

                mov    ebx,Offset BtrdBuff
                mov    edi,Offset CorrBuff
                mov    esi,Offset MergeBuff
                mov    ecx,bytes_in_buffer

      @@00:     sub    ecx,3
                cmp    ecx,0
                jl     @@01

                movsx  eax,byte ptr [ebx+ecx+2]
                shl    eax,16
                mov    ax,word ptr [ebx+ecx]
                mov    dword ptr btrd_sample,eax

                movsx  eax,byte ptr [edi+ecx+2]
                shl    eax,16
                mov    ax,word ptr [edi+ecx]
                mov    dword ptr corr_sample,eax

                fild   btrd_sample
                fmul   Settings.Scaling_Factor_inv
                fiadd  corr_sample
                fistp  this_sample

                mov    eax,this_sample
                mov    word ptr [esi+ecx],ax
                shr    eax,16
                mov    byte ptr [esi+ecx+2],al
                jmp    @@00

      @@01:     pop    esi
                pop    edi
                pop    ebx
          end;

      4 : asm
                Push   ebx
                push   edi
                push   esi

                mov    ebx,Offset BtrdBuff
                mov    edi,Offset CorrBuff
                mov    esi,Offset MergeBuff
                mov    ecx,bytes_in_buffer

      @@00:     sub    ecx,4
                cmp    ecx,0
                jl     @@01

                fild   dword ptr [ebx+ecx]
                fmul   Settings.Scaling_Factor_inv
                fiadd  dword ptr [edi+ecx]
                fistp  dword ptr [esi+ecx]
                jmp    @@00

      @@01:     pop    esi
                pop    edi
                pop    ebx
         end;
    end;
  //============================================================================
  end;
  //============================================================================
  {$ENDIF}
  //============================================================================
  procedure openfile(var thisfile : file; thisname : string; thisfilemode : cardinal; orvalue : cardinal);
  //============================================================================
  begin
  //============================================================================
    filemode:=thisfilemode;

    AssignFile(thisfile, thisname);
    if IOResult <> 0 then exit;

    case thisfilemode of
      0 : Reset(thisfile, 1);
      2 : Rewrite(thisfile,1);
    else
      lossyWAVerror('Invalid FileMode!',$12);
    end;
    if IOResult <> 0 then exit;

    filesOpened:=filesopened or orvalue;
  //============================================================================
  end;
  //============================================================================

  //============================================================================
  procedure wavIOExitProc(wavIOString : String; wavIOCode : Integer);
  //============================================================================
  var
    wi : integer;
  begin
  //============================================================================
    for wi:=0 to numchunks-1 do
      write(ErrOutput,chunkmap[wi].Header.ID,',');
    write(ErrOutput,CRLF,wavIOString,CRLF);
    closeWavIO;
    exitcode:=wavIOcode;
    halt;
  //============================================================================
  end;
  //============================================================================

  //============================================================================
  function readfromWAVEfile(buffpointer : pointer; bytestoread : Cardinal) : cardinal;
  //============================================================================
  begin
  //============================================================================
    readbytecount:=0;
    retries:=0;

    if Parameters.STDINPUT then
    while (readbytecount<bytestoread) and (not CantReadInWavAnyMore) do
    begin
      byteslefttoread:=bytestoread-readbytecount;

      if ReadFile(hSTDIN, pointer(cardinal(buffpointer)+readbytecount)^, byteslefttoread, thisreadbytecount,nil) then
        lastwinerror:=0
      else
        lastwinerror:=getlasterror();

      if thisreadbytecount=0 then
      begin
        inc(retries);
        timer.DelayTime:=PowersOf.two[min(7,retries)-11];
        ndelay;
      end
      else
        retries:=0;

      if (lastwinerror=error_broken_pipe) or (retries=32) then
        cantreadinwavanymore:=true;

      readbytecount:=readbytecount+thisreadbytecount;
    end
    else
    begin
      blockread(wavefile,buffpointer^,bytestoread,readbytecount);
      lastIOResult:=IOResult;
      if (lastIOResult<>0) or (readbytecount<bytestoread) then
        cantreadinwavanymore:=true;
    end;

    readfromWAVEfile:=readbytecount;
  //============================================================================
  end;
  //============================================================================

  //============================================================================
  function WordAlign(WAvalue : Cardinal) : Cardinal;
  //============================================================================
  begin
  //============================================================================
    WordAlign:=WAvalue + (WAvalue and 1);
  //============================================================================
  end;
  //============================================================================

  //============================================================================
  function WriteChunks(var outputfile : file) : longbool;
  //============================================================================
  var
    thischunk : Cardinal;
  begin
  //============================================================================
    writechunks:=false;

    if (Parameters.merging or Parameters.stdoutput) and (WAVEChunk.FACT.Header.Size<>0) then
      WAVEChunk.WAV.Header.Size:=WAVEChunk.WAV.Header.Size-(WAVEChunk.FACT.Header.Size+8);

    BlockWrite(outputfile, WAVEChunk.WAV, 12);
    if IOResult <> 0 then exit;

    for thischunk:=0 to numchunks-1 do
    begin
      if chunkmap[thischunk].cptr=-1 then
      begin
        case chunkmap[thischunk].header.cnum of
          $20746D66 : begin //'fmt '
                        BlockWrite(outputfile, WAVEChunk.FMT, WordAlign(WAVEChunk.FMT.Header.Size+8));
                        if IOResult <> 0 then
                          wavIOExitProc('Error writing ''fmt '' chunk.',$12);
                        if (not Parameters.merging) and (WaveChunk.Fact.Header.Size<>0) and (not Parameters.stdoutput) then
                        begin
                          BlockWrite(outputfile, WAVEChunk.FACT, WordAlign(WAVEChunk.FACT.Header.Size+8));
                          if IOResult <> 0 then
                            wavIOExitProc('Error writing ''fact'' chunk.',$12);
                        end;
                      end;
          $74636166 : begin //'fact'
                      end;
        end
      End
      else
      Begin
        Blockwrite(outputfile,chunkmap[thischunk].header,8);
        if IOResult <> 0 then
          wavIOExitProc('Error writing '''+chunkmap[thischunk].header.id+''' chunk.',$12);
        Blockwrite(outputfile,chunkspace^[chunkmap[thischunk].cptr],WordAlign(chunkmap[thischunk].header.size));
        if IOResult <> 0 then
          wavIOExitProc('Error writing '''+chunkmap[thischunk].header.id+''' chunk.',$12);
      end;
    end;

   if not Parameters.stdoutput then
     datachunkpos:=filepos(outputfile);
    BlockWrite(outputfile, WAVEChunk.DATA, 8);
    if IOResult <> 0 then exit;

    writechunks:=true;
  //============================================================================
  end;
  //============================================================================

  //============================================================================
  procedure RewriteDataChunk(var outputfile : file);
  //============================================================================
  begin
  //============================================================================
    thisfilepos:=filepos(outputfile);
    seek(outputfile,0);
    wavechunk.WAV.header.size:=thisfilepos-8;
    blockwrite(outputfile,wavechunk.WAV,8,writeresult);

    seek(outputfile,datachunkpos);
    wavechunk.data.header.size:=samplebytesread;
    blockwrite(outputfile,wavechunk.data,8,writeresult);
  //============================================================================
  end;
  //============================================================================

  //============================================================================
  procedure ReadChunks(var ThisFile : File; var ThisChunk : tlossyWAVChunk);
  //============================================================================
  begin
  //============================================================================
    fillchar(thischunk,sizeof(thischunk),0);
    numchunks:=0;
    endofchunkmap:=0;

    BlockRead(ThisFile,ChunkHeader,8,readresult);
    if readresult<>8 then
      wavIOExitProc('Error Reading WAV file Information.',$11);

    if chunkheader.id='RIFF' then
    begin
      move(chunkheader,thischunk.wav,8);
      blockread(thisfile,thischunk.wav.rifftype,4,readresult);
      if (readresult<>4) then
        wavIOExitProc('Error Reading WAV file Information.',$11);
    end
    else
      wavIOexitProc('File type is not RIFF.',$12);

    if ThisChunk.WAV.RIFFType <> 'WAVE' then
      wavIOExitProc('RIFF type is not WAVE.',$12);

    repeat
      BlockRead(ThisFile,ChunkHeader,8,readresult);
      if readresult<>8 then
        wavIOExitProc('Error Reading WAV file Information.',$11);

      case chunkheader.cnum of
        $20746D66 : begin //'fmt '
                      chunkmap[numchunks].Header:=chunkheader;
                      chunkmap[numchunks].Cptr:=-1;
                      inc(numchunks);
                      move(chunkheader,ThisChunk.FMT,8);
                      Case ThisChunk.FMT.Header.Size of
                        16,18,40 : Begin
                                   End;
                      else
                        WavIOExitProc('FMT Chunk incorrect size.',$12);
                      End;

                      blockread(ThisFile,ThisChunk.FMT.wFormatTag,WordAlign(thischunk.FMT.Header.Size),readresult);
                      if readresult<>wordalign(thischunk.FMT.Header.Size) then
                        wavIOExitProc('Error reading FMT chunk.',$12);

//                      if (ThisChunk.FMT.wFormatTag=1) and (ThisChunk.FMT.wBitsPerSample=40) then
//                      begin
//                        thischunk.fmt.Samples.wValidBitsPerSample:=thischunk.fmt.wBitsPerSample;
//                        thischunk.fmt.wFormatTag:=$FFFE;
//                        ThisChunk.FMT.SubFormat_Word[0]:=1;
//                      end;
                    end;
        $61746164 : begin //'data'
                      Move(ChunkHeader,ThisChunk.DATA,8);
                    end;
      else
        chunkmap[numchunks].Header:=chunkheader;
        chunkmap[numchunks].Cptr:=EndofChunkMap;
        If endofchunkmap>sizeof(chunkspace) then
          wavIOExitProc('Too much WAV information before ''data'' chunk!',$12);
        blockread(thisfile,chunkspace^[endofchunkmap],WordAlign(chunkheader.size),readresult);
        if readresult<>wordalign(chunkheader.size) then
          wavIOExitProc('Error Reading '''+chunkheader.ID+''' Chunk.',$12);
        endofchunkmap:=endofchunkmap+wordalign(chunkheader.size);

        if (chunkmap[numchunks].Header.Id='fact') then
        begin
          ThisChunk.FACT.Header:=ChunkMap[numchunks].Header;
          if thischunk.fact.header.size<factchunkMaxsize then
            move(chunkspace^[chunkmap[numchunks].Cptr],thischunk.fact.data,thischunk.fact.header.size);
          if copy(thischunk.fact.data,1,8)<>'lossyWAV' then
            thischunk.fact.header.size:=0
          else
            chunkmap[numchunks].cptr:=-1;
        end;

        inc(numchunks);
      end;
    until ChunkHeader.ID='data';

    if (ThisChunk.FMT.wFormatTag<>1) and ((ThisChunk.FMT.wformatTag=$FFFE) and (ThisChunk.FMT.SubFormat_Word[0]<>1)) then
      wavIOExitProc('WAV Data is not uncompressed PCM.',$12);
  //============================================================================
  end;
  //============================================================================

  //============================================================================
  function find_text(find_str : string; source_str : string) : integer;
  //============================================================================
  var
    ft_i : Integer;
    ft_l : Integer;
  begin
  //============================================================================
    ft_i:=0;
    ft_l:=length(find_str);

    if ft_l=0 then
      ft_i:=-1
    else
    begin
      repeat
        inc(ft_i);
      until (ft_i=(length(source_str)-ft_l)) or (find_str=copy(source_str,ft_i,ft_l));

      if ft_i=(length(source_str)-ft_l) then
        ft_i:=-1
    end;
    find_text:=ft_i;
    last_find_text:=ft_i;
  //============================================================================
  end;
  //============================================================================

  //============================================================================
  procedure MergeFiles;
  //============================================================================
  var
    thisread   : cardinal;
    numreads   : cardinal;
    lastread   : cardinal;
    readcount  : cardinal;
    this_error : Integer;
    temp_str   : String;
  begin
  //============================================================================
    filesOpened:=0;

    openfile(BTRDFile,Parameters.WAVInpDir + Parameters.lossyName,0,2);
    ReadChunks(BTRDFile,BTRDChunk);

    openfile(CORRFile,Parameters.WAVInpDir + Parameters.lwcdfName,0,2);
    ReadChunks(CORRFile,CORRChunk);

    write(ErrOutput,CRLF,'Merging: ',Parameters.lossyname,CRLF);
    write(ErrOutput,     '       & ',Parameters.lwcdfname,CRLF);

    if btrdchunk.wav.Header.Size<>corrchunk.wav.Header.Size then
      lossyWAVerror('Files not same length.',$12);

    if btrdchunk.data.Header.Size<>corrchunk.data.Header.Size then
      lossyWAVerror('Data not same length.',$12);

    if (btrdchunk.fact.header.size=0) then
      lossyWAVerror('FACT Data missing (lossy.wav).',$12);

    if (corrchunk.fact.header.size=0) then
      lossyWAVerror('FACT Data missing (lwcdf.wav).',$12);

    if btrdchunk.fact.data<>corrchunk.fact.data then
      lossyWAVerror('FACT Data mismatch.',$12);

    if copy(btrdchunk.fact.data,1,8)<>'lossyWAV' then
      lossyWAVerror('Not a lossyWAV FACT chunk.',$12);

    if (btrdchunk.fmt.Header.Size<>corrchunk.fmt.Header.Size) or
       (btrdchunk.fmt.wFormatTag<>corrchunk.fmt.wFormatTag) or
       (btrdchunk.fmt.wChannels<>corrchunk.fmt.wChannels) or
       (btrdchunk.fmt.nSamplesPerSec<>corrchunk.fmt.nSamplesPerSec) or
       (btrdchunk.fmt.nAvgBytesPerSec<>corrchunk.fmt.nAvgBytesPerSec) or
       (btrdchunk.fmt.nblockAlign<>corrchunk.fmt.nBlockAlign) or
       (btrdchunk.fmt.wBitsPerSample<>corrchunk.fmt.wBitsPerSample) then
      lossyWAVerror('FMT Data mismatch.',$12);

    Global.Bytes_Per_Sample:=(btrdchunk.fmt.wBitsPerSample+7) shr 3;
    if Global.Bytes_Per_Sample<2 then
      lossyWAVerror('Invalid bitdepth: '+inttostr(btrdchunk.fmt.wBitsPerSample),$12);

    lastread:=btrdChunk.Data.Header.Size mod BUFFER_SIZE;
    numreads:=btrdchunk.data.Header.Size div BUFFER_SIZE+cardinal(lastread>0);

    move(btrdchunk,wavechunk,sizeof(btrdchunk));
    openfile(WAVEFile,Parameters.WAVOutDir+Parameters.WAVName,2,1);
    if not writechunks(WAVEfile) then
      lossyWAVerror('Writing to output file.',$12);

    if find_text('--scale ',wavechunk.fact.data)<>-1 then
      temp_str:=copy(wavechunk.fact.data,last_find_text+7,8)
    else
      temp_str:='1.000000';

    val(temp_str,Settings.Scaling_Factor,this_error);
    if (this_error<>0) or (Settings.Scaling_Factor<=0) then
      lossyWAVerror('Invalid scale value: '+temp_str,$12);
    Settings.Scaling_Factor_inv:=1/Settings.Scaling_Factor;
    BUFFER_SIZEtouse:=BUFFER_SIZE-(BUFFER_SIZE mod (Global.Bytes_Per_Sample*btrdchunk.fmt.wChannels));

    for readcount:=0 to numreads-1 do
    begin
      thisread:=min(BUFFER_SIZEtouse,btrdchunk.data.Header.Size-Readcount*BUFFER_SIZEtouse);
      blockread(BTRDFile,BTRDBuff,thisread,readresult);
      blockread(CORRFile,CORRBuff,thisread,readresult);
      readbytecount:=readbytecount+readresult;
      combinelossybuffers(thisread);
      blockwrite(WAVEFile,MERGEBuff,thisread,readresult);
      write(ErrOutput,#13,'Progress: ',readbytecount/(btrdchunk.wav.Header.Size+8)*100:0:2,'%    ');
    end;

    close(CORRFile);
    readbytecount:=filepos(btrdfile);
    readresult:=65536;

    while readresult=65536 do
    begin
      blockread(BTRDFile,BTRDBuff,65536,readresult);
      blockwrite(WAVEFile,btrdbuff,readresult);
      readbytecount:=readbytecount+readresult;
      write(ErrOutput,#13,'Progress: ',readbytecount/(btrdchunk.wav.Header.Size+8)*100:0:2,'%    ');
    end;

    write(ErrOutput,CRLF);

    close(BTRDFile);
    close(WAVEFile);
  //============================================================================
  end;
  //============================================================================

  //============================================================================
  {$IFDEF USEASM}
  //============================================================================
  procedure readtransfer_two(pB,pEndB : pByte);
  //============================================================================
  asm
                push   ebx
                push   edi
                push   esi

                mov    edi,dword ptr BlockData.Pointer[3*pointer_size]
                mov    ecx,pB
                mov    esi,pEndB

      @@20:     sub    ebx,ebx
                sub    edx,edx

      @@21:     movsx  eax,word ptr [ecx]
                mov    dword ptr [edx+edi],eax

                add    ecx,2
                add    edx,CHANNEL_BYTE_SIZE
                add    ebx,1
                cmp    bx,WAVEChunk.FMT.wChannels
                jl     @@21

                add    edi,INTEGER_SIZE
                cmp    ecx,esi
                jl     @@20

                pop    esi
                pop    edi
                pop    ebx
  end;
  //============================================================================

  //============================================================================
  procedure readtransfer_three(pB,pEndB : pByte);
  //============================================================================
  asm
                push   ebx
                push   edi
                push   esi

                mov    edi,dword ptr BlockData.Pointer[3*pointer_size]
                mov    ecx,pB
                mov    esi,pEndB

      @@30:     sub    ebx,ebx
                sub    edx,edx

      @@31:     movsx  eax,word ptr [ecx+1]
                shl    eax,8
                mov    al,byte ptr [ecx]
                mov    dword ptr [edx+edi],eax

                add    ecx,3
                add    edx,CHANNEL_BYTE_SIZE
                add    ebx,1
                cmp    bx,word ptr WAVEChunk.FMT.wChannels
                jl     @@31

                add    edi,INTEGER_SIZE
                cmp    ecx,esi
                jl     @@30

                pop    esi
                pop    edi
                pop    ebx
  end;
  //============================================================================

  //============================================================================
  procedure readtransfer_four(pB,pEndB : pByte);
  //============================================================================
  asm
                push   ebx
                push   edi
                push   esi

                mov    edi,dword ptr BlockData.Pointer[3*pointer_size]
                mov    ecx,pB
                mov    esi,pEndB

      @@40:     sub    ebx,ebx
                sub    edx,edx

      @@41:     mov    eax,[ecx]
                mov    dword ptr [edx+edi],eax

                add    ecx,4
                add    edx,CHANNEL_BYTE_SIZE
                add    ebx,1
                cmp    bx,word ptr WAVEChunk.FMT.wChannels
                jl     @@41

                add    edi,INTEGER_SIZE
                cmp    ecx,esi
                jl     @@40

                pop    esi
                pop    edi
                pop    ebx
  end;
  //============================================================================
  {$ELSE}
  //============================================================================
  procedure readtransfer_two(pB,pEndB : pByte);
  //============================================================================
  var
    pS: PSmallInt;
    iSample: integer;
    iChannel : Integer;
    thissample : integer;
  begin
  //============================================================================
    iSample:=0;
    while pB <> pEndB do
    begin
      for iChannel:=0 to WAVEChunk.FMT.wChannels - 1 do
      begin
        pS:=pSmallInt(pB);
        inc(pb,2);
        thissample:=pS^;
        BlockData.Pointer[3]^[iChannel, iSample]:=thissample;
      end;
      iSample:=iSample + 1;
    end;
  //============================================================================
  end;
  //============================================================================

  //============================================================================
  procedure readtransfer_three(pB,pEndB : pByte);
  //============================================================================
  var
    pS: pSmallInt;
    iSample: integer;
    iChannel : Integer;
    thissample : integer;
  begin
  //============================================================================
    iSample:=0;
    while pB <> pEndB do
    begin
      for iChannel:=0 to WAVEChunk.FMT.wChannels - 1 do
      begin
        pS:=pSmallint(integer(pB)+1);
        thissample:=pS^*256;
        thissample:=thissample or pB^;
        inc(pB,3);
        BlockData.Pointer[3]^[iChannel, iSample]:=thissample;
      end;
      iSample:=iSample + 1;
    end;
  end;
  //============================================================================

  //============================================================================
  procedure readtransfer_four(pB,pEndB : pByte);
  //============================================================================
  var
    pI: PInteger;
    iSample: integer;
    iChannel : Integer;
    thissample : integer;
  begin
  //============================================================================
    iSample:=0;
    while pB <> pEndB do
    begin
      for iChannel:=0 to WAVEChunk.FMT.wChannels - 1 do
      begin
        pI:=pInteger(pB);
        inc(pB,4);
        thissample:=pI^;
        BlockData.Pointer[3]^[iChannel, iSample]:=thissample;
      end;
      iSample:=iSample + 1;
    end;
  end;
  //============================================================================
  {$ENDIF}
  //============================================================================
  function readNextNextCodecBlock : longbool;
  //============================================================================
  var
    pB, pStartB: PByte;
    thisblockreadlength : cardinal;
  begin
  //============================================================================
    readNextnextCodecBlock:=false;
    BlockData.Size.Next:=0;

    if cantreadinwavanymore then
      exit;

    if nrOfBlockInBuffNotYetFetched = 0 then // samples to be fetched aren't in inbuff
    begin
      if (nrOfFullBlocksReadIntoInBuff < nrOfBlocksInReadBuff) or (sampleByteLeftToRead=0) then // EOF encountered with last call to readNextSampleBlock
        exit; // no more data

      thisblockreadlength:=min(BUFFER_SIZEread,samplebytelefttoread);
      readbytecount:=0;
      retries:=0;

      readfromWAVEfile(@WAVBuff,thisblockreadlength);

      sampleByteLeftToRead:=sampleByteLeftToRead-readByteCount;
      samplebytesread:=samplebytesread+readbytecount;

      nrOfFullBlocksReadIntoInBuff:=readbyteCount DIV nrOfByteInOneBlockInBuff;
      nrOfByteOfTheLastIncompleteBlock:=integer(readbyteCount) mod - nrOfByteInOneBlockInBuff;

      nrOfBlockInBuffNotYetFetched:=1;
    end;

    pB:=@WAVBuff;
    Inc(pB, (nrOfBlockInBuffNotYetFetched - 1) * nrOfByteInOneBlockInBuff);
    pStartB:=pB; // points to first sample in Block

    if nrOfBlockInBuffNotYetFetched <= nrOfFullBlocksReadIntoInBuff then
    begin
      Inc(pB, nrOfByteInOneBlockInBuff);  // pB points to the spot after last sample in Block
      BlockData.Size.Next:=nrOfByteInOneBlockInBuff DIV (Cardinal(Global.Bytes_Per_Sample*Global.Channels));
      nrOfBlockInBuffNotYetFetched:=nrOfBlockInBuffNotYetFetched + 1;
      if nrOfBlockInBuffNotYetFetched > nrOfBlocksInReadBuff then
        nrOfBlockInBuffNotYetFetched:=0;
    end
    else
    begin
      if nrOfByteOfTheLastIncompleteBlock = 0 then
        exit; // no more data
      Inc(pB, nrOfByteOfTheLastIncompleteBlock);  // pB points to the spot after last sample in Block
      BlockData.Size.Next:=nrOfByteOfTheLastIncompleteBlock DIV (Cardinal(Global.Bytes_Per_Sample*Global.Channels));
      nrOfBlockInBuffNotYetFetched:=0; // force a try for a new Blockread with next call which will lead to a return False
    end;
    readtransfer(pStartB,pB);
    readNextNextCodecBlock:=true;
  //============================================================================
  end;
  //============================================================================
  {$IFDEF USEASM}
  //============================================================================
  Procedure WriteTransfer_Two(var outputcodecblock : MultiChannelCodecBlock; pB : PByte);
  //============================================================================
  Asm
                push   ebx
                push   edi
                push   esi

                mov    ecx,pb
                mov    edi,outputcodecblock
                mov    esi,BlockData.Size.This
                shl    esi,INTEGER_SHIFT
                add    esi,edi

      @@00:     sub    ebx,ebx
                sub    edx,edx

      @@01:     mov    eax,dword ptr [edx+edi]
                mov    word ptr [ecx],ax

                add    ecx,2
                add    edx,CHANNEL_BYTE_SIZE
                add    ebx,1
                cmp    bx,WAVEChunk.FMT.wChannels
                jl     @@01

                add    edi,INTEGER_SIZE
                cmp    edi,esi
                jl     @@00

                pop    esi
                pop    edi
                pop    ebx
  End;
  //============================================================================

  //============================================================================
  Procedure WriteTransfer_Three(var outputcodecblock : MultiChannelCodecBlock; pB : PByte);
  //============================================================================
  Asm
                push   ebx
                push   edi
                push   esi

                mov    ecx,pb
                mov    edi,outputcodecblock
                mov    esi,BlockData.Size.This
                shl    esi,INTEGER_SHIFT
                add    esi,edi

      @@00:     sub    ebx,ebx
                sub    edx,edx

      @@01:     mov    eax,dword ptr [edx+edi]
                mov    [ecx],word ptr ax
                shr    eax,8
                mov    [ecx+2],ah

                add    ecx,3
                add    edx,CHANNEL_BYTE_SIZE
                add    ebx,1
                cmp    bx,WAVEChunk.FMT.wChannels
                jl     @@01

                add    edi,INTEGER_SIZE
                cmp    edi,esi
                jl     @@00

                pop    esi
                pop    edi
                pop    ebx
  End;
  //============================================================================

  //============================================================================
  Procedure WriteTransfer_Four(var outputcodecblock : MultiChannelCodecBlock; pB : PByte);
  //============================================================================
  Asm
                push   ebx
                push   edi
                push   esi

                mov    ecx,pb
                mov    edi,outputcodecblock
                mov    esi,BlockData.Size.This
                shl    esi,INTEGER_SHIFT
                add    esi,edi

      @@00:     sub    ebx,ebx
                sub    edx,edx

      @@01:     mov    eax,dword ptr [edx+edi]
                mov    dword ptr [ecx],eax

                add    ecx,4
                add    edx,CHANNEL_BYTE_SIZE
                add    ebx,1
                cmp    bx,WAVEChunk.FMT.wChannels
                jl     @@01

                add    edi,INTEGER_SIZE
                cmp    edi,esi
                jl     @@00

                pop    esi
                pop    edi
                pop    ebx
  End;
  //============================================================================
  {$ELSE}
  //============================================================================
  Procedure WriteTransfer_All(var outputcodecblock : MultiChannelCodecBlock; pB : PByte);
  //============================================================================
  Var
    wt_i, wt_j : Integer;
    this_sample : Integer;
  Begin
  //============================================================================
    For wt_i:=0 to BlockData.Size.This-1 do
      for wt_j:=0 to WaveChunk.FMT.wChannels-1 do
      begin
        this_sample:=outputcodecblock[wt_j,wt_i];
        pB^:=this_sample and 255;
        inc(pb);
        if Global.Bytes_Per_Sample>1 then
        begin
          this_sample:=this_sample shr 8;
          pb^:=this_sample and 255;
          inc(pb);
          if Global.Bytes_Per_Sample>2 then
          begin
            this_sample:=this_sample shr 8;
            pb^:=this_sample and 255;
            inc(pb);
            if Global.Bytes_Per_Sample>3 then
            begin
              this_sample:=this_sample shr 8;
              pb^:=this_sample and 255;
              inc(pb);
            end;
          end;
        end;
      end;
  //============================================================================
  End;
  //============================================================================
  {$ENDIF}
  //============================================================================
  function writeNextCodecBlock(var outputfile : file; var OutBuff; var OutputCodecBlock : MultiChannelCodecBlock; var FilledOutputBuffer : Cardinal): longbool;
  //============================================================================
  var
    nrOfByteForOutBuff : Cardinal;
    pB: PByte;
  begin
  //============================================================================
    writeNextCodecBlock:=false;
    nrOfByteForOutBuff:=integer(Global.Bytes_Per_Sample) * WAVEChunk.FMT.wChannels * BlockData.Size.This;
    writebytecount:=0;

    if FilledOutputBuffer + nrOfByteForOutBuff > BUFFER_SIZEwrite then // not sufficient space in outBuff : flush outBuff
    begin
      BlockWrite(outputfile, OutBuff, FilledOutputBuffer);
      if IOResult <> 0 then
        exit;
      FilledOutputBuffer:=0;
    end;

    pB:=@OutBuff;
    Inc(pB, FilledOutputBuffer); // points to first free spot in outBuff

    WriteTransfer(outputcodecblock,pb);

    FilledOutputBuffer:=FilledOutputBuffer + nrOfByteForOutBuff;
    writeNextCodecBlock:=true;
  //============================================================================
  end;
  //============================================================================

  //============================================================================
  function writeNextBtrdCodecBlock: longbool;
  //============================================================================
  begin
  //============================================================================
    WriteNextBtrdCodecBlock:=WriteNextCodecBlock(BtrdFile,BtrdBuff,BlockData.BTRD,nrofByteFilledInBtrdBuff);
  //============================================================================
  end;
  //============================================================================

  //============================================================================
  function writeNextCorrCodecBlock : longbool;
  //============================================================================
  begin
  //============================================================================
    WriteNextCorrCodecBlock:=WriteNextCodecBlock(CorrFile,Corrbuff,BlockData.CORR,nrofByteFilledInCorrBuff);
  //============================================================================
  end;
  //============================================================================

  //============================================================================
  function closeWavIO: longbool;
  //============================================================================
  begin
  //============================================================================
    closeWavIO:=false;

    if filesOpened>1 then
    begin
      if (filesopened and 2=2) and (nrOfByteFilledInBtrdBuff > 0) then
      begin
        BlockWrite(BtrdFile, BtrdBuff, nrOfByteFilledInBtrdBuff);
        if filepos(Btrdfile) and 1=1 then
          BlockWrite(BtrdFile,Unit_Data_Packer2[0],1);
        if IOResult <> 0 then
          exit;
      end;

      if (filesopened and 4=4) and (nrOfByteFilledInCorrBuff > 0) then
      begin
        BlockWrite(CorrFile, CorrBuff, nrOfByteFilledInCorrBuff);
        if filepos(Corrfile) and 1=1 then
          BlockWrite(CorrFile,Unit_Data_Packer2[0],1);
        if IOResult <> 0 then
          exit;
      end;

      While not(eof(wavefile)) and (not CantReadInWAVAnyMore) do
      Begin
        if (readfromWAVEfile(@WAVBuff, BUFFER_SIZEread)<cardinal(BUFFER_SIZEread)) or (IOResult <> 0) then
          CantReadInWavAnyMore:=true;

        if readbytecount<>0 then
        Begin
          if (filesopened and 2=2) then
          begin
            BlockWrite(BtrdFile, WAVBuff, readbyteCount);
            if IOResult <> 0 then
              exit;
          end;

          if filesopened and 4=4 then
          Begin
            BlockWrite(CorrFile, WAVBuff, readbyteCount);
            if IOResult <> 0 then
              exit;
          End;
        End;
      End;

      closeFile(WAVEFile);
      if IOResult <> 0 then
        exit;
      filesopened:=filesopened xor 1;

      if filesopened and 2=2 then
      begin
        if filepos(Btrdfile) and 1=1 then
          BlockWrite(BtrdFile,Unit_Data_Packer2[0],1);
        if (Parameters.stdinput) and (not Parameters.stdoutput) then
          rewritedatachunk(BtrdFile);
        CloseFile(BtrdFile);
        if IOResult <> 0 then
          exit;
        filesopened:=filesopened xor 2;
      end;

      if filesopened and 4=4 then
      begin
        if filepos(Corrfile) and 1=1 then
          BlockWrite(CorrFile,Unit_Data_Packer2[0],1);
        if (Parameters.stdinput) and (not Parameters.stdoutput) then
          rewritedatachunk(CorrFile);
        closeFile(CorrFile);
        if ioresult<>0 then
          exit;
        filesopened:=filesopened xor 4;
      end;
    end;

    closeWavIO:=true;
  //============================================================================
  end;
  //============================================================================

  //============================================================================
  function openWavIO : Longbool;
  //============================================================================
  const
  {$IFDEF USEASM}
    writetransferprocs : array[2..4] of procedure(var outputcodecblock : MultiChannelCodecBlock; pB : PByte) = (writetransfer_two,writetransfer_three,writetransfer_four);
  {$ELSE}
    writetransferprocs : array[2..4] of procedure (var outputcodecblock : MultiChannelCodecBlock; pB : PByte) = (writetransfer_all,writetransfer_all,writetransfer_all);
  {$ENDIF}
    readtransferprocs : array[2..4] of procedure(pB,pEndB : pByte) = (readtransfer_two,readtransfer_three,readtransfer_four);
  begin
  //============================================================================
    OpenWavIO:=False;
    filesOpened:=0;
    Global.Codec_Block_size:=0;

    openfile(WAVEFile,Parameters.WAVInpDir+Parameters.WAVName,0,1);
    ReadChunks(WAVEFile,WAVEChunk);
    if wavechunk.data.header.size=0 then
      wavIOexitproc('Zero sample WAV file!?!',$12);

    Global.Bits_Per_Sample:=WAVEChunk.FMT.wBitsPerSample;

    if (Global.Bits_Per_Sample<9) or (Global.Bits_Per_Sample>32) then
    begin
      write(ErrOutput,'Input file has a bitdepth of ' + IntToStr(Global.Bits_Per_Sample) + ' bits per sample.'+CRLF);
      wavIOExitProc('This is not supported.',$12);
    end;
    Global.Bytes_Per_Sample:=(Global.Bits_Per_Sample+7) shr 3;

    readtransfer:=readtransferprocs[Global.Bytes_Per_Sample];
    writetransfer:=writetransferprocs[Global.Bytes_Per_Sample];

    Global.Channels       :=WAVEChunk.FMT.wChannels;
    Global.Wave_Size      :=WAVEChunk.WAV.Header.Size;
    Global.Sample_Rate    :=WAVEChunk.FMT.nSamplesPerSec;
    Global.Total_Samples  :=floor(1.0*Global.Wave_Size/Global.Channels/Global.Bytes_Per_Sample);
    Global.sample_rate_recip:=1/Global.sample_rate;

    if Global.Sample_Rate<22050 then
    begin
      write(ErrOutput,'Sample Rate too low : '+floattostrf(Global.sample_rate*oneover[1000],ffFixed,99,2)+'kHz (Min=22.05kHz).',CRLF);
      wavIOExitProc('This is not supported.',$12);
    end;

    if Global.Sample_Rate>384000 then
    begin
      write(ErrOutput,'Sample Rate too high : '+floattostrf(Global.Sample_Rate*oneover[1000],ffFixed,99,2)+'kHz (Max=384.001kHz).',CRLF);
      wavIOExitProc('This is not supported.',$12);
    end;

    //~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    // Codec_Block_Size to be set to power-of-two window length closest to 10ms.
    //~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    Global.Codec_Block_Duration:=OneOver[100];                                //desired codec-block length = 10msec
    Global.Codec_Block_Bits:=max(8,round(log2(Global.Codec_Block_duration*Global.Sample_Rate)));
    Global.Codec_Block_Size:=Powersof.TwoInt[Global.Codec_Block_Bits];
    Global.Codec_Block_Bit_Shift:=Global.Codec_Block_Bits-9;
    Global.Codec_Block_Duration:=Global.Codec_Block_Size/Global.Sample_Rate;  //actual codec-block length
    //~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

    //~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    If Global.Codec_Block_Bits>MAX_FFT_BIT_LENGTH-1 then
      write(ErrOutput,'Sample Rate too high : '+floattostrf(Global.Sample_Rate*oneover[1000],ffFixed,99,1)+'kHz (Max=409.6kHz).',CRLF);
    //~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

    sampleByteLeftToRead:=WAVEChunk.DATA.Header.Size;
    nrOfByteInOneBlockInBuff:=Global.Bytes_Per_Sample * Global.Channels * Global.Codec_Block_size;

    if Parameters.stdinput then
      nrOfBlocksInReadBuff:=1
    else
      nrOfBlocksInReadBuff:=(BUFFER_SIZE div nrOfByteInOneBlockInBuff);

    if Parameters.stdoutput then
      nrOfBlocksInWriteBuff:=1
    else
      nrOfBlocksInWriteBuff:=(BUFFER_SIZE div nrOfByteInOneBlockInBuff);

    BUFFER_SIZEread :=nrOfByteInOneBlockInBuff * nrOfBlocksInReadBuff;
    BUFFER_SIZEwrite:=nrOfByteInOneBlockInBuff * nrOfBlocksInWriteBuff;

    CantReadInWavAnyMore:=false;
    //==========================================================================

    if Parameters.checking then
    begin
      if (WAVEChunk.FACT.Header.Size>0) and (copy(WAVEChunk.FACT.Data,1,8)='lossyWAV') then
      begin
        exitcode:=$10;
        if (Parameters.verbosity) and (not Parameters.silent) then
        begin
          write(ErrOutput,'lossyWAV FACT Chunk found. File already processed.',CRLF);
          write(ErrOutput,WAVEChunk.FACT.Data);
        end;
      end
      else
      begin
        exitcode:=$00;
        if (Parameters.verbosity) and (not Parameters.silent) then
          write(ErrOutput,'lossyWAV FACT Chunk not found. File not marked as processed.',CRLF);
      end;
      closeWAVio;
      halt;
    end
    else
    begin
      if copy(WAVEChunk.FACT.data,1,8)='lossyWAV' then
      begin
        write(ErrOutput,WAVEChunk.FACT.Data);
        lossyWAVerror('lossyWAV FACT Chunk found. File already processed.',$10);
      end;
    end;

    //==========================================================================
    //create 'fact' chunk data string.
    //==========================================================================
    datetimetostring(Strings.DateStamp,'',timer.StartTime);
    WAVEChunk.FACT.Header.ID:='fact';
    DateTimeString:=version_string+strings.version_short+' @ '+Strings.DateStamp+', '+Strings.Parameter+CRLF+#0;
    if length(DateTimestring) and 1=1 then
      DateTimestring:=DateTimestring+#0;
    move(DateTimestring[1],WAVEChunk.FACT.data,length(DateTimestring));
    WAVEChunk.FACT.Header.Size:=length(DateTimestring);

    if wavechunk.wav.Header.size<MAX_WAVE_SIZE then
      inc(WAVEChunk.WAV.Header.Size,WAVEChunk.FACT.Header.Size+8);
    //==========================================================================

    firstsamplesread:=min(samplebytelefttoread,BUFFER_SIZEread);
    if (readfromWAVEfile(@WAVBuff, firstsamplesread)<firstsamplesread) or (cantreadinwavanymore) then
    begin
      CantReadInWavAnyMore:=true;
      exit;
    end;

    samplebytesread:=readbytecount;
    sampleByteLeftToRead:=sampleByteLeftToRead - readByteCount;
    nrOfFullBlocksReadIntoInBuff:=readbyteCount DIV nrOfByteInOneBlockInBuff;
    nrOfByteOfTheLastIncompleteBlock:=readbyteCount - nrOfFullBlocksReadIntoInBuff * nrOfByteInOneBlockInBuff;
    nrOfBlockInBuffNotYetFetched:=1;
    nrOfByteFilledInBtrdBuff:=0;
    nrOfByteFilledInCorrBuff:=0;

    //==========================================================================
    openfile(BtrdFile,Parameters.WAVOutDir + Parameters.lossyName,2,2);
    if not WriteChunks(btrdfile) then
      lossyWAVerror('Writing to output file.',$12);

    if (not Parameters.STDOUTPUT) and (Parameters.correction) then
    begin
      openfile(CorrFile,Parameters.WAVOutDir + Parameters.lwcdfName,2,4);
      if not WriteChunks(Corrfile) then
        lossyWAVerror('Writing to output file.',$12);
    end;
    OpenWavIO:=true;
    //==========================================================================

  //============================================================================
  end;
  //============================================================================

  //============================================================================
  Procedure nWAV_Init;
  //============================================================================
  Var
    co_i : Integer;
  begin
  //============================================================================
    new(chunkspace);

    for co_i:=0 to MAX_CHANNELS-1 do
      BlockData.CHANNEL_OFFSET[co_i]:=co_i*CHANNEL_BYTE_SIZE;
  //============================================================================
  end;
  //============================================================================

//==============================================================================
{$I+}
//==============================================================================
initialization
//==============================================================================
  Unit_Data_Packer2[0]:=0;
  {$IFDEF DISPLAY_VAR_DATA}
  write_unitname_and_datapointers(UNITNAME,@Unit_Data_Packer1,@Unit_Data_Packer2,sizeof(Unit_Data_Packer2));
  {$ENDIF}

  nWAV_Init;
//==============================================================================

//==============================================================================
finalization
//==============================================================================
  {$IFDEF WRITE_FINAL}
  Write(Erroutput,UNITNAME,CRLF);
  {$ENDIF}
  dispose(chunkspace);
//==============================================================================

//==============================================================================
end.
//==============================================================================

